%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXC:COMPILER.SL
% Description:  PSL Compiler.
% Author:       Harold Carr & Leigh Stoller
% Created:
% Modified:
% Mode:         Lisp
% Package:
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1987, University of Utah, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%
% This file loads in all the compiler modules for Sun4.
%

(compiletime (load fasl-decls))

(load datamachine
      pass-1
      pass-2-3
      comp-decls
      pass-1-lap
      solaris-lap
      solaris-cmac
      solaris-comp
%     backquote         % Needed for solaris LAP patterns.
      registers
      solaris-instrs
      lapopt
      faslout
      solaris-spec)

(de codefiletrailer ()
  (prog (s)
        (systemfaslfixup)
        (binarywrite codeout* (idifference (isub1 nextidnumber*)
                                           first-local-id-number))
        % Number of local IDs
        (foreach x in (car orderedidlist*) do
                 (progn (remprop x fasl-idnumber-property*)
                        (setq x (strinf (faslid2string x)))
                        (setq s (strlen x))
                        (binarywrite codeout* s)
                        (binarywriteblock codeout* (strbase x)
                                          (strpack s))))
        (binarywrite codeout* % S is size in words
(TIMES2 2
                     (setq s
                      (iquotient
                       (iplus2 currentoffset*
                        (isub1 addressingunitsperitem))
                       addressingunitsperitem))))
        (binarywrite codeout* initoffset*)
        (binarywriteblock codeout* codebase* s)
        (binarywrite codeout*
                    (setq s
                      (iquotient
                       (iplus2 bittableoffset*
                        (isub1 bittable-entries-per-word))
                       bittable-entries-per-word)))
        (binarywriteblock codeout* bittablebase* s)
        (deallocatefaslspaces)))

(de AppendOneConstant (ExpressionLabelPair)
        (unless !*writingasmfile
         (AddCodeLabel "double-word-alignment")) % W. N 22.1.96
         (AddCodeLabel (cdr ExpressionLabelPair))
         (AppendItem (car ExpressionLabelPair)))

(de AppendContents (ExpressionLabelPair)
  (prog (Expression UpperBound I)

        (when (and (FloatP (car ExpressionLabelPair)) (prop 'oddalign))
                (AddCode '(oddalign))) % odd double align floats

       (unless !*writingasmfile
        (AddCodeLabel "double-word-alignment")) % W. N 22.1.96
	(AddCodeLabel (cdr ExpressionLabelPair))
	(setq Expression (car ExpressionLabelPair))
	(cond ((PairP Expression)
	       (progn (AppendItem (car Expression))
		      (AppendItem (cdr Expression))))
	      ((StringP Expression)
	       (progn (AddFullWord (Size Expression))
		      (AddCode (list 'STRING Expression))))
	      ((or (VectorP Expression) (EVectorP Expression))
	       (progn (setq UpperBound (ISizeV Expression))
		      (AddFullWord UpperBound)
		      (setq I 0)
		      (while (ILEQ I UpperBound)
			     (progn (AppendItem (IGetV Expression I))
				    (setq I (IAdd1 I))))))
	      ((BigP Expression)
	       (progn (setq UpperBound (ISizeV Expression))
		      (AddFullWord UpperBound)
		      (AppendItem (IGetV Expression 0)) % sign is an id
                      (setq I 1)
		      (while (ILEQ I UpperBound)
%		     (progn (AppendItem (IGetV Expression I))
                          (progn (AddFullWord (sys2fixn (IGetV Expression I)))
				    (setq I (IAdd1 I))))))
	      ((FixP Expression)
	       (progn (AddFullWord 0)	% Header of full word fixnum.
		      (AddFullWord Expression)))
	      ((FloatP Expression)
	       (progn (AddFullWord 1)	% Header of float.
		      (AddCode (list 'FLOAT Expression)))))))


(de sys2fixn (n)
  (prog (fx)
        (setf savefixn-tag (tag n))
        (setf savefixn-inf (inf n))
        (setf n 0)
        (setq fx (gtfixn))
        (setf (fixval fx) (mkitem savefixn-tag savefixn-inf))
        (return (mkfixn fx))))


(setq *plap nil)
(setq *pcmac nil *pgwd nil)
(setq *writingfaslfile nil)
(setq *writingasmfile nil)

% End of file.
