%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           $pxc/i860-cmac.sl
% Title:          Patterns and predicates for Intel i860 PSL cmacro expansion
% Author:         Winfried Neun
% Created:        May 15, 1992
% Modified:
% Status:         Experimental
%
% Copyright (c) 1992, Konrad-Zuse-Zentrum Berin, All Rights Reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(on fast-integers)

(remprop '*WFIX 'opencode)
(remprop '*FQUOTIENT  'opencode)
(remprop '*fplus2  'opencode)
(remprop '*Ftimes2  'opencode)
(remprop '*wfloat  'opencode)
(remprop '*fdifference  'opencode)
(remprop '*flessp  'opencode)
(remprop '*fgreaterp  'opencode)
(remprop 'cons 'opencode)
(remprop 'ncons 'opencode)
(remprop 'xcons 'opencode)

(compiletime
  (progn (on eolinstringok)
     (dm * (u)  nil)
     (setq *syslisp t)
     (load if-system)
     nil))

(fluid '(AddressingUnitsPerItem
     CharactersPerWord
     StackDirection
     *ImmediateQuote
     AddressingUnitsPerFunctionCell))

(setq *ImmediateQuote t)

(flag '(Immediate UnImmediate Indirect displacement predecrement
    label extrareg postincrement Indexed absolute deferred reglist)
    'TerminalOperand)

% MkItem may be used when evaluating WConst expressions.

(de MkItem (TagPart InfPart) % ZIB
  (lor (lsh TagPart 27) (land InfPart 16#7ffffff)))

(de freep (asize) (if (and (wleq asize (wtimes2 4 4)) % AddressingUnitsPerItem))
			   (not &links-present&))
                      (setq *usefastframe t)
                      (setq *usefastframe nil)))

(de FiveP (x)  (eq x 5))

(de TwentySevenP (x)  (eq x 27))

(de compiler_evenp (x) (and (fixp x) (eq 0 (land (quotient x 4) 1))))

(de indexedp (x) (eqcar x 'indexed))

(de DeferrableP (Expression)
  (and (PairP Expression)
       (not (memq (first Expression)
              '(Deferred $Fluid $Global Fluid Global)))))

(de RegisterDeferredP (Expression)
  (and (EqCar Expression 'Deferred) (EqCar (second Expression) 'Reg)))


(de FakeRegisterNumberP (Expression)
  (and (IntP Expression) (GreaterP Expression 5)))

% NegativeImmediateP is for numbers whose negative is a one byte operand.

(de negp (value) (and (numberp value) (lessp value 0)))

(de displacementp (x) (and (pairp x) (eq (car x) 'displacement)))

(de NegativeImmediateP (Expression)
  (and (FixP Expression) (minusp Expression) (GreaterP Expression -64)))

% expansions of addressing modes.

(DefAnyreg CAR  ZIBanyregCAR)  % ZIB

(de ZIBAnyregCAR (reg source)
   (prog (areg)
      (setq source (resolveoperand reg source))
      (if (and (displacementp source)     % recursive case, e.g. caar
               (member (cadr source) *nextreglist*))
             (setq areg (cadr source))
             (setq areg (nexttempreg)))
      (if (regp source)  % its a reg or a frame reg
          (push `(andnoth 16#f800 ,source ,areg) Resultingcode*)

          (push `(*move ,source , areg) Resultingcode*)
          (push `(andnoth 16#f800 ,areg ,areg) Resultingcode*))
      (return `(displacement ,areg 0))))

(DefAnyreg CDR  ZIBanyregCDR)  % ZIB

(de ZIBAnyregCDR (reg source)
   (prog (areg)
      (setq source (resolveoperand reg source))
      (if (and (displacementp source)     % recursive case, e.g. cddar
               (member (cadr source) *nextreglist*))
             (setq areg (cadr source))
             (setq areg (nexttempreg)))
      (if (regp source)  % its a reg or a frame reg
          (push `(andnoth 16#f800 ,source ,areg) Resultingcode*)

          (push `(*move ,source , areg) Resultingcode*)
          (push `(andnoth 16#f800 ,areg ,areg) Resultingcode*))
      (return `(displacement ,areg 4))))

(fluid '(*nextreg* *nextreglist*))

(setq *nextreg* 0        % Four registers should be sufficient
      *nextreglist* '((reg t2) (reg t3) (reg t4) ))

(de nexttempreg (x)   % ZIB  :This saves Regs , non static assignment
     (setq *nextreg* (wplus2 *nextreg* 1))
     (when (eq *nextreg* 4) (setq *nextreg* 1))
     (nth *nextreglist* *nextreg*))

(DefAnyreg Memory  ZIBanyregMemory)  % ZIB

(de ZIBAnyregMemory (reg source offset)
  (prog (aReg cadrsource)
   (setq source (resolveoperand reg source))
   (setq offset (resolveoperand reg offset))
   (cond ((and (registerp source) (registerp offset))
             (return `(indexed ,source ,offset))) % i860 has double index
         ((registerp source)
      (return (if (eq 0 offset) `(Deferred ,source)
                     (if (i860-short-immediate-p offset)
			    `(displacement ,source ,offset)
			    (progn
			      (setq aReg (nexttempreg))
				              % take a new TempReg
			      (push `(*move ,offset ,aReg) Resultingcode*)
			     `(indexed ,source ,aReg) )))))
    % what shall that ?
      ((and (AddressConstantP source)(ZeroP offset))
	    (return  `(unimmediate ,SOURCE )))
      ((and (RegisterDeferredP source )(ZeroP offset))
         (return `(deferred (Displacement (UnDeferred ,SOURCE) 0)))))
    (setq aReg (if (and (displacementp source)
                    (member (setq cadrsource (cadr source))
				  *nextreglist*))
               cadrsource      (nexttempreg))) % take a new TempReg
    (push `(*move ,source ,aReg) Resultingcode*)
    (return `(displacement ,aReg  ,offset))))


(DefAnyreg QUOTE
       AnyregQUOTE
       ((Null) (REG NIL))
       ((EqTP) (reg t))
       ((InumP) SOURCE)
       (       (QUOTE SOURCE)))

(DefAnyreg FRAME ZIBAnyregFrame)

(de ZIBAnyregFRAME (dum number)
     (if *usefastframe (nth '((reg ff1) (reg ff2) (reg ff3) (reg ff4)) number)
        `(displacement (reg st) ,(times2 4 number))))

(DefAnyreg xFRAME ZIBAnyregxFrame)

(de ZIBAnyregxFRAME (dum number)   % regardless of *usefastframe
    `(displacement (reg st) ,(times2 4 number)))

(DefAnyreg REG ZIBAnyregREG)

(de ZIBanyregReg (reg source)
   (cond ((Fakeregisternumberp source) % argumentblock lies just below SYMVAL
             `(displacement (reg symvalcenter) ,(wtimes2 (wplus2 source -17) 4)))
         (t   `(REG ,SOURCE))))

(loadtime (remflag '(fluid global $fluid $global) 'Terminaloperand))

(defanyreg Fluid ZIBanyregfluid)

(defanyreg $Fluid ZIBanyregfluid)

(fluid '(*syslisp ))

(de ZIBanyregFluid (reg source)
   (cond ((and (idp source) (get source 'registername))
			 `(reg ,(get source 'registername)))
         (t       `($Fluid ,SOURCE))))

(defanyreg Global ZIBanyregglobal)

(defanyreg $Global ZIBanyregglobal)

(de ZIBanyregGlobal (reg source)
   (cond ((and (idp source) (get source 'registername))
			 `(reg ,(get source 'registername)))
         (t       `($global ,SOURCE))))

%-----------------------------------------------------------------------------

(de *LinkE (DeAllocCount FunctionName FunctionType NumberOfArguments)
  (if (FlagP FunctionName 'ForeignFunction)
      (list (list '*ForeignLink FunctionName FunctionType NumberOfArguments)
        '(*Exit 0))
      (list (list '*DeAlloc DeAllocCount)
        (list '*JCall FunctionName))))

(DefCMacro *Call
       ((InternallyCallableP)  (call (InternalEntry ArgOne))
			   (NOP))
	((asmp)   (ld!.l (displacement (reg SYMFNCcenter)
					 (times2 4 (idloc ArgOne)))
			(reg t4))
		  (calli (reg t4))
                  (*move (idloc ArgOne) (reg t3)))

        (         (ld!.l (entryaddress ArgOne) (reg t4))
                  (calli (reg t4))
                  (or (idloc ArgOne) (reg 0) (reg t3))))

(DefCMacro *JCall
       ((InternallyCallableP) (Br (InternalEntry ArgOne))
			      (nop))
        ((asmp)   (ld!.l (displacement (reg SYMFNCcenter)
			 (times2 4 (idloc ArgOne))) (reg t4))
                  (bri (reg t4))
                  (*move (idloc ArgOne) (reg t3) ))

        (         (ld!.l (entryaddress ArgOne) (reg t4))
                  (bri (reg t4))
                  (or (idloc ArgOne) (reg 0) (reg t3))))

(de displacement-p (x)
    (and (eqcar x 'displacement)
     (eqn (length x) 3) (i860-register-p (cadr x))))

(de deferred-p (x)
    (and (pairp x) (or (i860-register-p (car x))
                   (eq (car x) 'deferred))))

(remprop '*move 'optfn)

(DefCMacro *Move
       (Equal)
       ((regp regp)		(or  ArgOne (reg null) ArgTwo))
       ((indexedp regp)		(ld!.l ArgOne ArgTwo))
       ((regp anyp)		(*movex ArgOne ArgTwo))
       ((anyp regp)		(*movex ArgOne ArgTwo))
       (			(*move ArgOne (reg t5))
				(*move (reg t5) ArgTwo)))

% At Alloc and Dealloc we have to add room for the returnaddr register, since
% this must be saved by the user

(de *ALLOC (framesize)

    (setq NAlloc!* framesize)
    (cond ((freep (wtimes2 AddressingUnitsPerItem framesize)) NIL)
          (t  (when (eq 0 (wand framesize 1)) % set st on double word boundary
			 (setq framesize (iadd1 framesize)))
	      `((adds ,(minus (plus2 4 (times2 4 framesize))) (reg st) (reg st))
		(st!.l (reg returnaddr) (displacement (reg st) 0))))
)   )

(defcmacro *ALLOC)

(de *xalloc (framesize)      % alloc in any case, regardless of freep
    (when (eq 0 (wand framesize 1)) % set st on double word boundary
			 (setq framesize (iadd1 framesize)))
    (setq framesize (wtimes2 4 framesize))
    `((adds ,(minus (plus2 4 framesize)) (reg st) (reg st))
      (st!.l (reg returnaddr) (displacement (reg st) 0))))

(defcmacro *xALLOC)

(DefCMacro *DeAlloc
           ((FreeP))
           ((compiler_evenp) (ld!.l (displacement (reg st) 0) (reg returnaddr))
                    (*wplus2 (reg st) (plus2 8 ArgOne)))
           (    (ld!.l (displacement (reg st) 0) (reg returnaddr))
                (*wplus2 (reg st) (plus2 4 ArgOne))))

(de *xDeAlloc (framesize)
            (when (eq 0 (wand framesize 1)) % set st on double word boundary
			 (setq framesize (iadd1 framesize)))
           `((ld!.l (displacement (reg st) 0) (reg returnaddr))
             (*wplus2 (reg st) (plus2 4 ,(wtimes2 4 framesize)))))

(DefCMacro *xDeAlloc)

(de asmp () *writingasmfile)

(DefCMacro *Exit
         % leaf routine first
           ((Freep)  (bri (reg returnaddr))
                     (nop))
           (((lambda (x) (and (asmp) (compiler_evenp x))))
                (ld!.l (displacement (reg st) 0) (reg returnaddr))
		(bri (reg returnaddr))
                (addu  (plus2 8 ArgOne) (reg st) (reg st)))
           ((asmp)
		(ld!.l (displacement (reg st) 0) (reg returnaddr))
		(bri (reg returnaddr))
                (addu (plus2 4 ArgOne) (reg st) (reg st)))
           ((compiler_evenp)
		(ld!.l (displacement (reg st) 0) (reg returnaddr))
                (bri (reg returnaddr))
                (*wplus2 (reg st) (plus2 8 ArgOne))) 
           (    (ld!.l (displacement (reg st) 0) (reg returnaddr))
                (bri (reg returnaddr))
                (*wplus2 (reg st) (plus2 4 ArgOne))) 
)

(de deferredp (x) (and (pairp x) (eq (car x) 'deferred)))

(DefCMacro *Jump
       ((deferredp)        (*move (undeferred ArgOne) (reg t5))
			   (jmpl (deferred (reg t5)) (reg 0))
				(or (reg 0) 0 (reg 0))) % NO NOP!!!!! FOR LAPOPT
       ((displacementp)    (jmpl ArgOne (reg 0))
			   (nop))
       (                   (br ArgOne)
			   (nop)))


(DefCMacro *Lbl           (ArgOne))

(de regp (x) (or (eqcar x 'Reg)
              (and (eqcar x 'frame) (leq (cadr x) 12))))

(de *3op (a1 a2 instruction)
  (prog (ResultingCode*)
    (return (CMacroPatternExpand
           (list (ResolveOperand '(Reg t1) A1)
                 (ResolveOperand '(Reg t2) A2)
                  Instruction)
                 (get '*3op 'CMacroPatternTable)))))

(DefCMacro *3op
       ((regp regp) (ArgThree ArgOne ArgTwo ArgOne))
       ((regp AnyP)  (*move ArgTwo (reg T5))
                      (*3op ArgOne (reg t5) ArgThree))
       ((Anyp regp)  (*move ArgOne (reg T5))
                      (ArgThree (reg t5) ArgTwo (reg t5))
                      (*move (reg t5) ArgOne))
       (              (*move ArgTwo (reg T5))
                      (*move ArgOne (reg T6))
                      (ArgThree (reg t6) (reg t5) (reg t5))
                      (*move (reg t5) ArgOne)))

(DefCMacro *WPlus2
       ((*wplus3 ArgOne ArgOne ArgTwo)))

(de *wplus3(arg1 arg2 arg3)
    (expand3operandcmacro arg1 arg2 arg3 '*wplus3))

(DefCMacro *WPlus3
           ((regp regp regp) (adds ArgThree ArgTwo ArgOne))
           ((regp i860-short-immediate-p regp)
				 (adds ArgTwo ArgThree ArgOne))
           ((regp regp i860-short-immediate-p)
				 (adds ArgThree ArgTwo ArgOne))
           ((regp i860-short-immediate-p anyp)
				(*move ArgThree (reg t5))
				(adds  ArgTwo (reg t5) ArgOne))
           ((regp anyp i860-short-immediate-p)
				(*move ArgTwo (reg t5))
				(adds  ArgThree  (reg t5) ArgOne))
	   ((regp regp anyp)  (*move ArgThree (reg t5))
				(*wplus3  ArgOne ArgTwo (reg t5)))
           ((regp anyp anyp)  (*move ArgTwo ArgOne)
				(*wplus2 ArgOne ArgThree))
           ((anyp regp anyp)  (*move ArgThree (reg t5))
				(*wplus2 (reg t5) ArgTwo)
				(*move (reg t5) ArgOne))
           (                   (*move ArgTwo (reg t5))
				(*move ArgThree (reg t6))
				(adds (reg t5) (reg t6) (reg t5))
				(*move (reg t5) ArgOne)))

(DefCMacro *WDifference
       ((*3op ArgOne ArgTwo Subs)))

(de *wdifference3 (arg1 arg2 arg3)
    (expand3operandcmacro arg1 arg2 arg3 '*wdifference3))

% format ArgOne = ArgTwo - ArgThree

(DefCMacro *Wdifference3
           ((regp regp regp) (subs ArgTwo ArgThree ArgOne))
           ((regp i860-short-immediate-p regp)
				(subs ArgTwo ArgThree ArgOne))
           ((regp anyp regp) (*move ArgTwo (reg t5))
				(subs (reg t5) ArgThree ArgOne))
           ((regp anyp anyp)  (*move ArgTwo (reg t5))
				(*move ArgThree (reg t6))
				(subs (reg t5) (reg t6) ArgOne))
           ((anyp regp anyp)  (*move ArgThree (reg t5))
				(subs  ArgTwo (reg t5) (reg t5))
				(*move (reg t5) ArgOne))
           (                   (*move ArgTwo (reg t5))
				(*move ArgThree (reg t6))
				(subs (reg t5) (reg t6) (reg t5))
				(*move (reg t5) ArgOne)))

(loadtime (remprop 'wtimes2 'openfn))
(loadtime (remprop 'wquotient  'openfn))
(loadtime (remflag '(*wquotient *wremainder *wtimes2) 'mc))
(loadtime (remprop 'wremainder 'openfn))

(DefCMacro *WAnd
       ((*3op ArgOne ArgTwo and)))

(DefCMacro *WOr
       ((*3op ArgOne ArgTwo or)))

(DefCMacro *WXOr
       ((*3op ArgOne ArgTwo xor)))

(de fixplusp (x) (and (fixp x) (not (minusp x))))

(DefCMacro *AShift
    ((AnyP ZeroP))
    ((regp negp)     (shra ArgOne (minus ArgTwo) ArgOne))
    ((regp fixplusp) (shl ArgOne ArgTwo ArgOne))
    ((regp regp)    (subs argTwo (reg 0) (reg 0))
                      (bc!.t templabel)
                      (subs (reg 0) ArgTwo (reg t6)) % delay slot
		      (br templabel2)
                      (shl ArgTwo ArgOne  ArgOne)
                     templabel
                      (shra (reg t6) ArgOne ArgOne)
                     templabel2)
    ((regp anyp)     (*move ArgTwo (reg t5))
                      (*ashift ArgOne (reg t5)))
    ((Anyp fixp)      (*move ArgOne (reg t5))
                      (*ashift (reg t5) ArgTwo)
		      (*move (reg t5) ArgOne))
    (                 (*move ArgOne (reg t1))
                      (*ashift (reg t1) ArgTwo)
                      (*move (reg t1) ArgOne)))

(DefCMacro *WShift
    ((AnyP ZeroP))
    ((regp negp)     (shr (minus ArgTwo) ArgOne ArgOne))
    ((regp fixplusp) (shl ArgTwo ArgOne ArgOne))
    ((regp regp)    (subs argTwo (reg 0) (reg 0))
		      (bc!.t templabel)
		      (subs (reg 0) ArgTwo (reg t6)) % delay slot
		      (br templabel2)
                      (shl ArgTwo ArgOne  ArgOne)% delay slot
                     templabel
		      (shr (reg t6) ArgOne ArgOne)
		     templabel2)
    ((regp anyp)     (*move ArgTwo (reg t5))
                      (*wshift ArgOne (reg t5)))
    ((Anyp fixp)      (*move ArgOne (reg t5))
                      (*wshift (reg t5) ArgTwo)
		      (*move (reg t5) ArgOne))
    (                 (*move ArgOne (reg t1))
                      (*wshift (reg t1) ArgTwo)
                      (*move (reg t1) ArgOne)))

(DefCMacro *WNot
       ((regp regp)  (*move -1 (reg t6))
		      (andnot ArgTwo (reg t6) ArgOne))
       ((regp Anyp)  (*move ArgTwo ArgOne)
		      (*move -1 (reg t6))
                      (andnot ArgOne (reg t6) ArgOne))
       (              (*move ArgTwo (reg t5))
                      (*wnot (reg t5) (reg t5))
                      (*move (reg t5) ArgOne)))

(DefCMacro *WMinus
           ((regp regp) (subs (reg 0) ArgTwo ArgOne))
           ((regp Anyp)  (*move ArgTwo ArgOne)
			  (subs (reg 0) ArgOne ArgOne))
           (              (*move ArgTwo (reg t5))
			  (*wminus (reg t5) (reg t5))
			  (*move (reg t5) ArgOne)))

(DefCMacro *MkItem
       ((regp fixp)     (andnoth 16#f800 ArgOne ArgOne)
			 (orh (times 16#0800 ArgTwo) ArgOne ArgOne))
       ((regp regp)     (andnoth 16#f800 ArgOne ArgOne)
			 (shl 27 ArgTwo (reg t5))
			 (or ArgOne (reg t5) ArgOne))
       ((regp anyp)     (*move ArgTwo (reg t5))
			 (*Mkitem ArgOne (reg t5)))
       (                (*move ArgOne (reg t1))
			 (*mkitem (reg t1) ArgTwo)
			 (*move (reg t1) ArgOne)))

(de smallinump (x) (i860-short-immediate-p x))

(de immediate-p (x)    (fixp x))

(DefCmacro *tag
      ((regp regp) (shr 27 ArgTwo ArgOne))
      ((Anyp  regp)(*move ArgTwo (reg t1))
                    (shr 27 (reg t1) (reg t1))
                    (*move (reg t1) ArgOne))
      ((regp anyp) (*move ArgTwo (reg t1))
                    (shr 27 (reg t1) ArgOne))
      (             (*move ArgTwo (reg t1))
                    (*tag ArgOne (reg t1))))

(de *tag (ar1 ar2)
     (expand2operandCmacro ar1 ar2 '*tag))

(DefCMacro *JumpType
      (             (*Tag (Reg t5) ArgOne)
                    (*JumpEQ ArgThree (Reg t5) ArgTwo)))

(DefCMacro *JumpNotType
      (             (*Tag (Reg t5) ArgOne)
                    (*JumpNOTEQ ArgThree (Reg t5) ArgTwo)))

(DefCMacro *JumpInType
       ((regp Anyp) (shr 27 ArgOne (reg t5))
                     (subs ArgTwo (reg t5) (reg 0))
                     (bnc!.t ArgThree)
		     (nop)
                     (xor 31 (reg t5) (reg 0))
		     (bc!.t ArgThree)
                     (nop))
       (             (*move ArgOne (reg t5))
		     (shr 27 (reg t5) (reg t5))
                     (subs ArgTwo (reg t5) (reg 0))
                     (bnc!.t ArgThree)
                     (nop)
		     (xor 31 (reg t5) (reg 0))
                     (bc!.t ArgThree)
                     (nop)))

(DefCMacro *JumpNotInType
       ((regp Anyp) (shr 27 ArgOne (reg t5))
                     (subs ArgTwo (reg t5) (reg 0))
                     (bnc!.t Templabel)
                     (nop)
		     (xor 31 (reg t5) (reg 0))
		     (bnc!.t ArgThree)
		     (nop)
		     templabel)
       (            (*move ArgOne (reg t5))
		     (shr 27 (reg t5) (reg t5))
                     (subs ArgTwo (reg t5) (reg 0))
                     (bnc!.t Templabel)
                     (nop)
                     (xor 31 (reg t5) (reg 0))
                     (bnc!.t ArgThree)
		     (nop)
                     templabel))

(de posSmallInumP (x) (and (SmallInumP x) (greaterp x -1)))

% hier: werden nicht mehr gebraucht nach bootstrap!!
(de *JumpEQ (Label Arg1 Arg2)
  (Expand2OperandAndLabelCMacro Arg1 Arg2 Label '*JumpEQ))

(de *JumpNotEQ (Label Arg1 Arg2)
  (Expand2OperandAndLabelCMacro Arg1 Arg2 Label '*JumpNotEQ))

(de *JumpWLessP (Label Arg1 Arg2)
  (Expand2OperandAndLabelCMacro Arg1 Arg2 Label '*JumpWLessP))

(de *JumpWGreaterP (Label Arg1 Arg2)
  (Expand2OperandAndLabelCMacro Arg1 Arg2 Label '*JumpWGreaterP))

(de *JumpWLEQ (Label Arg1 Arg2)
  (Expand2OperandAndLabelCMacro Arg1 Arg2 Label '*JumpWLEQ))

(de *JumpWGEQ (Label Arg1 Arg2)
  (Expand2OperandAndLabelCMacro Arg1 Arg2 Label '*JumpWGEQ))

(defcmacro *jumpeq
     ((regp regp)       (xor ArgOne ArgTwo (reg 0))
			 (bc!.t ArgThree)
			 (nop))
     ((regp posSmallInumP) (xor ArgTwo ArgOne (reg 0))
			 (bc!.t ArgThree)
			 (nop))
     ((regp anyp)       (*move ArgTwo (reg t5))
			 (*jumpeq ArgThree ArgOne (reg t5)))
     ((anyp  regp)      (*move ArgOne (reg t5))
			 (*jumpeq ArgThree (reg t5) ArgTwo))
     ((anyp  posSmallInumP) (*move ArgOne (reg t5))
			 (*jumpeq ArgThree (reg t5) ArgTwo))
     (                   (*move ArgOne (reg t5))
			 (*move ArgTwo (reg t4))
			 (*jumpeq ArgThree (reg t5) (reg t4))))

(defcmacro *jumpnoteq
     ((regp regp)       (xor ArgOne ArgTwo (reg 0))
			 (bnc!.t ArgThree)
			 (nop))
     ((regp posSmallInumP) (xor ArgTwo ArgOne (reg 0))
			 (bnc!.t ArgThree)
			 (nop))
     ((regp anyp)       (*move ArgTwo (reg t5))
			 (*jumpnoteq ArgThree ArgOne (reg t5)))
     ((anyp  regp)      (*move ArgOne (reg t5))
			 (*jumpnoteq ArgThree (reg t5) ArgTwo))
     ((anyp posSmallInumP) (*move ArgOne (reg t5))
			 (*jumpnoteq ArgThree (reg t5) ArgTwo))
     (                   (*move ArgOne (reg t5))
			 (*move ArgTwo (reg t4))
			 (*jumpnoteq ArgThree (reg t5) (reg t4))))

(defcmacro *jumpwlessp
     ((regp regp)       (subs ArgOne ArgTwo (reg 0))
			 (bc!.t ArgThree)
			 (nop))
     ((SmallInumP Regp) (subs ArgOne ArgTwo (reg 0))
			 (bc!.t ArgThree)
			 (nop))
     ((regp anyp)       (*move ArgTwo (reg t5))
			 (*jumpwlessp ArgThree ArgOne (reg t5)))
     ((anyp  regp)      (*move ArgOne (reg t5))
			 (*jumpwlessp ArgThree (reg t5) ArgTwo))
     ((SmallInumP Anyp) (*move ArgTwo (reg t5))
			 (*jumpwlessp ArgThree ArgOne (reg t5)))
     (                   (*move ArgOne (reg t5))
			 (*move ArgTwo (reg t4))
			 (*jumpwlessp ArgThree (reg t5) (reg t4))))

(defcmacro *jumpwgreaterp
     ((regp regp)       (subs ArgTwo ArgOne (reg 0))
			 (bc!.t ArgThree)
			 (nop))
     ((regp SmallInumP) (subs ArgTwo ArgOne (reg 0))
			 (bc!.t ArgThree)
			 (nop))
     ((regp anyp)       (*move ArgTwo (reg t5))
			 (*jumpwgreaterp ArgThree ArgOne  (reg t5)))
     ((anyp  regp)      (*move ArgOne (reg t5))
			 (*jumpwgreaterp ArgThree (reg t5) ArgTwo))
     ((anyp  SmallInumP) (*move ArgOne (reg t5))
			 (*jumpwgreaterp ArgThree (reg t5) ArgTwo))
     (                   (*move ArgOne (reg t5))
			 (*move ArgTwo (reg t4))
			 (*jumpwgreaterp ArgThree (reg t5) (reg t4))))

(defcmacro *jumpwleq
     ((regp regp)       (subs ArgTwo ArgOne (reg 0))
			 (bnc!.t ArgThree)
			 (nop))
     ((regp SmallInumP) (subs ArgTwo ArgOne (reg 0))
			 (bnc!.t ArgThree)
			 (nop))
     ((regp anyp)       (*move ArgTwo (reg t5))
			 (*jumpwleq ArgThree ArgOne (reg t5)))
     ((anyp  regp)      (*move ArgOne (reg t5))
			 (*jumpwleq ArgThree (reg t5) ArgTwo))
     ((anyp  SmallInumP) (*move ArgOne (reg t5))
			 (*jumpwleq ArgThree (reg t5) ArgTwo))
     (                   (*move ArgOne (reg t5))
			 (*move ArgTwo (reg t4))
			 (*jumpwleq ArgThree (reg t5) (reg t4))))

(defcmacro *jumpwgeq
     ((regp regp)       (subs ArgOne ArgTwo (reg 0))
			 (bnc!.t ArgThree)
			 (nop))
     ((SmallInumP Regp) (subs ArgOne ArgTwo (reg 0))
			 (bnc!.t ArgThree)
			 (nop))
     ((regp anyp)       (*move ArgTwo (reg t5))
			 (*jumpwgeq ArgThree ArgOne  (reg t5)))
     ((anyp  regp)      (*move ArgOne (reg t5))
			 (*jumpwgeq ArgThree (reg t5) ArgTwo))
     ((SmallInumP Anyp) (*move ArgTwo (reg t5))
			 (*jumpwgeq ArgThree ArgOne (reg t5)))
     (                   (*move ArgOne (reg t5))
			 (*move ArgTwo (reg t4))
			 (*jumpwgeq ArgThree (reg t5) (reg t4))))

(de fluid-arg-p (x)
    (and (pairp x)
     (member (car x) '(!$fluid !$global fluid global))))

(DefCMacro *Push
       ( (*cerror "NO push ")))

(DefCMacro *Pop
       ( (*cerror "NO pop ")))

(de *loc (ArgOne ArgTwo)
   (let ((resolvedarg (resolveoperand '(reg t2) ArgTwo)))
      (cond ((deferred-p resolvedarg)
         `((*move (undeferred ,resolvedarg) ,ArgOne))) % get the address.
         ((displacement-p resolvedarg)
            `((adds ,(caddr resolvedarg) ,(cadr resolvedarg) ,ArgOne)))
         ((eqcar resolvedarg 'cdr)
            `((adds 4 ,(cadr resolvedarg) ,ArgOne)))
         (t `((*movex ,resolvedarg ,ArgOne))))))

(DefCMacro *loc)

%t6 is used, because it is not used in lap code, but it destroyed in cmacros.
%look at test file mask.s in ~burdorf/i860 for example.

(de X5p (u) (eq u 5))

(de X27p (u) (eq u 27))

(DefCMacro *Field

  ((regp regp zerop fixp)(shr (difference 32 ArgFour) ArgTwo ArgOne))
  ((regp regp X5p X27p)  (andnoth 16#f800 ArgTwo  ArgOne))
  ((regp regp fixp fixp) (shl ArgThree ArgTwo ArgOne)
			 (shr (difference 32 ArgFour) ArgOne ArgOne))
  ((regp anyp fixp fixp) (*move ArgTwo (reg t5))
			 (shl ArgThree (reg t5) (reg t5))
			 (shr (difference 32 ArgFour) (reg t5) ArgOne))
  ((anyp regp fixp fixp) (*field (reg t5) ArgTwo ArgThree ArgFour)
			 (*move (reg t5) ArgOne))
  (			 (*move ArgTwo (reg t6))
			 (*field (reg t5) (reg t6) ArgThree ArgFour)
			 (*move (reg t5) ArgOne)))

(DefCMacro *SignedField

  ((regp regp fixp fixp) (shl ArgThree ArgTwo ArgOne)
			 (shra (difference 32 ArgFour) ArgOne ArgOne))
  ((regp anyp fixp fixp) (*move ArgTwo (reg t5))
			 (shl ArgThree (reg t5) (reg t5))
			 (shra (difference 32 ArgFour) (reg t5) ArgOne))
  ((anyp regp fixp fixp) (*Signedfield (reg t5) ArgTwo ArgThree ArgFour)
			 (*move (reg t5) ArgOne))
  (			 (*move ArgTwo (reg t6))
			 (*Signedfield (reg t5) (reg t6) ArgThree ArgFour)
			 (*move (reg t5) ArgOne)))

% following is used by garbage collector only. Not very useful to
% optimize other cases than put inf field or put tag field (with constant)


(DefCMacro *PutField

 ((regp regp fiveP TwentysevenP)(andh 16#f800 ArgOne (reg t6))
				(andnoth 16#f800 ArgTwo ArgTwo)
				(or    ArgTwo (reg t6) ArgTwo))
 ((regp regp Zerop fivep)       (shl   27 ArgOne (reg t5))
				(andnoth 16#f800 ArgTwo ArgTwo)
				(or    (reg t5) ArgTwo ArgTwo))
 ((inump regp Zerop fivep)      (shl (reg 0) ArgOne (reg t5))
				(shl 27 (reg t5) (reg t5))
				(andnoth 16#f800 ArgTwo ArgTwo)
				(or  (reg t5) ArgTwo ArgTwo))
 (      (*move -1 (reg t5))
        (*move  ArgOne (reg t4))
        (shl  (difference 32 ArgFour) (reg t5) (reg t5))
        (*wshift (reg t5) (minus ArgThree)) % for 0 opt
        (shl  (difference 32 (plus ArgThree ArgFour)) (reg t4) (reg t4))
        (and  (reg t5) (reg t4) (reg t4))
        (*move ArgTwo (reg t6))
        (andnot  (reg t5) (reg t6) (reg t6))
        (or   (reg t4) (reg t6) (reg t5))
        (*move (reg t5) ArgTwo)))


(deflist '((Byte    ((*wplus2 (reg 1) (reg 2))
		     (ld!.b (displacement (reg 1) 0) (reg 1))))
       (PutByte     ((*wplus2 (reg 1) (reg 2))
		     (st!.b (reg 3) (displacement (reg 1) 0))))
       (HalfWord    ((shl 1 (reg 2) (reg t1))
		     (*wplus2 (reg 1) (reg t1))
                     (ld!.s (displacement (reg 1) 0) (reg 1))))
       (PutHalfWord ((shl 1 (reg 2) (reg t1))
		     (*wplus2 (reg t1) (reg 1))
                     (st!.s (reg 3) (displacement (reg t1) 0)))))
  'OpenCode)

(&OneReg '(Byte PutByte HalfWord PutHalfWord))

(on fast-integers)

(for (in function '(idapply0 idapply1 idapply2 idapply3 idapply4))
     (from register 1)
     (do
      (put function 'opencode
        `((*move ($fluid SYMFNC) (reg t1))
          (shl 5 (reg ,register) (reg ,register))
          (shr 3 (reg ,register) (reg t3))
	  (*wplus2 (reg t1) (reg t3))
	  (ld!.l (displacement (reg t1) 0) (reg t1))
          (calli (reg t1))
          (or (reg 0) (reg 0) (reg 0)))) % NO NOP!!!!! FOR LAPOPT
      (put function 'exitopencode
        `((*move ($fluid SYMFNC) (reg t1))
          (shl 5 (reg ,register) (reg ,register))
          (shr 3 (reg ,register) (reg t2))
          (*wplus2 (reg t2) (reg t1))
          (ld!.l (displacement (reg t2) 0) (reg t1))
          (bri (reg t1))
          (or (reg 0) (reg 0) (reg 0)))) % NO NOP!!!!! FOR LAPOPT
))

(for (in function '(addressapply0 addressapply1 addressapply2
                addressapply3 addressapply4))
     (from register 1)
     (do
      (put function 'opencode
         `((calli (reg ,register))
           (or (reg 0) (reg 0) (reg 0)))) % NO NOP!!!!! FOR LAPOPT
      (put function 'exitopencode
         `((bri (reg ,register))
           (or (reg 0) (reg 0) (reg 0)))) % NO NOP!!!!! FOR LAPOPT
      ))


(defcmacro *fast-apply-load
   ( (*move ArgOne (reg t1)) ))

(put 'fast-idapply    'opencode
     '((*move (fluid symfnc) (reg t2))
       (andnoth 16#f800 (reg t1) (reg t3))
       (shl 2 (reg t3) (reg t1)) % by 4 to get offset into symfnc table
       (*wplus2 (reg t2) (reg t1))
       (ld!.l (displacement (reg t2) 0) (reg t2))
       (calli (reg t2))
       (nop)))

(put 'fast-codeapply    'opencode
     '((andnoth 16#f800 (reg t1) (reg t2))
       (calli (reg t2))
       (nop)))

(put 'fast-idapply    'exitopencode
     '((*move (fluid symfnc) (reg t2))
       (andnoth 16#f800 (reg t1) (reg t3))
       (shl 2 (reg t3) (reg t1)) % by 4 to get offset into symfnc table
       (*wplus2 (reg t2) (reg t1))
       (ld!.l (displacement (reg t2) 0) (reg t2))
       (bri (reg t2))
       (nop)))

(put 'fast-codeapply    'exitopencode
     '((andnoth 16#f800 (reg t1) (reg t2))
       (bri (reg t2))
       (nop)))

(De *LamBind (Regs Fluids)

 % be careful, code is somewhat tricky

 (prog (n lng list cregs cfluids cadrcfluids lab initload freeregs)
      (setq lab (gensym))
      (setq n 0)
      (setq Regs (rest Regs))   % Remove "Registers" from the front
      (setq Fluids (rest fluids)) % Remove NONLOCALVARS
      (setq lng (wtimes2 (length Regs) 8)) % two words per BndStk entry *4

   %looking for free register to preload values of fluid s

      (setq freeregs (mapcan '((reg 1)(reg 2)(reg 3)(reg 4)(reg 5))
                   (function (lambda (x)
				 (cond ((member x regs) nil)
				        (t (cons x nil))
      )              ) )         )  )
      (setq freeregs (nconc (list '(reg T3) '(reg T4)) freeregs))
      (setq cfluids fluids) % copy of fluids

preload  (setq initload
       (progn (setq cadrcfluids
                (nconc cadrcfluids (cons (car freeregs) nil)))
        (nconc initload `((*move ,(car cfluids) ,(car freeregs))))
       )   )
       (setq freeregs (cdr freeregs))
       (setq cfluids (cdr cfluids))

       (cond ((and freeregs cfluids) (go preload)))

       (setq freeregs (nconc cadrcfluids cfluids)) %end of preloadcode

       % freeregs contains the list of preloaded regs
       % and not preloaded fluids if those exist

      (setq list `((*move ($fluid BndStkPtr) (Reg t1))
               (*wplus3 (Reg t2) ,lng (Reg t1))
               (subs  (reg BndstkUpperBound)(reg t2)(reg 0))
               (bnc!.t ,lab)
               (*move (Reg t2) ($fluid BndstkPtr))
	       (*call Bstackoverflow)
		,lab)) %start of code

      (setq list (append initload list))

  loop
      (setq cregs (car Regs))
      (setq cfluids (car Fluids))
      (setq cadrcfluids (cadr cfluids))
      (when (or (eq cadrcfluids 't) (eq cadrcfluids 'nil))
           (stderror "T and NIL cannot be rebound"))
      (setq n (wplus2 n 8))
      (Setq list (append list
     `((*move ,(car freeregs) (displacement (Reg t1) ,n))
       (*move (quote ,Cadrcfluids)
                   (displacement (reg t1) ,(wplus2 n -4)))
       (*move ,cregs ,cfluids)
      )          ))
      (setq fluids (cdr Fluids))
      (setq freeregs (cdr freeregs))
      (cond ((setq regs (cdr Regs)) (go loop)))
      (return list)
)    )

(defcmacro !*lambind)

(De *ProgBind (Fluids)

 % be careful, code is somewhat tricky

 (prog (n lng list cfluids cadrcfluids lab initload freeregs)
      (setq lab (gensym))
      (setq n 0)
      (setq Fluids (rest fluids)) % Remove NONLOCALVARS
      (setq lng (wtimes2 (length Fluids) 8)) % two words per BndStk entry
				          % * 4 addressingunits
      (setq freeregs '((reg t3)(reg T4)(reg 1)(reg 2)(reg 3)(reg 4)(reg 5)))
      (setq cfluids fluids) % copy of fluids

preload  (setq initload
       (progn (setq cadrcfluids
                (nconc cadrcfluids (cons (car freeregs) nil)))
        (nconc initload `((*move ,(car cfluids) ,(car freeregs))))
       )   )
       (setq freeregs (cdr freeregs))
       (setq cfluids (cdr cfluids))

       (cond ((and freeregs cfluids) (go preload)))

       (setq freeregs (nconc cadrcfluids cfluids)) %end of preloadcode

       % freeregs contains the list of preloaded regs
       % and not preloaded fluids if those exist

      (setq list `((*move ($fluid BndStkPtr) (Reg t1))
		(*wplus3 (Reg t2) ,lng (Reg t1))
		(subs (reg BndstkUpperBound)(reg t2)(reg 0))
		(bnc!.t ,lab)
		(*move (Reg t2) ($fluid BndstkPtr))
		(*call Bstackoverflow)
                ,lab)) %start of code

     (setq list (append initload list))

 loop
      (setq cfluids (car Fluids))
      (setq cadrcfluids (cadr cfluids))
      (when (or (eq cadrcfluids 't) (eq cadrcfluids 'nil))
           (stderror "T and NIL cannot be rebound"))
      (setq n (wplus2 n 8))
      (Setq list (append list
             `((*move ,(car freeregs) (displacement (Reg t1) ,n))
               (*move (quote ,Cadrcfluids)
			   (displacement (reg t1)  ,(wplus2 n -4)))
               (*move (quote nil) ,cfluids)
      )          ))
      (setq freeregs (cdr freeregs))
      (cond ((setq Fluids (cdr Fluids)) (go loop)))
      (return list)
)    )
(defcmacro *progbind)

(De *FreeRstr (Fluids)

 (prog (n lng list cfluids listfluids lab initload freeregs)
      (setq lab (gensym))
      (setq n 0)
      (setq Fluids (rest fluids)) % Remove NONLOCALVARS
      (setq lng (wtimes2 (length Fluids) 2)) % two words per BndStk entry
      (setq freeregs '((reg t3)(reg t4)(reg 2)(reg 3)(reg 4)(reg 5)))
      (setq cfluids fluids) % copy of fluids
      (setq n (wtimes2 4 (wdifference 2 lng)))
      (setq lng (wtimes2 lng 4)) % * addressingunitperitem
      (setq initload (list '(*move (reg Bndstkptr) (reg t1))))

preload  (setq initload
       (progn (setq listfluids
               (if freeregs
                (nconc listfluids (cons (car freeregs) nil))
                (nconc listfluids (cons nil nil))) )
        (nconc initload
         (if freeregs
     `((*move (displacement (reg t1) ,n) ,(car freeregs))) nil)
       )   ))
       (setq n (wplus2 n 8))
       (when freeregs (setq freeregs (cdr freeregs)))
       (setq cfluids (cdr cfluids))

       (cond (cfluids (go preload)))

       (setq freeregs listfluids ) %end of preloadcode

       % freeregs contains the list of preloaded regs
       % and nil if not enough regs available

      (setq list `((*move ($fluid BndStkPtr) (Reg t1))
                (*wplus3 (Reg t2) ,(minus lng) (Reg t1))
                (subs (reg t1) (reg BndstkLowerBound)(reg 0))
                (bnc!.t ,lab)
                (*move (Reg t2) ($fluid BndstkPtr))
                (*call Bstackunderflow)
                ,lab)) %start of code

     (setq list (append initload list))
     (setq n 0)

 loop
      (setq cfluids (car Fluids))
      (setq n (wplus2 n 8))

  % insert reloaded register or memory reference

      (setq list (append list
           (if (car freeregs) `((*move ,(car freeregs) ,cfluids ))
                `((*move (displacement (Reg t2) ,n) ,cfluids )))

      )          )
      (setq freeregs (cdr freeregs))
      (cond ((setq Fluids (cdr Fluids)) (go loop)))
      (return list)
)    )
(defcmacro *freerstr)

(setq *unsafebinder t)   % has to save Registers across calls

(global '(*writingasmfile))

(de !*xjumpon (register lowerbound upperbound labellist)
    (for (from x lowerbound upperbound)
         (join
          `((bte ,x ,register ,(pop labellist))
            (nop)))))

(de !*jumpon (register lowerbound upperbound labellist)
   (if *writingasmfile
       (*xjumpon register lowerbound upperbound labellist)
     (PROG (X LL LL2)
       (setq ll  (gensym))
       (setq ll2 (stringgensym))
       (SETQ X
        (if (and (weq lowerbound 0) (weq upperbound 31) *syslisp)
                  % jumpon on tags (most probably)
      `((shl 3 ,register (reg t5))  % 8 bytes per jumptable entry
        (move-relocatable-object ,ll2 (reg t1))
	(*wplus2 (reg t1) (reg t5))
        (bri (reg t1))
        (nop)
       ,ll2)
      `((xor  ,upperbound ,register (reg 0))
        (bc!.t ,(lastcar labellist))
        (nop)
	(subs ,upperbound ,register (reg 0))
        (bc!.t (label ,ll))
	(nop)
	(xor ,Lowerbound ,register  (reg 0))
	(bc!.t (label ,ll2))
	(nop)
        (subs ,Lowerbound ,register  (reg t5))
        (bnc!.t (label ,ll))
        (nop)
        (shl 3 (reg t5) (reg t5))  % 8 bytes per jumptable entry
	(move-relocatable-object ,ll2 (reg t1))
	(*wdifference (reg t1) (reg t5))
        (bri (reg t1))
        (nop)
       ,ll2) ) )
      Loop  (Setq x (nconc X `((Br ,(car Labellist)) ,(list 'nop))) )
            (setq Labellist (cdr Labellist))
            (cond (Labellist (go loop)))

            (setq x (nconc x `((*lbl (Label ,ll))) ))
            (return x)
)) )  )

(defcmacro !*jumpon)


% Call foreign fns (loaded by oload) by putting SymFunc cell addr in a reg,
% then doing "calls" instr indirect via the target address of the jmp
% instruction stored there.  Use old direct link for asm (kernel) compile.

(de *foreignlink (functionname functiontype numberofarguments)

         `((*move (quote T) ($fluid *kernelmode))
           (*move (reg st) (reg t1))
	   (adds -100 (reg st) (reg st))
           (andnot 15 (reg st) (reg st))
	   (st!.l  (reg t1) (displacement (reg st) 0))
	   (call (foreignentry ,functionname))
	   (nop)
           (ld!.l (displacement (reg st) 0) (reg st))
	   (*move (quote NIL) ($fluid *kernelmode))))

(defcmacro !*foreignlink)

(defcmacro *movex)
   % Movex is of type cmacro

(de *movex (src dst)
     (let (tmp)
      (cond ((i860-register-p dst)
          (cond ((i860-register-p src)  `((shl (reg 0) ,src ,dst)))
		((eqcar src 'idloc)     `((or ,src (reg zero) ,dst)))
                ((eqcar src 'quote)
                     (if *writingasmfile           %% asmout and lap generate
			 `((orh  (hi ,src) (reg 0) ,dst)  %% different code.
			   (or   (lo ,src) ,dst ,dst))
			 `((move-relocatable-object ,src ,dst))))
                ((i860-absolute-p src)
                 (if *writingasmfile           %% asmout and lap generate
                     `((orh (hi ,src) (reg 0) ,dst)  %% different code.
			(ld!.l (displacement ,dst (lo ,src)) ,dst))
                     `((move-relocatable-object-and-load ,dst ,src ))))
		((i860-fluid-p src)
		     (if *writingasmfile           %% asmout and lap generate
			`((orh (hi ,src) (reg 0) ,dst)%% different code.
			  (ld!.l (displacement ,dst (lo ,src)) ,dst))
			`((ld!.l (valueaddress ,src) ,dst))))
                ((i860-short-immediate-p src) `((adds ,src (reg 0) ,dst)))
                ((or (i860-relocatable-immediate-p src)
                     (i860-foreignentry-p src))
                 (if *writingasmfile           %% asmout and lap generate
		      `((orh  (hi ,src) (reg 0) ,dst)  %% different code.
			(or   (lo ,src) ,dst ,dst))
                     `((move-relocatable-object ,src ,dst))))
                ((i860-immediate-p src)
                     `((orh (hi ,src) (reg 0) ,dst)
			(or  (lo ,src) ,dst ,dst)))
                ((i860-huge-displacement-p src)
		      `((orh (hi ,(caddr src)) (reg 0) ,dst)
			(adds ,(cadr src) ,dst ,dst)
			(ld!.l (displacement ,dst (lo ,(caddr src))) ,dst)))
                ((and (i860-deferred-p src) (i860-indirect-p (cadr src)))
			(car (load-src-arg src dst)))
                ((i860-indirect-p src) `((ld!.l ,src ,dst)))
                (t  (error 0
                 (list "Unknown addressing mode for *movex:" src dst)))))
         ((i860-register-p src)
          (cond ((or (i860-absolute-p dst) (i860-foreignentry-p src))
                 (if *writingasmfile
		      `((orh (hi ,dst) (reg 0) (reg t6))
			(st!.l ,src (displacement (reg t6) (lo ,dst))))
                     `((move-relocatable-object-and-store ,dst (reg t6) ,src))))
		((indexedp dst)
			`((*wplus3 (reg t6) ,(cadr dst) ,(caddr dst) )
			  (st!.l ,src (displacement (reg t6) 0))))
		((and (i860-fluid-p dst) *writingasmfile)
		        `((orh (hi ,dst) (reg 0) (reg t6))
			  (st!.l ,src (displacement (reg t6) (lo ,dst)))))
		((i860-fluid-p dst) `((st!.l ,src (valueaddress ,dst ))))
                ((and (i860-deferred-p dst) (i860-indirect-p (cadr dst)))
                 (setq tmp (load-src-arg (cadr dst) '(reg t6)))
                 (append (car tmp) `((st!.l ,src (deferred (reg t6))))))
                ((i860-huge-displacement-p dst)
                 `((orh (hi ,(caddr dst)) (reg 0) (reg t6))
		   (adds ,(cadr dst) (reg t6) (reg t6))
                   (st!.l ,src
			 (displacement (reg t6) (lo ,(caddr dst))) )))
                ((i860-indirect-p dst) `((st!.l ,src ,dst)))
                (t (error 0
			 (list "Unknown addressing mode:" src dst)))))
         (t (print (list " Unsaeglich: Move " src dst))
))))
