%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           PXC:Mips-cmac.sl
% Title:          Patterns and predicates for Mips PSL cmacro expansion
% Author:         Winfried Neun, ZIB Berlin
% Created:        September 15, 1989
% Modified:
% Status:         Experimental
% Mode:           Lisp
%
% Copyright (c) 1989, Konrad-Zuse-Zentrum Berlin, All Rights Reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(on fast-integers)

(compiletime
  (progn (on eolinstringok)
	 (dm * (u)  nil)
	 (load if-system)
	 (setq *syslisp t)
	 nil))

(global '(*writingasmfile))

(fluid '(AddressingUnitsPerItem
	 CharactersPerWord
	 StackDirection
	 *ImmediateQuote
         *usefastframe
	 AddressingUnitsPerFunctionCell))

(loadtime (on immediatequote))

(flag '(Immediate UnImmediate Indirect displacement
	label extrareg Indexed deferred)           'TerminalOperand)

% MkItem may be used when evaluating WConst expressions.

(de MkItem (TagPart InfPart) 
	  (lor (lsh TagPart 27) (land InfPart 16#7ffffff)))
  
(de regp (x) (eqcar x 'reg))

(de freep (asize) (if (and (wleq asize 4) (not &links-present&))
                      (setq *usefastframe t) 
                      (setq *usefastframe nil)))

(de FiveP (x)  (eq x 5))

(de TwentySevenP (x)  (eq x 27))

(de evenp (x) (and (fixp x) (eq 0 (land x 1)))) 

(de indexedp (x) (eqcar x 'indexed))
 
(de DeferrableP (Expression)
  (and (PairP Expression)
       (not (memq (first Expression)
		  '(Deferred $Fluid $Global Fluid Global)))))

(de RegisterDeferredP (Expression)
  (and (EqCar Expression 'Deferred) (EqCar (second Expression) 'Reg)))

(de FakeRegisterNumberP (Expression)
      (and (IntP Expression) (wgreaterP Expression 5)))

% NegativeImmediateP is for numbers whose negative is a one byte operand.

(de negp (value) (and (numberp value) (lessp value 0)))

(de NegativeImmediateP (Expression)
  (and (FixP Expression) (minusp Expression) (GreaterP Expression -64)))

% expansions of addressing modes.

(DefAnyreg CAR  ZIBanyregCAR)
 
(de ZIBAnyregCAR (reg source)
   (prog (areg)
      (setq source (resolveoperand reg source))
      (if (and (displacementp source)     % recursive case, e.g. caar
	       (member (cadr source) *nextreglist*))
	     (setq areg (cadr source))
	     (setq areg (nexttempreg)))
      (if (regp source) (return `(displacement ,source 0))
  
          (push `(*move ,source , areg) Resultingcode*))
      (return `(displacement ,areg 0))))
 
(DefAnyreg CDR  ZIBanyregCDR)  % ZIB

(de ZIBAnyregCDR (reg source)
   (prog (areg)
      (setq source (resolveoperand reg source))
      (if (and (displacementp source)     % recursive case, e.g. cddar
	       (member (cadr source) *nextreglist*))
	     (setq areg (cadr source))
	     (setq areg (nexttempreg)))
      (if (regp source) (return `(displacement ,source 8))
          (push `(*move ,source , areg) Resultingcode*))

      (return `(displacement ,areg 8))))
 
(fluid '(*nextreg* *nextreglist*))

(setq *nextreg* 0        % Three registers should be sufficient
      *nextreglist* '((reg t2) (reg t3) (reg t7)))
  
(de nexttempreg ()   % ZIB  :This saves Regs , non static assignment
     (setq *nextreg* (wplus2 *nextreg* 1))
     (when (eq *nextreg* 4) (setq *nextreg* 1))
     (nth *nextreglist* *nextreg*)) 

(DefAnyreg Memory  ZIBanyregMemory)  % ZIB 

(de ZIBAnyregMemory (reg source offset)
  (prog (aReg cadrsource)
   (setq source (resolveoperand reg source))
   (setq offset (resolveoperand reg offset))
   (cond ((and   (registerp source) (intp offset)
                 (mips-short-immediate-p  offset))
               (return   `(displacement ,source offset) ))
         ((registerp source)
               (setq AReg (nexttempreg)) % take a new TempReg
                (push `(*move ,offset ,aReg) Resultingcode*)
                (push `(*wplus2 ,areg ,source) Resultingcode*)
               (return   `(displacement ,areg 0) ))
	  ((and (AddressConstantP source)(ZeroP offset)) 
		        (return  `(unimmediate ,SOURCE)))
	  ((and (RegisterDeferredP source)(ZeroP offset))
	     (return `(deferred (Displacement (UnDeferred ,SOURCE) 0)))))
    (setq aReg (if (and (displacementp source) 
		       (member (setq cadrsource (cadr source)) *nextreglist*))
		   cadrsource      (nexttempreg))) % take a new TempReg
    (push `(*wplus3 ,areg ,source ,offset) Resultingcode*)
    (return `(displacement ,aReg  ,0))))


(DefAnyreg QUOTE
	   AnyregQUOTE
	   ((Null) (REG NIL))
	   ((EqTP) (reg t))  
	   ((InumP) SOURCE)
	   (       (QUOTE SOURCE)))

(DefAnyreg FRAME ZIBAnyregFrame)  % ZIB

(de ZIBAnyregFRAME (dum number)
     (if *usefastframe (nth '((reg ff1) (reg ff2) (reg ff3) (reg ff4)) number)
        `(displacement (reg st) ,(times2 8 number))))

(DefAnyreg xFRAME ZIBAnyregxFrame) 

(de ZIBAnyregxFRAME (dum number)   % regardless of *usefastframe
    `(displacement (reg st) ,(times2 4 number)))

(DefAnyreg REG ZIBAnyregREG)
  
(de ZIBanyregReg (reg source)
   (cond ((Fakeregisternumberp source) % argumentblock lies just below SYMVAL 
	     `(indexed (reg symvalcenter) ,(wtimes2 (wplus2 source -8017) 4)))
	 (t   `(REG ,SOURCE))))

(loadtime (remflag '(fluid global $fluid $global) 'Terminaloperand))

(defanyreg Fluid ZIBanyregfluid)

(defanyreg $Fluid ZIBanyregfluid)

(fluid '(*syslisp ))

(de ZIBanyregFluid (reg source)
   (cond ((and (idp source) (get source 'registername))
		         `(reg ,(get source 'registername)))
	 (t       `($Fluid ,SOURCE))))

(defanyreg Global ZIBanyregglobal)

(defanyreg $Global ZIBanyregglobal)

(de ZIBanyregGlobal (reg source)
   (cond ((and (idp source) (get source 'registername))
		         `(reg ,(get source 'registername)))
	 (t       `($global ,SOURCE))))

%-----------------------------------------------------------------------------

(DefCMacro *Call
       ((InternallyCallableP)
		  (jal (InternalEntry ArgOne))
		  (NOP))

	((asmp)   (lw (reg t4) (plus2 SYMFNC (times2 4 (idloc argone))) )
		  (*move (idloc argone) (reg t3))
		  (jalr (reg returnaddr) (reg t4)))

	(         (lw (reg t4) (entryaddress argone))
		  (nop)
		  (jalr (reg returnaddr) (reg t4))
		  (ori (reg t3) (reg 0) (idloc argone) )))
  
(DefCMacro *JCall
	((InternallyCallableP) (j (InternalEntry ArgOne))
		                  (nop))

	((asmp)   (lw (reg t4) (plus2 SYMFNC (times2 4 (idloc argone))) )
		  (*move (idloc argone) (reg t3))
		  (jr (reg t4)))

	(         (lw (reg t4) (entryaddress argone))
		  (nop)
		  (jr (reg t4))
		  (ori (reg t3) (reg 0) (idloc argone) )))

(de quote-fix-p (x)
    (and (eqcar x 'quote) (fixp (cadr x))))

(de displacement-p (x)
    (and (eqcar x 'displacement)
	 (eqn (length x) 3) (mips-register-p (cadr x))))

(de deferred-p (x)
    (and (pairp x) (or (mips-register-p (car x))
		       (eq (car x) 'deferred))))

(remprop '*move 'optfn)
 
(DefCMacro *Move
	(Equal)
	((regp regp)                (or ArgTwo (reg ZERO) ArgOne))
	((Mips-short-immediate-p regp)
		                    (addi ArgTwo (reg zero) ArgOne))
	((indexedp regp)            (lw ArgTwo ArgOne) (nop)) 
	((regp indexedp)            (sw ArgOne ArgTwo)) 
	((regp anyp)                (*movex ArgOne ArgTwo)) 
	((anyp regp)                (*movex ArgOne ArgTwo)) 
	(                           (*move ArgOne (reg t5))
		                    (*move (reg t5) ArgTwo)))

% At Alloc and Dealloc we have to add room for the returnaddr register, since
% this must be saved by the user

(de *ALLOC (framesize)
    
    (setq NAlloc!* framesize)
    (cond ((freep framesize) NIL)
          (t  %(when (eq 0 (wand framesize 1)) % set st on double word boundary
              %          (setq framesize (iadd1 framesize)))
             `((addiu (reg st) (reg st) ,(minus (plus2 8 (times2 8 framesize))))               (sw (reg returnaddr) (displacement (reg st) 0))))
)   )


(defcmacro *ALLOC)

(de *xalloc (framesize)      % alloc in any case, regardless of freep
%   (when (eq 0 (wand framesize 1)) % set st on double word boundary
%                        (setq framesize (iadd1 framesize)))
    (setq framesize (wtimes2 4 framesize))
    `((addiu (reg st) (reg st) ,(minus (plus2 8 framesize)))
      (sw (reg returnaddr) (displacement (reg st) 0))))
 
 
(defcmacro *xALLOC)
 
(DefCMacro *DeAlloc
           ((FreeP))
%          ((evenp) (lw (reg returnaddr) (displacement (reg st) 0))
%                   (*wplus2 (reg st) (plus2 8 (times2 4 ArgOne))))
           (    (lw (reg returnaddr) (displacement (reg st) 0))
                (*wplus2 (reg st) (plus2 4 (times2 4 ArgOne)))))
 
(de *xDeAlloc (framesize)
%           (when (eq 0 (wand framesize 1)) % set st on double word boundary
%                        (setq framesize (iadd1 framesize)))
           `((lw (reg returnaddr) (displacement (reg st) 0))
             (*wplus2 (reg st) (plus2 4 (times2 4 ,framesize)))))
 
(DefCMacro *xDeAlloc)

(de asmp () *writingasmfile)
 
(DefCMacro *Exit
         % leaf routine first
           ((Freep)  (jr (reg returnaddr))
                     (nop))
       %   (((lambda (x) (and (asmp) (evenp x))))
       %        (lw (reg returnaddr) (displacement (reg st) 0))
       %        (addiu  (reg st) (reg st) (plus2 16 (times2 8 ArgOne)))
       %        (jr (reg returnaddr))
       %        (nop))
           ((asmp)
                (lw (reg returnaddr) (displacement (reg st) 0))
                (addiu  (reg st) (reg st) (plus2 8 (times2 8 ArgOne)))
                (jr (reg returnaddr))
                (nop))
       %   ((evenp)
       %        (lw (reg returnaddr) (displacement (reg st) 0))
       %        (nop)
       %        (jr (reg returnaddr))
       %        (*wplus2 (reg st) (plus2 16 (times2 8 ArgOne))))
           (    (lw (reg returnaddr) (displacement (reg st) 0))
                (nop)
                (jr (reg returnaddr))
                (*wplus2 (reg st) (plus2 8 (times2 8 ArgOne))))
)

(de deferredp (x) (eqcar x 'deferred))

(de displacementp (x) (eqcar x 'displacement))

(DefCMacro *Jump
	   ((deferredp)        (*move (undeferred ArgOne) (reg t5))
		               (jr (reg t5))
		               (or (reg 0) (reg 0) (reg 0)))
		                                 % NO NOP!!!!! FOR LAPOPT
	   (                   (j ArgOne)
		               (nop)))
     

(DefCMacro *Lbl           (ArgOne))
	     
(de *3op (a1 a2 instruction imminstruction)
  (prog (ResultingCode*)
    (return (CMacroPatternExpand
	       (list (ResolveOperand '(Reg t1) A1)
		     (ResolveOperand '(Reg t2) A2)
		      Instruction imminstruction)
		(get '*3op 'CMacroPatternTable)))))

(DefCMacro *3op
	   ((regp regp) (ArgThree ArgOne ArgOne ArgTwo))
	   ((regp mips-short-positive-immediate-p) 
		        (ArgFour ArgOne ArgOne ArgTwo))
	   ((regp AnyP) (*move ArgTwo (reg T5))
		        (*3op ArgOne (reg t5) ArgThree))
	   ((Anyp regp) (*move ArgOne (reg T5))
		        (ArgThree (reg t5) (reg t5) ArgTwo)
		        (*move (reg t5) ArgOne))
	   ((Anyp mips-short-positive-immediate-p)
		        (*move ArgOne (reg T6))
		        (ArgFour (reg t5) (reg t6) ArgTwo)
		        (*move (reg t5) ArgOne))
	   (            (*move ArgTwo (reg T5))
		        (*move ArgOne (reg T6))
		        (ArgThree (reg t5) (reg t6) (reg t5))
		        (*move (reg t5) ArgOne)))

(DefCMacro *WPlus2
           ((regp regp) (Addu ArgOne ArgOne ArgTwo))
           ((regp mips-short-immediate-p)
                        (addiu ArgOne ArgOne ArgTwo))
           ((regp AnyP) (*move ArgTwo (reg T5))
                        (*wplus2 ArgOne (reg t5)))
           ((Anyp regp) (*move ArgOne (reg T5))
                        (Addu (reg t5) (reg t5) ArgTwo)
                        (*move (reg t5) ArgOne))
           ((Anyp mips-short-immediate-p)
                        (*move ArgOne (reg T6))
                        (Addiu (reg t5) (reg t6) ArgTwo)
                        (*move (reg t5) ArgOne))
           (            (*move ArgTwo (reg T5))
                        (*move ArgOne (reg T6))
                        (Addu (reg t5) (reg t6) (reg t5))
                        (*move (reg t5) ArgOne)))

(de *wplus3(arg1 arg2 arg3)
    (expand3operandcmacro arg1 arg2 arg3 '*wplus3))

(DefCMacro *WPlus3
	   ((regp regp regp) (addu ArgOne ArgTwo ArgThree))
	   ((regp mips-short-immediate-p regp)
		             (addiu ArgOne ArgThree ArgTwo))
	   ((regp regp mips-short-immediate-p)
		             (addiu ArgOne ArgTwo ArgThree))
	   ((regp mips-short-immediate-p anyp)
		             (*move ArgThree (reg t5))
		             (addiu ArgOne (reg t5) ArgTwo))
	   ((regp anyp mips-short-immediate-p)
		             (*move ArgTwo (reg t5)) 
		             (addiu ArgOne (reg t5) ArgThree))
	   ((regp anyp anyp) (*move ArgTwo ArgOne)
		             (*wplus2 ArgOne ArgThree))
	   ((anyp regp anyp) (*move ArgThree (reg t5))
		             (*wplus2 (reg t5) ArgTwo)
		             (*move (reg t5) ArgOne))
	   (                 (*move ArgTwo (reg t5))
		             (*move ArgThree (reg t6))
		             (addu (reg t5) (reg t6) (reg t5))
		             (*move (reg t5) ArgOne)))

(DefCMacro *WDifference
	   ( (*wdifference3 ArgOne ArgOne ArgTwo )))

(de *wdifference3 (arg1 arg2 arg3)
    (expand3operandcmacro arg1 arg2 arg3 '*wdifference3))

% format ArgOne = ArgTwo - ArgThree

(DefCMacro *Wdifference3
	   ((regp regp regp) (subu ArgOne ArgTwo ArgThree))
	   ((regp regp mips-short-immediate-p)
		             (addiu Argone ArgTwo (minus Argthree)))
	   ((regp anyp regp) (*move ArgTwo (reg t5))
		             (subu ArgOne (reg t5) ArgThree))
	   ((regp anyp anyp) (*move ArgTwo (reg t5))
		             (*move ArgThree (reg t6))
		             (subu ArgOne (reg t5) (reg t6)))
	   ((anyp regp anyp) (*move ArgThree (reg t5))
		             (subu (reg t5) ArgTwo (reg t5))
		             (*move (reg t5) ArgOne))
	   (                 (*move ArgTwo (reg t5))
		             (*move ArgThree (reg t6))
		             (subu (reg t5) (reg t5) (reg t6))
		             (*move (reg t5) ArgOne)))

(defcmacro *wtimes2)

(de *wtimes2 (arg1 arg2) 
		(*wtimes3 arg1 arg1 arg2))

(de *wtimes3 (arg1 arg2 arg3)
    (expand3operandcmacro arg1 arg2 arg3 '*wtimes3))

% format ArgOne = ArgTwo * ArgThree

(DefCMacro *wtimes3
	   ((anyp anyp onep) (*move ArgTwo ArgOne))
	   ((regp regp regp)
		    (mult  ArgTwo Argthree)
		    (mflo  Argone))
	   ((regp regp anyp)
		    (*move argthree (reg t1))
		    (*wtimes3 ArgOne ArgTwo (reg t1)))
	   (        (*move ArgTwo (Reg t1))
		    (*move ArgThree (reg t2))
		    (mult (Reg t1) (reg t2))
		    (mflo (reg t1))
		    (*move (reg t1) ArgOne)))
 
(defcmacro *wquotient)
 
(de *wquotient (arg1 arg2) (*wquotient3 arg1 arg1 arg2))
 
(de *wquotient3 (arg1 arg2 arg3)
    (expand3operandcmacro arg1 arg2 arg3 '*wquotient3))
 
% format ArgOne = ArgTwo * ArgThree
 
(DefCMacro *wquotient3
         ((anyp anyp onep)         (*move argtwo argone))
         ((regp fixp fixp)  (*move (quotient ArgTwo ArgThree) ArgOne))
	 ((regp regp powerof2p)  (*wquotientshift ArgOne ArgTwo ArgThree))
         ((regp regp regp)       (ddiv ArgTwo Argthree)
				 (mflo  Argone)
                                 (or (reg 1) (reg 1) (reg 1)))
	 ((regp regp anyp)        (*move argthree (reg t1))
				 (*wquotient3 ArgOne ArgTwo (reg t1)))
	   (        (*move ArgTwo (Reg t1))
		    (*move ArgThree (reg t2))
                    (ddiv (Reg t1) (reg t2))
		    (mflo (reg t1))
                    (or (reg 1) (reg 1) (reg 1))
		    (*move (reg t1) ArgOne)))

(de *wquotientshift (arg1 arg2 arg3)
	(setq arg3 (powerof2p arg3))
	`((srl (reg t5) ,arg2 ,(difference 32 arg3))
	  (*wplus2 (reg t5) ,arg2)
	  (sra ,arg1 (reg t5) ,arg3 )))

(defcmacro *wquotientshift)

(put 'wremainder 'opencode
          '((ddiv (reg 1) (reg 2))
	    (mfhi (reg 1))
	    (or (reg 1) (reg 1) (reg 1))))

(DefCMacro *WAnd	   ((*3op ArgOne ArgTwo and andi)))

(DefCMacro *WOr  	   ((*3op ArgOne ArgTwo or ori)))

(DefCMacro *WXOr	   ((*3op ArgOne ArgTwo xor xori)))

(de fixplusp (x) (and (fixp x) (not (minusp x))))

(DefCMacro *AShift
	((AnyP ZeroP))
	((regp negp)     (sra ArgOne ArgOne (minus ArgTwo) ))
	((regp fixplusp) (sll ArgOne ArgOne ArgTwo))
	((regp regp)     (bltz ArgTwo templabel)
		          (nop)
		          (sllv ArgOne ArgOne ArgTwo)
		          (j templabel2)
		          (nop)
		         templabel
		          (sub (reg t6) (reg NULL) ArgTwo) % delay slot
		          (srav ArgOne ArgOne (reg t6))
		         templabel2)
	((regp anyp)     (*move ArgTwo (reg t5))
		         (*ashift ArgOne (reg t5)))
	((Anyp fixp)     (*move ArgOne (reg t5))
		         (*ashift (reg t5) ArgTwo)
		         (*move (reg t5) ArgOne))
	(                (*move ArgOne (reg t1))
		         (*ashift (reg t1) ArgTwo)
		         (*move (reg t1) ArgOne)))

(DefCMacro *WShift
	((AnyP ZeroP))
        ((regp neg32p)     (dsrl ArgOne ArgOne (minus ArgTwo)))
        ((regp fixplus32p) (dsll ArgOne ArgOne ArgTwo))
        ((regp fixplusp) (dsll32 ArgOne ArgOne ArgTwo))
        ((regp negp)     (dsrl32 ArgOne ArgOne (minus ArgTwo)))
        ((regp regp)     (bltz ArgTwo templabel)
                          (nop)
                          (dsllv ArgOne ArgOne ArgTwo)
		          (j templabel2)
		          (nop)
		         templabel
                          (dsub (reg t6) (reg NULL) ArgTwo) % delay slot
                          (dsrlv ArgOne ArgOne (reg t6))
		         templabel2)
	((regp anyp)     (*move ArgTwo (reg t5))
		         (*wshift ArgOne (reg t5)))
	((Anyp fixp)     (*move ArgOne (reg t5))
		         (*wshift (reg t5) ArgTwo)
		         (*move (reg t5) ArgOne))
	(                (*move ArgOne (reg t1))
		         (*wshift (reg t1) ArgTwo)
		         (*move (reg t1) ArgOne)))
		                  
(DefCMacro *WNot
	   ((regp regp)  (nor ArgOne ArgTwo (reg zero)))
	   ((regp Anyp)  (*move ArgTwo ArgOne)
		         (nor ArgOne ArgOne (reg zero)))
	   (             (*move ArgTwo (reg t5))
		         (*wnot (reg t5) (reg t5))
		         (*move (reg t5) ArgOne))) 

(DefCMacro *WMinus
           ((regp regp)  (dsub ArgOne (reg 0) ArgTwo))
	   ((regp Anyp)  (*move ArgTwo ArgOne)
                         (dsub ArgOne (reg 0) ArgOne))
	   (             (*move ArgTwo (reg t5))
		         (*wminus (reg t5) (reg t5))
		         (*move (reg t5) ArgOne))) 

(DefCMacro *MkItem
       ((regp fixp)      (and ArgOne (reg infmask) ArgOne)
		          (*speclui ArgTwo)      % lui with tag to t5
		          (or ArgOne (reg t5) ArgOne))
       ((regp regp)      (and ArgOne (reg infmask) ArgOne)
		          (sll (reg t5) ArgTwo 27 )
		          (or ArgOne (reg t5) ArgOne))
       ((regp anyp)      (*move ArgTwo (reg t5))
		          (*Mkitem ArgOne (reg t5)))
       (                 (*move ArgOne (reg t1))
		          (*mkitem (reg t1) ArgTwo)
		          (*move (reg t1) ArgOne))) 

(de *speclui (num)          % avoid shift of constant at compiletime
	`((lui (reg t5) ,(wshift num 11)))) 

(defcmacro *speclui)  % support for *mkitem

(de smallinump (x) (mips-short-immediate-p x))

(de immediate-p (x)    (fixp x))

(DefCmacro *tag
          ((regp regp) (dsrl32 ArgOne ArgTwo 27 ))
	  ((Anyp  regp)(*move ArgTwo (reg t1))
                        (dsrl (reg t1) (reg t1) 27)
		        (*move (reg t1) ArgOne))
	  ((regp anyp) (*move ArgTwo (reg t1))
                        (dsrl ArgOne (reg t1) 27))
	  (             (*move ArgTwo (reg t1))
		        (*tag ArgOne (reg t1))))

(de *tag (ar1 ar2)
     (expand2operandCmacro ar1 ar2 '*tag))

(DefCMacro *JumpType
	  (             (*Tag (Reg t6) ArgOne)
		        (*JumpEQ ArgThree (Reg t6) ArgTwo)))

(DefCMacro *JumpNotType
	  (             (*Tag (Reg t6) ArgOne) 
		        (*JumpNOTEQ ArgThree (Reg t6) ArgTwo)))

(DefCMacro *JumpInType
	   ((regp regp) (srl (Reg t5) ArgOne 27)
		         (subu (Reg t6) (reg t5) ArgTwo)
		         (blez (reg t6) ArgThree)
		         (*move 31 (reg t6))
		         (beq  (reg t5) (Reg t6) Argthree) % negint-tag)
		         (nop))
	   ((regp Anyp) (srl (Reg t5) ArgOne 27)
		         (*Move Argtwo (reg t6)) 
		         (subu (Reg t6) (reg t5) (reg t6))
		         (blez (reg t6) ArgThree)
		         (*move 31 (reg t6))
		         (beq  (reg t5) (Reg t6) Argthree) % negint-tag)
		         (nop))
	   (            (*move ArgOne (reg t5))
		         (*move ArgTwo (reg t6))        
		         (srl (Reg t5) (Reg t5) 27)
		         (subu (reg t6) (reg t5) (Reg t6))
		         (blez (reg t6) ArgThree)
		         (*move 31 (reg t6))
		         (beq (reg t6) (reg t5) Argthree) % negint-tag)
		         (nop)))

(DefCMacro *JumpNotInType % should optimize with slti instr
	   ((regp Regp) (srl (reg t5) ArgOne 27)
		         (*move 31 (reg t6))
		         (beq (reg t5) (reg t6) templabel)
		         (subu  (reg t6) (reg t5) ArgTwo)
		         (bgtz (reg t6) ArgThree)
		         (nop)
		        templabel)
	   ((regp Anyp) (srl (reg t5) ArgOne 27)
		         (*move 31 (reg t6))
		         (beq (reg t5) (reg t6) templabel)
		         (*move ArgTwo (reg t6))
		         (subu  (reg t6) (reg t5) (Reg t6))
		         (bgtz (reg t6) ArgThree)
		         (nop)
		        templabel)
	   (            (*move ArgOne (reg t5))
		         (srl (reg t5) (reg t5) 27)
		         (*move  31 (reg t6))
		         (beq (reg t5) (reg t6) templabel)
		         (*move ArgTwo (reg t6))
		         (subu (reg t6) (reg t5) (reg t6))
		         (bgtz (reg t6) ArgThree)
		         (nop)
		        templabel))

%---------------------------------------------------------------------
%   General Purpose Jumpers. (EQ ane BNE have different formats
%   FORMAT: Operand1 Operand2 Address JumpOpCode

(DefCMacro *JumpIf  
	 ((regp regp)       (slt (reg t6) ArgOne ArgTwo)
		             (ArgFour (reg t6) (reg 0) ArgThree)
		             (nop))
	 ((regp SmallInumP) (slti (reg t6) ArgOne ArgTwo) 
		             (ArgFour (reg t6) (reg 0) ArgThree) 
		             (nop))
	 ((regp anyp)       (*move ArgTwo (reg t5)) 
		             (slt (reg t6) ArgOne (reg t5))  
		             (ArgFour (reg t6) (reg 0) ArgThree)  
		             (nop))
	 ((anyp  regp)      (*move ArgOne (reg t5)) 
		             (slt (reg t6) (reg t5) ArgTwo)  
		             (Argfour (reg t6) (reg 0) ArgThree)  
		             (nop))
	 ((anyp  SmallInumP) (*move ArgOne (reg t5))
		             (slti (reg t6) (reg t5) ArgTwo) 
		             (ArgFour (reg t6) (reg 0) ArgThree) 
		             (nop))
	 (                   (*move ArgOne (reg t5))
		             (*move ArgTwo (reg t4))
		             (slt (reg t6) (reg t5) (reg t4)) 
		             (ArgFour (reg t6) (reg 0) ArgThree) 
		             (nop)))

(de *JumpIF (ArgOne ArgTwo Label Instruction)
  (prog (ResultingCode*)
    (return (CMacroPatternExpand
	       (list (ResolveOperand '(Reg t1) ArgOne)
		     (ResolveOperand '(Reg t2) ArgTwo)
		     (ResolveOperand '(Reg Error) Label)
		      Instruction)
		     (get '*JumpIF 'CMacroPatternTable)))))
 
(DefCMacro *JumpIfeq
	 ((regp  smallinump) (addi (reg t5) ArgOne (minus argtwo))
                             (ArgFour (reg t5) (reg zero) ArgThree)  
		             (nop))
	 ((anyp  smallinump) (*move ArgOne (reg t5)) 
		             (addi (reg t5) (reg t5) (minus argtwo))
                             (ArgFour (reg t5) (reg zero) ArgThree)  
		             (nop))
         ((regp regp)       (ArgFour ArgOne ArgTwo ArgThree)
                             (nop))
         ((regp anyp)       (*move ArgTwo (reg t5))
                             (ArgFour ArgOne (Reg t5) ArgThree)
                             (nop))
         ((anyp  regp)      (*move ArgOne (reg t5))
                             (ArgFour (reg t5) ArgTwo ArgThree)
                             (nop))
	 (                  (*move ArgOne (reg t5))
		             (*move ArgTwo (reg t4))
		             (Argfour (reg t5) (reg t4) ArgThree) 
		             (nop)))

(de *JumpIFeq (ArgOne ArgTwo Label Instruction)
  (prog (ResultingCode*)
    (return (CMacroPatternExpand
	       (list (ResolveOperand '(Reg t1) ArgOne)
		     (ResolveOperand '(Reg t2) ArgTwo)
		     (ResolveOperand '(Reg Error) Label)
		      Instruction)
		     (get '*JumpIFeq 'CMacroPatternTable)))))
 

(defcmacro *jumpeq)

(de *jumpeq (Lbl ArgOne ArgTwo)
	   (*jumpifEQ ArgOne ArgTwo lbl 'BEQ))

(defcmacro *jumpnoteq)

(de *jumpnoteq(Lbl ArgOne ArgTwo)
	    (*jumpifEQ ArgOne ArgTwo Lbl 'BNE))

(defcmacro *jumpwlessp)

(de *jumpwlessp (Lbl ArgOne ArgTwo)
	    (*jumpif ArgOne ArgTwo lbl 'BNE))

(defcmacro *jumpwgreaterp)

(de *jumpwgreaterp (Lbl ArgOne ArgTwo)
	    (*jumpif ArgTwo ArgOne lbl 'BNE))

(defcmacro *jumpwleq)

(de  *jumpwleq(Lbl ArgOne ArgTwo) 
	    (*jumpif ArgTwo ArgOne lbl 'Beq))

(defcmacro *jumpwgeq)

(de *jumpwgeq (Lbl ArgOne ArgTwo) 
	    (*jumpif ArgOne ArgTwo lbl 'BEQ))


(de fluid-arg-p (x)
    (and (pairp x) (member (car x) '($fluid $global fluid global))))

(DefCMacro *Push   ( (*cerror "NO push ")))

(DefCMacro *Pop    ( (*cerror "NO pop ")))

(de *loc (ArgOne ArgTwo)
   (let ((resolvedarg (resolveoperand '(reg t2) ArgTwo)))
      (cond ((deferred-p resolvedarg)
	     `((*move (undeferred ,resolvedarg) ,ArgOne))) % get the address.
	     ((displacement-p resolvedarg)
		`((addi ,ArgOne ,(cadr resolvedarg) ,(caddr resolvedarg) )))
	     ((eqcar resolvedarg 'cdr)
		`((addi ,ArgOne ,(cadr resolvedarg) 4 )))
	     (t `((*movex ,resolvedarg ,ArgOne))))))
		
(DefCMacro *loc)

(DefCMacro *Field

  ((regp regp zerop fixp)   (srl ArgOne ArgTwo (difference 32 ArgFour)))
  ((regp regp fivep twentysevenp)
		            (and ArgOne ArgTwo (reg infmask)))
  ((regp regp fixp fixp)    (sll ArgOne ArgTwo ArgThree)  
		             (srl ArgOne ArgOne (difference 32 ArgFour)))
  ((regp anyp fixp fixp)    (*move ArgTwo (reg t5))
		             (sll (reg t5) (reg t5) ArgThree)  
		             (srl ArgOne (reg t5) (difference 32 ArgFour)))
  ((anyp regp fixp fixp)    (*field (reg t5) ArgTwo ArgThree ArgFour)
		             (*move (reg t5) ArgOne))
  (                          (*move ArgTwo (reg t6))      
		             (*field (reg t5) (reg t6) ArgThree ArgFour)
		             (*move (reg t5) ArgOne)))

(DefCMacro *SignedField

  ((regp regp fixp fixp)    (sll ArgOne ArgTwo ArgThree)  
		             (sra ArgOne ArgOne (difference 32 ArgFour)))
  ((regp anyp fixp fixp)    (*move ArgTwo (reg t5))
		             (sll (reg t5) (reg t5) ArgThree)  
		             (sra ArgOne (reg t5) (difference 32 ArgFour)))
  ((anyp regp fixp fixp)    (*Signedfield (reg t5) ArgTwo ArgThree ArgFour)
		             (*move (reg t5) ArgOne))
  (                          (*move ArgTwo (reg t6))      
		             (*Signedfield (reg t5) (reg t6) ArgThree ArgFour)
		             (*move (reg t5) ArgOne)))

% following is used by garbage collector only. Not very useful to
% optimize other cases than put inf field or put tag field (with constant)
		     
(DefCMacro *PutField

 ((regp regp fiveP TwentysevenP) (and (reg t6) ArgOne (reg infmask))
		                  (nor (reg t5) (reg infmask) (reg zero))
		                  (and ArgTwo (reg t5) ArgTwo) 
		                  (or  ArgTwo (reg t6) ArgTwo))
 ((regp regp Zerop fivep)        (sll (reg t5)  ArgOne 27)
		                  (and ArgTwo ArgTwo (reg infmask))
		                  (or   ArgTwo (reg t5) ArgTwo))
 ((inump regp Zerop fivep)       (*move  ArgOne (reg t5))
		                  (sll   (reg t5) (reg t5) 27)
		                  (and ArgTwo (reg infmask) ArgTwo)
		                  (or  ArgTwo (reg t5) ArgTwo))
 (          (*move -1 (reg t5))
	    (*move   ArgOne (reg t1))
	    (sll   (reg t5) (reg t5) (difference 32 ArgFour))
	    (*wshift (reg t5) (minus ArgThree)) % for 0 opt
	    (sll  (reg t1) (reg t1) (difference 32 (plus ArgThree ArgFour)))
	    (and  (reg t1) (reg t5) (reg t1))
	    (*move ArgTwo (reg t6))
	    (nor (reg t5) (reg t5) (reg zero)) % negate mask
	    (and  (reg t6) (reg t5) (reg t6))
	    (or   (reg t5) (reg t1) (reg t6))
	    (*move (reg t5) ArgTwo)))

	  
(deflist '((Byte        ((sll (reg 1) (reg 1) 2)
		         (*wplus3 (reg t1) (reg 1) (reg 2))
		         (lb (reg 1) (displacement (reg t1) 0))
		         (nop)))
	   (PutByte     ((sll (reg t1) (reg 1) 2)
		         (*wplus3 (reg t1) (reg t1) (reg 2))
		         (sb (reg 3) (displacement (reg t1) 0))))
	   (HalfWord    ((sll (reg 1) (reg 1) 2)
		         (sll (reg t1) (reg 2) 1) 
		         (*wplus3 (reg t1) (reg 1) (reg t1))
		         (lh (reg 1) (displacement (reg t1) 0))
		         (nop)))
	   (PutHalfWord ((sll (reg t2) (reg 1) 2)
		         (sll (reg t1) (reg 2) 1)
		         (*wplus3 (reg t1) (reg t2) (reg t1))
		         (sh (reg 3) (displacement (reg t1) 0)))))
  'OpenCode)

(&OneReg '(Byte PutByte HalfWord PutHalfWord))

(on fast-integers)

(for (in function '(idapply0 idapply1 idapply2 idapply3 idapply4))
     (from register 1)
     (do
      (put function 'opencode
	    `((*move (fluid symfnc) (reg t1))
	      (*wand (reg ,register) (reg infmask))
	      (*wplus2 (reg t1) (reg ,register))
	      (sll (reg t1) (reg t1) 2)
	      (lw (reg t2) (indexed (reg t1) 0))
	      (nop)
	      (jalr (reg returnaddr) (reg t2))
	      (*move (reg ,register) (reg t3))))
      (put function 'exitopencode 
	    `((*move (fluid symfnc) (reg t1))
	      (*wand (reg ,register) (reg infmask))
	      (*wplus2 (reg t1) (reg ,register))
	      (sll (reg t1) (reg t1) 2)
	      (lw (reg t2) (indexed (reg t1) 0)) 
	      (nop) 
	      (jr (reg t2))
	      (*move (reg ,register) (reg t3))))
))

(for (in function '(addressapply0 addressapply1 addressapply2
		    addressapply3 addressapply4))
     (from register 1)
     (do 
      (put function 'opencode
	     `( (sll (reg ,register) (reg ,register) 2)
		(jalr (reg returnaddr) (reg ,register))
		(or (reg 0) (reg 0) (reg 0)))) % NO NOP!!!!! FOR LAPOPT
      (put function 'exitopencode 
	     `( (sll (reg ,register) (reg ,register) 2)
		(jr (reg ,register))
		(or (reg 0) (reg 0) (reg 0)))) % NO NOP!!!!! FOR LAPOPT
      ))


% *feq, *fgreaterp and *flessp can only occur once in a function.

(deflist '((*wfix ((sll (reg 1) (reg 1) 2)
		   (lwc1 (reg f1) (deferred (reg 1)))
		   (lwc1 (reg f0) (displacement (reg 1) 4))
		   (nop)
		   (cvt!.w!.d (reg f2) (reg f0))
		   (mfc1 (reg 1) (reg f2))))
	   (*wfloat ((mtc1 (reg 2) (reg f2))
		     (sll (reg 1) (reg 1) 2)
		     (cvt!.d!.w (reg f0) (reg f2))
		     (swc1 (reg f1) (deferred (reg 1)))
		     (swc1 (reg f0) (displacement (reg 1) 4))))
	   (*fgreaterp
		    ((sll (reg 2) (reg 2) 2)
		     (sll (reg 1) (reg 1) 2)
		     (lwc1 (reg f1) (deferred (reg 1)))
		     (lwc1 (reg f0) (displacement (reg 1) 4))
		     (lwc1 (reg f3) (deferred (reg 2)))
		     (lwc1 (reg f2) (displacement (reg 2) 4))
		     (*move (quote t) (reg 1))
		     (c!.ngt!.d (reg f0) (reg f2))
		     (nop)
		     (bc1F *donefgreaterp*)
		     (nop)
		     (*move (reg nil) (reg 1))
		*donefgreaterp*))
	   (*flessp ((sll (reg 2) (reg 2) 2)
		     (sll (reg 1) (reg 1) 2)
		     (lwc1 (reg f1) (deferred (reg 1)))
		     (lwc1 (reg f0) (displacement (reg 1) 4))
		     (lwc1 (reg f3) (deferred (reg 2)))
		     (lwc1 (reg f2) (displacement (reg 2) 4))
		     (*move (quote t) (reg 1))
		     (c!.lt!.d (reg f0) (reg f2))
		     (nop)
		     (bc1T *doneflessp*)
		     (nop)
		     (*move (reg nil) (reg 1))
		*doneflessp*))
	   (*fplus2 ((sll (reg 3) (reg 3) 2)
		     (sll (reg 2) (reg 2) 2)
		     (lwc1 (reg f1) (deferred (reg 2)))
		     (lwc1 (reg f0) (displacement (reg 2) 4))
		     (lwc1 (reg f3) (deferred (reg 3)))
		     (lwc1 (reg f2) (displacement (reg 3) 4))
		     (sll (reg 1) (reg 1) 2)
		     (add!.d (reg f0) (reg f2) (reg f0))
		     (swc1 (reg f1) (deferred (reg 1)))
		     (swc1 (reg f0) (displacement (reg 1) 4))))
	   (*fdifference 
		    ((sll (reg 3) (reg 3) 2)
		     (sll (reg 2) (reg 2) 2)
		     (lwc1 (reg f1) (deferred (reg 2)))
		     (lwc1 (reg f0) (displacement (reg 2) 4))
		     (lwc1 (reg f3) (deferred (reg 3)))
		     (lwc1 (reg f2) (displacement (reg 3) 4))
		     (sll (reg 1) (reg 1) 2)
		     (sub!.d (reg f0) (reg f0) (reg f2))
		     (swc1 (reg f1) (deferred (reg 1)))
		     (swc1 (reg f0) (displacement (reg 1) 4))))
	   (*ftimes2 ((sll (reg 3) (reg 3) 2)
		     (sll (reg 2) (reg 2) 2)
		     (lwc1 (reg f1) (deferred (reg 2)))
		     (lwc1 (reg f0) (displacement (reg 2) 4))
		     (lwc1 (reg f3) (deferred (reg 3)))
		     (lwc1 (reg f2) (displacement (reg 3) 4))
		     (sll (reg 1) (reg 1) 2)
		     (mul!.d (reg f0) (reg f2) (reg f0))
		     (swc1 (reg f1) (deferred (reg 1)))
		     (swc1 (reg f0) (displacement (reg 1) 4))))
	   (*fquotient 
		    ((sll (reg 3) (reg 3) 2)
		     (sll (reg 2) (reg 2) 2)
		     (lwc1 (reg f1) (deferred (reg 2)))
		     (lwc1 (reg f0) (displacement (reg 2) 4))
		     (lwc1 (reg f3) (deferred (reg 3)))
		     (lwc1 (reg f2) (displacement (reg 3) 4))
		     (sll (reg 1) (reg 1) 2)
		     (div!.d (reg f0) (reg f0) (reg f2))
		     (swc1 (reg f1) (deferred (reg 1)))
		     (swc1 (reg f0) (displacement (reg 1) 4)))))
	 'opencode)


(defcmacro *fast-apply-load
   ( (*move ArgOne (reg t1)) ))

(put 'fast-idapply    'opencode  % id is passed in t1 Register
     '((*move (fluid symfnc) (reg t2))
       (and (reg t3) (reg infmask) (reg t1))
       (*wplus2 (reg t2) (reg t3))
       (sll (reg t2) (reg t2) 2) % by 4 to get offset into symfnc table
       (lw (reg t2) (indexed (reg t2) 0))
       (nop)
       (jalr (reg returnaddr) (reg t2))
       (nop)))

(put 'fast-codeapply    'opencode
     '((and (reg t1) (reg infmask) (reg t1))
       (sll (reg t1) (reg t1) 2)
       (jalr (reg returnaddr) (reg t1))
       (or (reg 0) (reg 0) (reg 0)))) % NO NOP!!!!! FOR LAPOPT

(put 'fast-idapply    'exitopencode
     '((*move (fluid symfnc) (reg t2))
       (and (reg t3) (reg infmask) (reg t1))
       (*wplus2 (reg t2) (reg t3))
       (sll (reg t2) (reg t2) 2) % by 4 to get offset into symfnc table
       (lw (reg t2) (indexed (reg t2) 0)) 
       (nop)
       (jr (reg t2))
       (nop)))

(put 'fast-codeapply    'exitopencode
     '((and (reg t1) (reg infmask) (reg t1))
		         %codepointer passed, so just strip off tag
       (sll (reg t1) (reg t1) 2) 
       (jr (reg t1))
       (or (reg 0) (reg 0) (reg 0)))) % NO NOP!!!!! FOR LAPOPT
 
(De *LamBind (Regs Fluids)
 
 % be careful, code is somewhat tricky
 
 (prog (n lng list cregs cfluids cadrcfluids genlabel initload freeregs)
      (setq genlabel (gensym))
      (setq n 0)
      (setq Regs (rest Regs))     % Remove "Registers" from the front
      (setq Fluids (rest fluids)) % Remove NONLOCALVARS
      (setq lng (wtimes2 (length Regs) 8)) % two words per BndStk entry *4
 
   %looking for free register to preload values of fluid s
 
      (setq freeregs (mapcan '((reg 1)(reg 2)(reg 3)(reg 4)(reg 5))
		       (function (lambda (x)
		                    (cond ((member x regs) nil)
		                           (t (cons x nil))
      )              ) )         )  )
      (setq freeregs (nconc (list '(reg T3) '(reg T7) ) freeregs))
      (setq cfluids fluids) % copy of fluids
 
preload  (setq initload
	   (progn (setq cadrcfluids
		    (nconc cadrcfluids (cons (car freeregs) nil)))
	    (nconc initload `((*move ,(car cfluids) ,(car freeregs))))
       )   )
       (setq freeregs (cdr freeregs))
       (setq cfluids (cdr cfluids))
 
       (cond ((and freeregs cfluids) (go preload)))
 
       (setq freeregs (nconc cadrcfluids cfluids)) %end of preloadcode
 
       % freeregs contains the list of preloaded regs
       % and not preloaded fluids if those exist
 
      (setq list `((sll (reg t1) (reg BndStkPtr) 2)
		   (*wplus3 (Reg t2) ,lng (Reg t1))
		   (sll (reg t5) (reg BndstkUpperBound) 2)
		   (sub  (reg t5) (reg t2) (reg t5))
		   (blez (reg t5) ,genlabel)
		   (srl (reg BndstkPtr) (reg t2) 2)
		   (*call Bstackoverflow)
		  ,genlabel  )) %start of code
 
      (setq list (append initload list))
 
  loop
      (setq cregs (car Regs))
      (setq cfluids (car Fluids))
      (setq cadrcfluids (cadr cfluids))
      (when (or (eq cadrcfluids 't) (eq cadrcfluids 'nil))
	       (stderror "T and NIL cannot be rebound"))
      (setq n (wplus2 n 8))
      (Setq list (append list
	 `((*move ,(car freeregs) (indexed (Reg t1) ,n))
	   (*move (quote ,Cadrcfluids)
		       (indexed (reg t1) ,(wplus2 n -4)))
	   (*move ,cregs ,cfluids)
      )          ))
      (setq fluids (cdr Fluids))
      (setq freeregs (cdr freeregs))
      (cond ((setq regs (cdr Regs)) (go loop)))
      (return list)
)    )

(defcmacro !*lambind)

(De *ProgBind (Fluids)
 
 % be careful, code is somewhat tricky
 
 (prog (n lng list cfluids cadrcfluids genlabel initload freeregs)
      (setq genlabel (gensym))
      (setq n 0)
      (setq Fluids (rest fluids)) % Remove NONLOCALVARS
      (setq lng (wtimes2 (length Fluids) 8)) % two words per BndStk entry
		                             % * 4 addressingunits
      (setq freeregs '((reg t3)(reg T7)
                       (reg 1)(reg 2)(reg 3)(reg 4)(reg 5)))
      (setq cfluids fluids) % copy of fluids
 
preload  (setq initload
	   (progn (setq cadrcfluids
		    (nconc cadrcfluids (cons (car freeregs) nil)))
	    (nconc initload `((*move ,(car cfluids) ,(car freeregs))))
       )   )
       (setq freeregs (cdr freeregs))
       (setq cfluids (cdr cfluids))
 
       (cond ((and freeregs cfluids) (go preload)))
 
       (setq freeregs (nconc cadrcfluids cfluids)) %end of preloadcode
 
       % freeregs contains the list of preloaded regs
       % and not preloaded fluids if those exist
 

      (setq list `((sll (reg t1) (reg BndStkPtr) 2)
		   (*wplus3 (Reg t2) ,lng (Reg t1))
		   (sll (reg t5) (reg BndstkUpperBound) 2)
		   (sub  (reg t5) (reg t2) (reg t5))
		   (blez (reg t5) ,genlabel)
		   (srl (reg BndstkPtr) (reg t2) 2)
		   (*call Bstackoverflow)
		  ,genlabel  )) %start of code
 
     (setq list (append initload list))
 
 loop
      (setq cfluids (car Fluids))
      (setq cadrcfluids (cadr cfluids))
      (when (or (eq cadrcfluids 't) (eq cadrcfluids 'nil))
	       (stderror "T and NIL cannot be rebound"))
      (setq n (wplus2 n 8))
      (Setq list (append list
		 `((*move ,(car freeregs) (indexed (Reg t1) ,n))
		   (*move (quote ,Cadrcfluids)
		               (indexed (reg t1)  ,(wplus2 n -4)))
		   (*move (quote nil) ,cfluids)
      )          ))
      (setq freeregs (cdr freeregs))
      (cond ((setq Fluids (cdr Fluids)) (go loop)))
      (return list)
)    )
(defcmacro *progbind)
 
(De *FreeRstr (Fluids)
 
 (prog (n lng list cfluids listfluids initload freeregs genlabel)
      (setq genlabel (gensym))
      (setq n 0)
      (setq Fluids (rest fluids)) % Remove NONLOCALVARS
      (setq lng (wtimes2 (length Fluids) 2)) % two words per BndStk entry
      (setq freeregs '((reg t3)(reg t7) (reg 2)(reg 3)(reg 4)(reg 5)))
      (setq cfluids fluids) % copy of fluids
      (setq n (wtimes2 4 (wdifference 2 lng)))
      (setq lng (wtimes2 lng 4)) % * addressingunitperitem
      (setq initload (list '(sll (reg t1) (reg Bndstkptr) 2)))
 
preload  (setq initload
	   (progn (setq listfluids
		   (if freeregs
		    (nconc listfluids (cons (car freeregs) nil))
		    (nconc listfluids (cons nil nil))) )
	    (nconc initload
	     (if freeregs
	 `((*move (indexed (reg t1) ,n) ,(car freeregs))) nil)
       )   ))
       (setq n (wplus2 n 8))
       (when freeregs (setq freeregs (cdr freeregs)))
       (setq cfluids (cdr cfluids))
 
       (cond (cfluids (go preload)))
 
       (setq freeregs listfluids ) %end of preloadcode
 
       % freeregs contains the list of preloaded regs
       % and nil if not enough regs available
 
      (setq list `((*wplus3 (Reg t2) (Reg t1) (minus ,lng))
		   (sll (reg  t5) (reg BndstklowerBound) 2)
		   (sub  (reg t5) (reg t2) (reg t5))
		   (bgez (reg t5) ,genlabel)
		   (srl (reg BndstkPtr) (reg t2) 2)
		   (*call Bstackunderflow)
		  ,genlabel  )) %start of code
 
     (setq list (append initload list))
     (setq n 0)
 
 loop
      (setq cfluids (car Fluids))
      (setq n (wplus2 n 8))
 
  % insert reloaded register or memory reference
 
      (setq list (append list
	       (if (car freeregs) `((*move ,(car freeregs) ,cfluids ))
		    `((*move (indexed (Reg t2) ,n) ,cfluids )))

      )          )
      (setq freeregs (cdr freeregs))
      (cond ((setq Fluids (cdr Fluids)) (go loop)))
      (return list)
)    )
(defcmacro *freerstr)
 
(setq *unsafebinder t)   % has to save Registers across calls
 
(global '(*writingasmfile))

(de !*xjumpon (register lowerbound upperbound labellist)
    (for (from x lowerbound upperbound)
	 (join
	  `((*move ,x (reg t2))
	    (beq ,register (reg t2) ,(pop labellist))
	    ))))

(de !*jumpon (register lowerbound upperbound labellist)
   (if *writingasmfile
       (*xjumpon register lowerbound upperbound labellist)
     (PROG (X LL LL2)
       (setq ll  (gensym)) 
       (setq ll2 (stringgensym)) 
       (SETQ X
	(if (and (weq lowerbound 0) (weq upperbound 31)) 
		                 % jumpon on tags (most probably) 
      `((sll (reg t5) ,register 3)  % 8 bytes per jumptable entry
	(move-relocatable-object ,ll2 (reg t1))
	(*wplus2 (reg t1) (reg t5))
	(jr (reg t1))
	(nop)
       ,ll2) 
      `((addi  (reg t1) ,register ,(minus upperbound))
	(bgtz  (reg t1) (label ,ll))
	(addi  (reg t5) ,register ,(minus Lowerbound))
	(bltz  (reg t5) (label ,ll))
	(sll (reg t5) (reg t5) 3)  % 8 bytes per jumptable entry
	(move-relocatable-object ,ll2 (reg t1))
	(*wplus2 (reg t1) (reg t5))
	(jr (reg t1))
	(nop) 
       ,ll2) ))
      Loop  (Setq x (nconc X `((j ,(car Labellist)) ,(list 'nop))) )
	    (setq Labellist (cdr Labellist))
	    (cond (Labellist (go loop)))
  
	    (setq x (nconc x `((*lbl (Label ,ll))) )) 
	    (return x) 
)  )  )

(defcmacro !*jumpon)

% Call foreign fns (loaded by oload) by putting SymFunc cell addr in a reg,
% then doing "calls" instr indirect via the target address of the jmp 
% instruction stored there.  Use old direct link for asm (kernel) compile.

(de *foreignlink (functionname functiontype numberofarguments)

% Following mips Architecture Manual, the Parameter past the forth have
% to be passed in stack. We therefore need some extra space, we get it
% by doing an extra alloc.
% We have to allow the foreignfunction to put their arguments in stack too.
  
 (if (wgreaterp numberofarguments 4)
     `((*alloc 1)            % This is a rare case (hopefully)
       (addiu (reg st) (reg st) -100)
       (*move (reg st) (reg t7)) 
       (*wshift (reg st) -3)  % adjust on double word boundary 
       (*wshift (reg st) 3) 
       (*move (reg t7) (indexed (reg st) 80)) 
       (*move (reg 15) (reg t5)) (sw (reg t5) (indexed (reg st) 56))
       (*move (reg 14) (reg t5)) (sw (reg t5) (indexed (reg st) 52))
       (*move (reg 13) (reg t5)) (sw (reg t5) (indexed (reg st) 48))
       (*move (reg 12) (reg t5)) (sw (reg t5) (indexed (reg st) 44))
       (*move (reg 11) (reg t5)) (sw (reg t5) (indexed (reg st) 40))
       (*move (reg 10) (reg t5)) (sw (reg t5) (indexed (reg st) 36))
       (*move (reg 9)  (reg t5)) (sw (reg t5) (indexed (reg st) 32))
       (*move (reg 8)  (reg t5)) (sw (reg t5) (indexed (reg st) 28))
       (*move (reg 7)  (reg t5)) (sw (reg t5) (indexed (reg st) 24))
       (*move (reg 6)  (reg t5)) (sw (reg t5) (indexed (reg st) 20))
       (sw (reg 5) (indexed (reg st) 16))
       (sw (reg 4) (indexed (reg st) 12))
       (sw (reg 3) (indexed (reg st) 8))
       (sw (reg 2) (indexed (reg st) 4))
       (*move (reg retval) ($fluid *kernelmode))
       (sw (reg 1) (indexed (reg st) 0))
       ,(if *writingasmfile `(jal (foreignentry ,functionname))
                            `(*call ,functionname))
       (nop)
       (*move (reg retval) (reg 1))
       (*move (indexed (reg st) 80) (reg st)) 
       (nop)
       (addiu (reg st) (reg st) 100)
       (*movex ($fluid symval) (reg retval))
       (sll (reg retval) (reg retval) 2)
       (lw (reg retval) (displacement (reg retval) 464))
		            % restore value of reg t
       (*move (reg NIL)  ($fluid *kernelmode))
       (*dealloc 1))

      `((addiu (reg st) (reg st) -100)
        (*move (reg st) (reg t7))
        (*wshift (reg st) -3)  % adjust on double word boundary
        (*wshift (reg st) 3)
        (*move (reg t7) (indexed (reg st) 80))
        (*move (reg r28) (indexed (reg st) 84)) % global pointer
        (*move (reg retval) ($fluid *kernelmode))
       ,(if *writingasmfile `(jal (foreignentry ,functionname))
                            `(*call ,functionname)) 
        (nop)
        (*move (indexed (reg st) 84) (reg r28)) % global pointer
        (*move (indexed (reg st) 80) (reg st))
	(addiu (reg st) (reg st) 100)
	(*move (reg retval) (reg 1))   %store return value in r1
	(*movex ($fluid symval) (reg retval)) %restore value of reg t
	(sll (reg retval) (reg retval) 2) 
	(lw (reg retval) (displacement (reg retval) 464)) % that is (fluid t)
        (*move (reg NIL)  ($fluid *kernelmode))
)))


(defcmacro !*foreignlink)

(defcmacro *movex)
   % Movex is of type cmacro

(de *movex (src dst)
   (let (tmp)
     (cond ((mips-register-p dst)
	      (cond ((mips-register-p src)  `((or ,dst ,src (reg 0))))
		    ((eqcar src 'idloc)     `((addi ,dst (reg zero) ,src)))
                    ((eqcar src 'quote)
                     (if *writingasmfile           %% asmout and lap generate
                         `((lw ,dst ,src))       %% different code.
                         `((move-relocatable-object ,src ,dst))))
		    ((mips-absolute-p src)
		     (if *writingasmfile           %% asmout and lap generate
		         `((lw ,dst ,src))         %% different code.
		         `((move-relocatable-object-and-load ,dst ,src)
		           (nop))))
		    ((mips-fluid-p src) 
		        (if *writingasmfile
		         `((lw  ,dst ,src ))
		         `((lw ,dst (valueaddress ,src))
		           (nop))))
		    ((mips-short-immediate-p src) `((addi ,dst (reg 0) ,src)))
		    ((or (mips-relocatable-immediate-p src)
		         (mips-foreignentry-p src))
		     (if *writingasmfile           %% asmout and lap generate
		         `((la ,dst ,src))         %% different code.
		         `((move-relocatable-object ,src ,dst)
		           (nop))))
		    ((mips-immediate-p src)
		        (if *writingasmfile
		         `((li ,dst ,src))
		         `((lui ,dst (hi ,src) )
		           (ori ,dst ,dst (lo ,src)))))
		    ((mips-huge-displacement-p src)
		         `((lui ,dst (hi ,(caddr src)))
		           (ori ,dst ,dst (lo ,(caddr src)))
		           (add ,dst ,dst ,(cadr src))
		           (lw ,dst (displacement ,dst 0))
		           (nop)))
		    ((mips-indirect-p src) `((lw ,dst ,src) (nop))) 
		    (t  (error 0
		     (list "Unknown addressing mode for *movex:" src dst)))))
	    ((mips-register-p src)
	      (cond
		  ((or (mips-absolute-p dst) (mips-foreignentry-p src))
		   (if *writingasmfile `((sw  ,src ,dst))
		    `((move-relocatable-object-and-store ,dst (reg t6) ,src))))
		  ((and (mips-fluid-p dst) *writingasmfile)
		         `((sw  ,src ,dst)))
		  ((mips-fluid-p dst) `((sw ,src (valueaddress ,dst ))))
		  ((mips-huge-displacement-p dst)
		     (if *writingasmfile
		      `((la (reg t6) ,(caddr dst))
		        (add (reg t6) (reg t6) ,(cadr dst))
		        (sw ,src (displacement (reg t6) 0)))
		      `((lui (reg t6) (hi ,(caddr dst)))
		        (ori (reg t6) (reg t6) (lo ,(caddr dst)))
		        (add (reg t6) (reg t6) ,(cadr dst))
		        (sw ,src (displacement (reg t6) 0)))))
		  ((mips-indirect-p dst) `((sw ,src ,dst)))
		  (t (error 0 (list "Unknown addressing mode:" src dst)))))
	   (t (error 0 (list " Unsaeglich: Move " src dst))
))))

 % Basic Predicates

(de mips-immediate-p(r)
  (cond ((fixp r) t)
	((stringp r) t)
	((idp r)  (cond ((get r '*register-code) nil)
		        ((get r '*f-register-code) nil)
		        ((get r '*g-register-code) nil)
		         (t     t)))  %% label
	(t nil)))

(de mips-relocatable-immediate-p (r)  (eqcar r 'IMMEDIATE))

(de mips-short-immediate-p (r)
	  (and (fixp r) (greaterp r -32767) (lessp r 32767)))

(de mips-short-positive-immediate-p (r)
          (and (posintp r) (lessp r 32767)))

(de mips-huge-displacement-p (r)
	   (and (eqcar r 'displacement)
		(numberp (caddr r))
		(or (greaterp (caddr r) 32767) (lessp (caddr r) -32768))))

(de mips-deferred-p (r)    (eqcar r 'deferred))

(de mips-indirect-p (r)
  (and (pairp r)
       (or (mips-register-p (car r))
	   (eq (car r) 'deferred)
	   (eq (car r) 'displacement))))

(de mips-double-indirect-p (r)
   (and (eqcar r 'deferred) (mips-indirect-p (cadr r))))

(de mips-fluid-p (r)
	  (and (pairp r) (member (car r)  '($fluid $global fluid global))))

(de mips-absolute-p (r)
	  (or (eqcar r 'quote) (stringp r)))

(de mips-foreignentry-p (r)
	  (eqcar r 'foreignentry))

(de *LinkE (DeAllocCount FunctionName FunctionType NumberOfArguments)
  (if (FlagP FunctionName 'ForeignFunction)
      (list (list '*ForeignLink FunctionName FunctionType NumberOfArguments)
	    '(*Exit 0))
      (list (list '*DeAlloc DeAllocCount)
	    (list '*JCall FunctionName))))
