%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXNK:TRAP.SL
% Description:  Signal handling for Sparc
% Author:       Vicki O'Day, HP Labs/CRC
% Created:      27-Feb-84
% Modified:     2-Jan-85 13:13:16 (Vicki O'Day)
% Mode:         Lisp
% Package:
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1984, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 5-Nov-90 (Winfried Neun) sparc-sigcp was cleared too late. Fixed for gc
% 5-Mar-90  (Winfried Neun)
%  added and refined (x-)code-address-to-symbol
% 18-Aug-86 (Leigh Stoller)
%  Reworked so that signals can be added more easily. Also added calls to
%  external functions which set up the trap, and reinit the trap after a
%  signal is received.
% 2-Jan-85 10:21:47 (Vicki O'Day)
%  Instead of just calling stderror with a message about the kind of trap,
%  we now find the name of the routine that caused the trap and report
%  that as well.  CODE-ADDRESS-TO-SYMBOL is used to figure out the name.
% 27-Dec-84 10:55:56 (Vicki O'Day)
%  Added sigunwind, to pop the signal handler's stack frame before calling
%  stderror.  (Many thanks to Jim Davis for deciphering the contents of this
%  frame.)
% 17-Dec-84 10:24:41 (Vicki O'Day)
%  Added lots more signals to catch, to try to prevent those occasional
%  dumps out of Nmode.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime (flag '(signal-done build-trap-message) 'internalfunction))

(fluid '(*bruch* sparc-sigcp sparc-sigaddr sparc-sigcode sparc-signumber
         *dump sparc-sigpc sparc-dumparray*))

% this is the interface for toploop

(de initializeinterrupts ()
          % setting float interrupts first, as described in
          % ieee_handler (3M)

	  % all signals are initiated expect for those in the following list
          % they are ignored

    (for (from i 1 31) (do
       (when (not (memq i '(1 15 18 19 20 21 22 28))) (sigsetup i)))))

% this installs the handler

(setq sparc-dumparray* (gtwarray 50))

(de sigsetup (signumber)
    (psl_sigset signumber (symfnc (id2int 'signalhandler))))

(setq *bruch* 0)

(de signalhandler (signumber sigcode sigcp)
     (prog (handel i)
	(when (and (eq signumber 2) (not (eq *bruch* 0)))
		(setq *bruch* -17)
		(return 0)) % a weaponed interrupt
         (setq handel (symfnc (id2int 'lisprpv)))
         (setq sparc-signumber signumber)
	 (setq sparc-sigcode sigcode)
	 (setq sparc-sigpc (wgetv sigcp 2))
         (setq i 0)
	 (while (wlessp i 32)
		(wputv sparc-dumparray* i (wgetv sigcp (wplus2 i 5)))
		(setq i (iadd1 i)))
         (wputv sigcp 2 handel)  % the pc
         (wputv sigcp 31 handel) % reg 27
)    )            %        'Brutal' version like Cray X-MP for UNICOS

(de lisprpv()
   (prog (mess)
    (setq *bruch* 0)
    (smalldump)
    (setq mess (atsoc sparc-signumber '((1 . "Hangup")
                          (2 . "Interrupt")
                          (3 . "Quit")
                          (4 . "Illegal Instruction")
                          (5 . "Trace Trap")
                          (6 . "IOT Instruction")
                          (7 . "EMT Instruction")
                          (8 . "Floating Point Exception")
                          (10 . "Bus Error")
			  (11 . "Segmentation Violation")
                          (12 . "Bad Args to System Call")
                          (13 . "Write on Pipe With None to Read")
                          (14 . "Alarm Clock")
                          (15 . "Software termination signal")  )))
    (when (eq sparc-signumber 8)
          (cond ((eq sparc-sigcode 4)
                         (setq mess '( 0 . "Floating Point divide by zero")))
                ((eq sparc-sigcode 8)
                         (setq mess '( 0 . "Floating Point overflow")))
                ((eq sparc-sigcode 16#cc)
                         (setq mess '( 0 . "Floating Point underflow")))))
    (setq sparc-sigaddr 0 sparc-sigcp 0) % for gc
    (if (null sparc-signumber) (setq mess  "Interrupt")
    (setq mess (if mess (cdr mess)
                  (bldmsg " Unknown signal type %d"  sparc-signumber))))
    (setq sparc-signumber nil)
    (build-trap-message mess sparc-sigpc)
    (setq sparc-sigaddr 0 sparc-sigcp 0) % for gc
    (return (stderror))))

(de smalldump()
  (when *dump
    (console-print-string "*** UNEXPECTED INTERRUPT ***")
    (console-newline)
    (console-print-string "*** System signal number: 16#")
    (unixputn sparc-signumber)
    (console-print-string " , signal code: 16#")
    (unixputn sparc-sigcode)
    (console-newline)
    (console-print-string "*** signal pc: 16#")
    (unixputn sparc-sigpc)
    (console-print-string " , signal Address: 16#")
    (unixputn sparc-sigaddr)
    (console-print-string " , context: 16#")
    (unixputn sparc-sigcp)
    (console-newline)
    (setq sparc-sigaddr 0 sparc-sigcp 0) % for gc
    (console-print-string "** reg  0-7 :")
    (unixputn 0) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 0)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 1)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 2)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 3)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 4)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 5)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 6)) (console-print-string " ")
     (console-newline)
    (console-print-string "** reg  8-15:")
    (unixputn (wgetv sparc-dumparray* 7)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 8)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 9)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 10)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 11)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 12)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 13)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 14)) (console-print-string " ")
     (console-newline)
    (console-print-string "** reg 16-23:")
    (unixputn (wgetv sparc-dumparray* 15)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 16)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 17)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 18)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 19)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 20)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 21)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 22)) (console-print-string " ")
     (console-newline)
    (console-print-string "** reg 24-31:")
    (unixputn (wgetv sparc-dumparray* 23)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 24)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 25)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 26)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 27)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 28)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 29)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 30)) (console-print-string " ")
     (console-newline)
))

(compiletime (flag '(x-code-address-to-symbol) 'internalfunction))

(de build-trap-message (trap-type trap-addr)
    (let (extra-info)
      (setf extra-info (bldmsg "%w%w" " in "
                                 (x-code-address-to-symbol (inf trap-addr))))
      (bldmsg "%w%w" trap-type extra-info)))

(fluid '(code-address* closest-address* closest-symbol*))

(de x-code-address-to-symbol (code-address*)
  (let ((closest-symbol* ()) (closest-address* 0))
        (mapobl #'(lambda (symbol)
                 (when (fcodep symbol)
                       (let ((address (inf (getfcodepointer symbol))))
                            (when (and (ileq address code-address*)
                                       (igreaterp address closest-address*))
                                  (setq closest-address* address)
                                  (setq closest-symbol* symbol))))))
       closest-symbol*))

(de code-address-to-symbol (ad) (x-code-address-to-symbol ad))
