%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXNK:PRINTF.SL                                              
% Description:  Formatted print routine                                    
% Author:       Eric Benson                                                
% Created:      27 August 1981                                             
% Modified:     29-Oct-84 09:10:42 (Vicki O'Day)
% Mode:         Lisp                                                       
% Package:                                                                 
% Status:       Experimental (Do Not Distribute)                           
%
% (c) Copyright 1984, Hewlett-Packard Company, all rights reserved.        
% (c) Copyright 1981, University of Utah                                   
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:                                                               
%
% 02-Mar-84 09:20:52 (Jim Ambras/CRC)                                      
%  Corrected file header.                                                  
% 24-Feb-84 15:42:09 (Nancy Kendzierski)                                   
%  Added "format" arguments to errorprintf and bldmsg, since they were     
%  supposed to be able to handle as manya arguments as PSL could pass ...  
% 06-Jan-84 17:10:01 (Tim Tillson)                                         
%  Un-rlisp'ed Sam's D-register model version                              
% 4-Nov-83 Sam Sands                                                       
%  Removed 15th parameter for D register model                             
% 17-Sep-82 16:01:01, BENSON                                               
%  Added ChannelPrintF                                                     
% 3-May-82 10:45:11, BENSON                                                
%  %L prints nothing for NIL                                               
% 23-Feb-82 21:40:31, BENSON                                               
%  Added %x for hex                                                        
% 1-Dec-81 16:11:11, BENSON                                                
%  Changed to cause error for unknown character                            
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime (load io-decls token-decls))

(compiletime
  (flag '(printf1 printf2) 'internalfunction))

(fluid '(formatforprintf* in* out*))

% First, lambda-bind FormatForPrintF!*                                     
% That's a no-no in the multiple environment model, since lambda
% binding causes heap grabbage, hence the dynamic extent trick with
% the so-called fluid
(de hugo (formatforprintf a1 a2 a3 a4 a5 a6 a7 a8 a9 a10 a11 a12 a13)
  % scs
  (let ((saveformatforprintf* formatforprintf*))
    (setq formatforprintf* formatforprintf)
    (unwind-protect
      (printf1 formatforprintf* a1 a2 a3 a4 a5 a6 a7 a8 a9 a10 a11 a12 a13))
      (setq formatforprintf* saveformatforprintf*)))

% scs                                                                      
% Then, push all the registers on the stack and set up a pointer to them   

(lap '((*entry printf1 expr 15)
       (*alloc 15)
        (*move (reg 2) (frame 1))
        (*move (reg 3) (frame 2))
        (*move (reg 4) (frame 3))
        (*move (reg 5) (frame 4))
        (*move (reg 6) (frame 5))
        (*move (reg 7) (frame 6))
        (*move (reg 8) (frame 7))
        (*move (reg 9) (frame 8))
        (*move (reg 10) (frame 9))
        (*move (reg 11) (frame 10))
        (*move (reg 12) (frame 11))
        (*move (reg 13) (frame 12))
        (*move (reg 14) (frame 13))
        (*move (reg 15) (frame 14))
       (*wplus3 (reg 1) (reg st) -36)
       %  scs
       (*call printf2)
       (*exit 15)))

(compiletime (put 'memo 'opencode 
			'((ldw (displacement (reg sr0) (reg 1) 0) (reg 1)))))

%  scs                                                                     
% Finally, actual printf, with 1 argument, pointer to array of parameters  
(de printf2 (printfargs)
  %. Formatted print                                                       
  %                                                                        
  % Format is a string, either in the heap or not, whose characters will be
  % written on the currently selected output channel.  The exception to this is
                                                                           
  % that when a % is encountered, the following character is interpreted as a
                                                                           
  % format character, to decide how to print one of the other arguments.  The
                                                                           
  % following format characters are currently supported:                   
  %	%b - blanks; take the next argument as integer and print that many 
  %		blanks                                                     
  %	%c - print the next argument as a single character                 
  %	%d - print the next argument as a decimal integer                  
  %     %e - EVALs the next argument for side-effect -- most useful if the
  %            thing EVALed does some printing                             
  %	%f - fresh-line, print end-of-line char if not at beginning of line
  %	%l - same as %w, except lists are printed without top level parens 
  %	%n - print end-of-line character                                   
  %	%o - print the next argument as an octal integer                   
  %	%p - print the next argument as a Lisp item, using Prin1           
  %     %r - print the next argument as a Lisp item, using ErrPrin (`FOO')
  %	%s - print the next argument as a string                           
  %	%t - tab; take the next argument as an integer and                 
  %		print spaces to that column                                
  %	%w - print the next argument as a Lisp item, using Prin2           
  %	%x - print the next argument as a hexidecimal integer              
  %	%% - print a %                                                     
  %                                                                        
  % If the character is not one of these (either upper or lower case), then an
                                                                           
  % error occurs.                                                          
  %                                                                        
  (prog (uplim i ch upch)
        (setq uplim (strlen (strinf formatforprintf*)))
        (setq i 0)
        (while (wleq i uplim)
          (setq ch (strbyt (strinf formatforprintf*) i))
          (if (wneq ch (char !%))
            (writechar ch)
            (prog nil
                  (setq i (wplus2 i 1))
                  (setq ch (strbyt (strinf formatforprintf*) i))
                  (setq upch (if (lowercasechar ch)
                          (raisechar ch)
                          ch))
                  (case upch 
                        (((char b)) 
                         (progn (spaces (memo printfargs))
                                (setq printfargs 
                                 (loc (wgetv printfargs stackdirection)))))
                        (((char c)) 
                         (progn (writechar (memo printfargs))
                                (setq printfargs 
                                 (loc (wgetv printfargs stackdirection)))))
                        (((char d)) 
                         (progn (writesysinteger (memo printfargs) 10)
                                (setq printfargs 
                                 (loc (wgetv printfargs stackdirection)))))
                        (((char e)) 
                         (progn (eval (memo printfargs))
                                (setq printfargs 
                                 (loc (wgetv printfargs stackdirection)))))
                        (((char f)) (when (wgreaterp (posn) 0)
                           (writechar (char eol))))
                        (((char l)) 
                         (progn (prin2l (memo printfargs))
                                (setq printfargs 
                                 (loc (wgetv printfargs stackdirection)))))
                        (((char n)) (writechar (char eol))) 
                        (((char o)) 
                         (progn (writesysinteger (memo printfargs) 8)
                                (setq printfargs 
                                 (loc (wgetv printfargs stackdirection)))))
                        (((char x)) 
                         (progn (writesysinteger (memo printfargs) 16)
                                (setq printfargs 
                                 (loc (wgetv printfargs stackdirection)))))
                        (((char p)) 
                         (progn (prin1 (memo printfargs))
                                (setq printfargs 
                                 (loc (wgetv printfargs stackdirection)))))
                        (((char r)) 
                         (progn (errprin (memo printfargs))
                                (setq printfargs 
                                 (loc (wgetv printfargs stackdirection)))))
                        (((char s)) 
                         (progn (writestring (memo printfargs))
                                (setq printfargs 
                                 (loc (wgetv printfargs stackdirection)))))
                        (((char t)) 
                         (progn (tab (memo printfargs))
                                (setq printfargs 
                                 (loc (wgetv printfargs stackdirection)))))
                        (((char w)) 
                         (progn (prin2 (memo printfargs))
                                (setq printfargs 
                                 (loc (wgetv printfargs stackdirection)))))
                        (((char !%)) (writechar (char !%)))
                        (nil 
                         (stderror 
                          (bldmsg '"Unknown character code for PrintF: %r" 
                           (mkid ch)))))))
          (setq i (wplus2 i 1)))))

