%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXNK:TRAP.SL
% Description:  Signal handling for Sparc
% Author:       Vicki O'Day, HP Labs/CRC
% Created:      27-Feb-84
% Modified:     2-Jan-85 13:13:16 (Vicki O'Day)
% Mode:         Lisp
% Package:
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1984, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 5-Nov-90 (Winfried Neun) sparc-sigcp was cleared too late. Fixed for gc
% 5-Mar-90  (Winfried Neun)
%  added and refined (x-)code-address-to-symbol
% 25-Aug-89 (Winfried Neun)
%  Added call to ieee_handler in initializeinterrupts
% 10-Mar-89 (Winfried Neun)
%  Handling for sparc (SUN4). Complete redesign.
%  Also installed fluid *bruch* to allow gc (etc) to turn off
%  cntrl-c interrupt detection, which destroy system.
% 18-Aug-86 (Leigh Stoller)
%  Reworked so that signals can be added more easily. Also added calls to
%  external functions which set up the trap, and reinit the trap after a
%  signal is received.
% 2-Jan-85 10:21:47 (Vicki O'Day)
%  Instead of just calling stderror with a message about the kind of trap,
%  we now find the name of the routine that caused the trap and report
%  that as well.  CODE-ADDRESS-TO-SYMBOL is used to figure out the name.
% 27-Dec-84 10:55:56 (Vicki O'Day)
%  Added sigunwind, to pop the signal handler's stack frame before calling
%  stderror.  (Many thanks to Jim Davis for deciphering the contents of this
%  frame.)
% 17-Dec-84 10:24:41 (Vicki O'Day)
%  Added lots more signals to catch, to try to prevent those occasional
%  dumps out of Nmode.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime (flag '(signal-done build-trap-message) 'internalfunction))

(fluid '(*bruch* sparc-sigcp sparc-sigaddr sparc-sigcode sparc-signumber
         *dump sparc-sigpc sparc-dumparray*))

% this is the interface for toploop

(de initializeinterrupts ()
          % setting float interrupts first, as described in
          % ieee_handler (3M)

    (ieee_handler (strbase (strinf "set"))
                  (strbase (strinf "common"))
		  (symfnc (id2int 'signalhandler)))

	  % all signals are initiated expect for those in the following list
          % they are ignored

    (for (from i 1 31) (do
       (when (not (memq i '(1 15 18 19 20 21 22 28))) (sigsetup i)))))

% this installs the handler

(de sigsetup (signumber)
    (psl_sigset signumber (symfnc (id2int 'signalhaendler))))

% this is the 'handler'

(setq sparc-dumparray* (gtwarray 40))

(lap '((*entry signalhaendler expr 0)
%       We save the information which is in 'scratch' registers.
%
         (*move (reg o0) (reg o4)) % will not be executed
         (*move (reg o0) (fluid sparc-signumber))
         (*move (reg o3) (fluid sparc-sigaddr))
         (*move (reg o1) (fluid sparc-sigcode))
         (*move (reg o2) (fluid sparc-sigcp))   % save signal context
         (*move (indexed (reg o2) 16) (fluid sparc-sigpc))

% No alloc, trap handler will never come back!

         (*jumpnoteq (label templab) (reg o0) (quote 2)) % interrupt
         (*jumpeq    (label templab) ($fluid *bruch*) 0)
%
% Now we have a 'weaponed' interrupt. This is used e.g. by gc to remain
% operable. Return Immediately and mark with a negative value in *bruch* !!

%
         (*move -17 ($fluid *bruch*))
         (*move 0 (fluid sparc-sigaddr))
         (*move 0 (fluid sparc-sigcp))
         (jmpl (indexed (reg o7) 8) (reg g0))
         (or (reg g0) 0 (reg g0))

    templab
         (*move ($fluid sparc-dumparray*) (reg o2))
         (*move (reg 2) (indexed (reg o2) 4))
         (*move (reg 3) (indexed (reg o2) 8))
         (*move (reg 4) (indexed (reg o2) 12))
         (*move (reg 5) (indexed (reg o2) 16))
         (*move (reg g6) (indexed (reg o2) 20))
         (*move (reg g7) (indexed (reg o2) 24))
         (*move (reg i0) (indexed (reg o2) 28))
         (*move (reg i1) (indexed (reg o2) 32))
         (*move (reg i2) (indexed (reg o2) 36))
         (*move (reg i3) (indexed (reg o2) 40))
         (*move (reg i4) (indexed (reg o2) 44))
         (*move (reg i5) (indexed (reg o2) 48))
         (*move (reg i6) (indexed (reg o2) 52))
         (*move (reg i7) (indexed (reg o2) 56))

         (*move ($fluid sparc-sigcp) (reg o3))
         (*move (indexed (reg o3) 24) (reg o1))  %the g1
         (*move (reg o1) (indexed (reg o2) 0))   % moving the
         (*move (indexed (reg o3) 8) (reg o3))   % l's and the i's
         (*move (indexed (reg o3) 0) (reg o1))
         (*move (reg o1) (indexed (reg o2) 60))
         (*move (indexed (reg o3) 4) (reg o1))
         (*move (reg o1) (indexed (reg o2) 64))
         (*move (indexed (reg o3) 8) (reg o1))
         (*move (reg o1) (indexed (reg o2) 68))
         (*move (indexed (reg o3) 12) (reg o1))
         (*move (reg o1) (indexed (reg o2) 72))
         (*move (indexed (reg o3) 16) (reg o1))
         (*move (reg o1) (indexed (reg o2) 76))
         (*move (indexed (reg o3) 20) (reg o1))
         (*move (reg o1) (indexed (reg o2) 80))
         (*move (indexed (reg o3) 24) (reg o1))
         (*move (reg o1) (indexed (reg o2) 84))
         (*move (indexed (reg o3) 28) (reg o1))
         (*move (reg o1) (indexed (reg o2) 88))
         (*move (indexed (reg o3) 32) (reg o1))
         (*move (reg o1) (indexed (reg o2) 92))
         (*move (indexed (reg o3) 36) (reg o1))
         (*move (reg o1) (indexed (reg o2) 96))
         (*move (indexed (reg o3) 40) (reg o1))
         (*move (reg o1) (indexed (reg o2) 100))
         (*move (indexed (reg o3) 44) (reg o1))
         (*move (reg o1) (indexed (reg o2) 104))
         (*move (indexed (reg o3) 48) (reg o1))
         (*move (reg o1) (indexed (reg o2) 108))
         (*move (indexed (reg o3) 52) (reg o1))
         (*move (reg o1) (indexed (reg o2) 112))
         (*move (indexed (reg o3) 56) (reg o1))
         (*move (reg o1) (indexed (reg o2) 116))
         (*move (indexed (reg o3) 60) (reg o1))
         (*move (reg o1) (indexed (reg o2) 120))

         (*jcall signal-done)
))

(setq *bruch* 0)

(de signal-done ()
     (prog (handel)
         (setq handel (symfnc (id2int 'lisprpv)))
         (wputv sparc-sigcp 3 handel)
         (wputv sparc-sigcp 4 (wplus2 handel 4)) % Overwrite pc and npc
)    )            %        'Brutal' version like Cray X-MP for UNICOS

(de lisprpv()
   (prog (mess)
    (setq *bruch* 0)
    (smalldump)
    (setq mess (atsoc sparc-signumber '((1 . "Hangup")
                          (2 . "Interrupt")
                          (3 . "Quit")
                          (4 . "Illegal Instruction")
                          (5 . "Trace Trap")
                          (6 . "IOT Instruction")
                          (7 . "EMT Instruction")
                          (8 . "Floating Point Exception")
                          (10 . "Bus Error")
			  (11 . "Segmentation Violation")
                          (12 . "Bad Args to System Call")
                          (13 . "Write on Pipe With None to Read")
                          (14 . "Alarm Clock")
                          (15 . "Software termination signal")  )))
    (when (eq sparc-signumber 8)
          (cond ((eq sparc-sigcode 16#c8)
                         (setq mess '( 0 . "Floating Point divide by zero")))
                ((eq sparc-sigcode 16#d4)
                         (setq mess '( 0 . "Floating Point overflow")))
                ((eq sparc-sigcode 16#cc)
                         (setq mess '( 0 . "Floating Point underflow")))))
    (setq sparc-sigaddr 0 sparc-sigcp 0) % for gc
    (setq mess (if mess (cdr mess)
                  (bldmsg " Unknown signal type %d"  sparc-signumber)))
    (build-trap-message mess sparc-sigpc)
    (setq sparc-sigaddr 0 sparc-sigcp 0) % for gc
    (return (stderror))))

(de smalldump()
  (when *dump
    (console-print-string "*** UNEXPECTED INTERRUPT ***")
    (console-newline)
    (console-print-string "*** System signal number: 16#")
    (unixputn sparc-signumber)
    (console-print-string " , signal code: 16#")
    (unixputn sparc-sigcode)
    (console-newline)
    (console-print-string "*** signal pc: 16#")
    (unixputn sparc-sigpc)
    (console-print-string " , signal Address: 16#")
    (unixputn sparc-sigaddr)
    (console-print-string " , context: 16#")
    (unixputn sparc-sigcp)
    (console-newline)
    (setq sparc-sigaddr 0 sparc-sigcp 0) % for gc
    (console-print-string "*** G registers :")
    (unixputn 0) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 0)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 1)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 2)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 3)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 4)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 5)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 6)) (console-print-string " ")
     (console-newline)
    (console-print-string "*** O registers :")
    (unixputn (wgetv sparc-dumparray* 7)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 8)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 9)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 10)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 11)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 12)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 13)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 14)) (console-print-string " ")
     (console-newline)
    (console-print-string "*** L registers :")
    (unixputn (wgetv sparc-dumparray* 15)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 16)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 17)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 18)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 19)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 20)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 21)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 22)) (console-print-string " ")
     (console-newline)
    (console-print-string "*** I registers :")
    (unixputn (wgetv sparc-dumparray* 23)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 24)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 25)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 26)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 27)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 28)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 29)) (console-print-string " ")
    (unixputn (wgetv sparc-dumparray* 30)) (console-print-string " ")
     (console-newline)
))

(compiletime (flag '(x-code-address-to-symbol) 'internalfunction))

(de build-trap-message (trap-type trap-addr)
    (let (extra-info)
      (setf extra-info (bldmsg "%w%w" " in "
                                 (x-code-address-to-symbol (inf trap-addr))))
      (bldmsg "%w%w" trap-type extra-info)))

(fluid '(code-address* closest-address* closest-symbol*))

(de x-code-address-to-symbol (code-address*)
  (let ((closest-symbol* ()) (closest-address* 0) numb)
    (if (wgreaterp code-address* symfnc)
       (progn
        (setq numb (wquotient (wdifference code-address* symfnc)
                              addressingunitsperfunctioncell))
        (setq closest-symbol*
           (bldmsg "calling %w" (cond ((wlessp numb maxsymbols) (int2id numb))
                                      (t nil)))))
        (mapobl #'(lambda (symbol)
                 (when (fcodep symbol)
                       (let ((address (inf (getfcodepointer symbol))))
                            (when (and (ileq address code-address*)
                                       (igreaterp address closest-address*))
                                  (setq closest-address* address)
                                  (setq closest-symbol* symbol)))))))
       closest-symbol*))

(de code-address-to-symbol (ad) (x-code-address-to-symbol ad))
