%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           PNK:SYMBOL-VALUES.SL
% Title:          ValueCell, UnboundP, MakeUnbound and Set
% Author:         Eric Benson
% Created:        20 August 1981
% Modified:       29-Aug-84 11:02:05 (Brian Beach)
% Status:         Experimental
% Mode:           Lisp
% Package:        Kernel
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% Copyright (c) 1981 Eric Benson
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 29 May 1986 W. Neun  , ZIB Berlin
%  Added features to read and write fluids that are stored in registers
%  this code is automatically generated from registernames
% 26 Jan 1984 1001-PST (Brian Beach)
%  Corrected file name in header.
% 01-Dec-83 15:02:42 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(compiletime (load vector-instructions))
 
(fluid '(fluid-reg-vector))
 
(de unboundp (u)
  % Does U not have a value?
  (if (idp u)
    (if (weq (tag (symval (idinf u))) unbound-tag)
      t
      nil)
    (noniderror u 'unboundp)))
 
(de makeunbound (u)
  % Make U an unbound ID
  (if (idp u)
    (setf (symval (idinf u)) (mkitem unbound-tag (idinf u)))
    (noniderror u 'makeunbound)))
 
%
% following macros generates code to do the id-to-reg conversion
% for eval.
% with  valuecell  all fluids are shown, but
% with  set        only those fluids can be set, that are flagged
%                  *user-variable
 
(compiletime
 (progn
 
(fluid '(collectedregs*))
 
(de collectregs-read (x)
     (when (and (get x 'registername)
                (not (memq x '(t1 t2 t3 t4 t5 t6 t7))))
        (setq collectedregs*
         (cons `((eq u (quote ,x)) ,x) collectedregs*))))
 
(dm idtoreg-conversion-read ()
       (setq collectedregs* nil)
       (mapobl (function collectregs-read))
       (cons 'cond (nconc collectedregs* `((t v))  )))
 
(de collectregs-write (x)
     (when (and (get x 'registername) (flagp x '*user-variable))
        (setq collectedregs*
         (cons `((eq exp (quote ,x)) (setq ,x val)) collectedregs*))))
 
(dm idtoreg-conversion-write ()
       (setq collectedregs* nil)
       (mapobl (function collectregs-write))
       (cons 'cond (nconc collectedregs*
                          `((t (setf (symval (idinf exp)) val))) )))
 
(dm build-fluid-reg-vector ()
       (setq collectedregs* nil)
       (mapobl (function (lambda (x)
                  (when (and (get x 'registername)
                             (not (memq x '(t1 t2 t3 t4 t5 t6 t7))))
                   (setq collectedregs* (cons x collectedregs*))))))
       (list 'quote collectedregs*))
 
(dm build-lng-fluid-reg-vector ()
       (setq collectedregs* 0)
       (mapobl (function (lambda (x)
                  (when (and (get x 'registername)
                             (not (memq x '(t1 t2 t3 t4 t5 t6 t7))))
                   (setq collectedregs* (plus2 collectedregs* 1))))))
       collectedregs*)
 
(define-constant lng-of-fluid-reg-vector (build-lng-fluid-reg-vector))
)) %end compiletime
 
(setq fluid-reg-vector (build-fluid-reg-vector))
 
(de valuecell (u)
  % Safe access to SymVal entry
  (prog (v)
        % This guy is called from Eval
        (return (if (idp u)
                       (progn
                         (setq v (symval (idinf u)))
                         (if (weq (tag v) unbound-tag)
                            (progn (VsetVl lng-of-fluid-reg-vector)
                                (Vload (Vreg 1)
                                       (wplus2 fluid-reg-vector 1) 1)
                                (VSminusV (Vreg 2) u (Vreg 1))
                                (Setq v (VtestZ (Vreg 2)))
                                (if (weq v 0)
                                  (setq v (continuableerror '99
                                  (bldmsg '"%r is an unbound ID" u) u))
                                  (setq v (idtoreg-conversion-read))
                         )  )   )
                         v)
                  (noniderror u 'valuecell)))))
 
% This version of SET differs from the Standard Lisp report in that Exp is
% not declared fluid, in order to maintain compatibility between compiled
% and interpreted code.
(de set (exp val)
  % Assign Val to ID Exp
  (if (idp exp)
    (if (not (or (null exp) (weq exp 't)))
      (progn (idtoreg-conversion-write)
             val)
      (stderror '"T and NIL cannot be SET"))
    (noniderror exp 'set)))
