%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           PNK:STRING-GENSYM.SL
% Title:          Complement to GenSym, makes a string instead of ID
% Author:         Eric Benson
% Created:        14 January 1982
% Modified:       29-Aug-84 11:03:46 (Brian Beach)
% Status:         Experimental
% Mode:           Lisp
% Package:        Kernel
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1982, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 19-Dec-86 (Leigh Stoller)
%  Added copystring to setq of stringgensym* below to make sure the string
%  in allocated in heap, not bps.
% 26 Jan 1984 1001-PST (Brian Beach)
%  Corrected file name in header.
% 01-Dec-83 15:02:12 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
% Edit by Cris Perdue,  9 Feb 1983 1620-PST
% Modified to avoid using the CHAR macro in a top level form
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(fluid '(stringgensym*))

% Make sure that the string ends up in heap, not bps, so we can unexec over it.
(setq stringgensym* (copystring (symnam (inf 'L0000))))

% Copy to force into heap /csp
(compiletime
  (flag '(stringgensym1) 'internalfunction))

(de stringgensym ()
  %. Generate unique string
  (stringgensym1 4))

(de stringgensym1 (n)
  %. Auxiliary function for StringGenSym
  (let (ch)
       (cond ((greaterp n 0) (if (lessp (setq ch (indx stringgensym* n))
                                  (char !9))
                (progn (setindx stringgensym* n (plus ch 1))
                       (totalcopy stringgensym*))
                (progn (setindx stringgensym* n (char !0))
                       (stringgensym1 (difference n 1)))))
             (t % Increment starting letter

                (progn (setindx stringgensym* 0
                        (plus (indx stringgensym* 0) 1))
                       (stringgensym))))))

