%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXC/MULS32.SL
% Description:  support for BIGNUM package with double INUM operations
%               lap source for SUN SPARC processor
% Author:       H. Melenk
% Created:      11 Sept  1989
% Modified:
% Mode:         Lisp
% Status:       Experimental
%
% (c) Copyright 1989, Konrad-Zuse-Zentrum, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Double length multiply and divide for the  processor.
% These are the open codes which embed the extended arithmetic 
% operations such that they are accessible from compiled LISP top level.
%
% Full 32 bit version : 28-June 1993 (H. Melenk)

(compiletime (flag '(wquotientdouble*) 'internalfunction))

(fluid '(*second-value* carry* doublearray*))
 
%----------------- unsigned add and subtract operations -------------------

% add and set carry
%
%   (a + b) -> (carry*,result)

(bothtimes
(put 'addAndSetCarry 'opencode
       '(
         (ADDcc (reg 2)(reg 1)(reg 1))
           % move cf to carry*
         (ADDX (reg 0)(reg 0)(reg t1))
         (*MOVE (reg t1) ($FLUID carry*))
       ))
)


% add with carry
%
%   (a + b + carry*) -> (carry*,result)

(put 'addwithcarry 'opencode
       '(
           % move carry* to register CF
         (*MOVE ($fluid carry*)(reg t1))
         (SUBcc (reg 0)(reg t1)(reg 0))
           % add with carry
         (ADDXcc (reg 2)(reg 1)(reg 1))
           % move cf to carry*
         (ADDX (reg 0)(reg 0)(reg t1))
         (*MOVE (reg t1) ($FLUID carry*))
       ))

% add and add carry
%
%   (a + b) -> result,  (*second-value* + carry) -> *second-value*

(put 'addAndAddCarry 'opencode
       '(
         (*move ($FLUID *second-value*) (reg t1))
           % add (reg 2) to (reg 1)
         (ADDcc (reg 2)(reg 1)(reg 1))
           % add carry to reg 3
         (ADDX (reg 0)(reg t1)(reg t1))
         (*move (reg t1) ($FLUID *second-value*))
       ))

% subtract with borrow
%
%  )(a - (b + carry!*)) -> result, carry*=1 if borrow

(bothtimes
(put 'subtractwithborrow 'opencode
       '(
           % move carry* to cf
         (*MOVE ($fluid carry*)(reg t1))
         (SUBcc (reg 0)(reg t1)(reg 0))
           % subtract with borrow
         (SUBXcc (reg 1)(reg 2)(reg 1))
           % move new borrow to carry*
         (ADDX (reg 0)(reg 0)(reg t1))
         (*MOVE (reg t1) ($FLUID carry*))
       ))
)

%------------------- unsigned greaterp ---------------------

% code based on integer carry; 
 
(bothtimes

(put 'ugreaterp* 'opencode 
   % returns 1 if arg1 > arg2 unsigned.
'( 
   (SUBCC (reg 2)(reg 1)(reg 2))      % compare, setting carry if r1>r2
   (ADDX (reg 0)(reg 0)(reg 1))       % move carry to reg 1
))

)

(ds ugreaterp(a b)(eq 1 (ugreaterp* a b)))

%-------------------------- multiply, divide -------------------------

%-------------------------- supersparc -------------------------------

(de supersparc() (copyd 'WxTimes2 'wwWxTimes2)
                 (copyd 'wquotientdouble 'wwwquotientdouble)
		 (copyd 'wdivide 'wwdivide)
                 (copyd 'wtimesdouble 'wwwtimesdouble))

(lap '((*entry wwwtimesdouble expr 2)
     % Called with two parameters of maximum bbase size
     % the return value are the low order bits of the product.
     % The hig order bits are placed in fluid variable.
       (*alloc 0)
       (umul (reg 1) (reg 2) (reg 1))
       (rdy (reg 0) (reg 0) (reg t1))
       (*MOVE (reg t1)($FLUID *second-value*))
       (*exit 0)))

(lap '((*entry wwWxTimes2 expr 2)   % extended times
        (smul (reg 1) (reg 2) (reg 1))
        (rdy (reg 0) (reg 0) (reg 2))
        (*move (reg 2) ($fluid *second-value*))
        (*exit 0)))

(lap '((*entry wwwquotientdouble expr 3)
              % adjusting the words first
        (wry (reg 1) (reg 0) (reg 0))
          % now we can divide and spread the results
        (UDIV (reg 2) (reg 3) (reg 1)) % quotient is in REG 1
        (smul (reg 1) (reg 3) (reg 4))
        (*wdifference (reg 2) (reg 4))
        (*MOVE (reg 2) ($FLUID *second-value*))
        (*exit 0)))

(de wwdivide (x y) 
        (cond ((wgreaterp (tag x) 15) (wwwdivide -1 x y)) % sign
	      (t (wwwdivide 0 x y))))

(lap '((*entry wwwdivide expr 3)
	(wry (reg 1) (reg 0) (reg 0))
        (SDIV (reg 2) (reg 3) (reg 1)) % quotient is in REG 1
        (smul (reg 1) (reg 3) (reg 4))
        (*wdifference (reg 2) (reg 4))
        (*MOVE (reg 2) ($FLUID *second-value*))
        (*exit 0)))


%-------------------------- standard Sparc ---------------------------



%--------------------------------------------------------------------
%  Unsigned divide 64:32 -> (32,32) in software
%--------------------------------------------------------------------

% (copyd 'wtimesdouble 'wuxtimes2)

(lap '((*entry wtimesdouble expr 3)
       (*alloc 2)                   %saving reg 4 and 5
       (*move (reg 4)(frame 1))
       (*move (reg 5)(frame 2))
       (*call wuxtimes2)
       (*move (frame 1) (reg 4))
       (*move (frame 2)(reg 5))
       (*exit 2)))

(commentoutcode

    % pattern for the lap code below

(de wquotientdouble (arg1 arg2 arg3)

   (prog (quot rem locarry hicarry i)

     (setq quot 0)
     (setq rem arg1)

     (ifor (from i 0 31 1) (do

            % now shift the remainder left by 1 and add the most significant
            % bit of arg2. shift arg2 left one bit
            % if sum is greater or equal divisor, add a bit '1' to remainder
            % else add a '0'.
            % do it until arg2 is eaten up. quot is the result (quotient),
            %          rem is the remainder (in *second-value*)

            (progn
             (setq quot (wshift quot 1))
             (setq locarry (wshift arg2 -31))
             (setq arg2 (wshift arg2 1))
             (setq hicarry (wshift rem -31))
             (setq rem (wplus2 (wshift rem 1) locarry))
             (when (or (not (ugreaterp arg3 rem)) 
                       (eq hicarry 1))
                   (setq rem (wdifference rem arg3))
                   (setq quot (wplus2 quot 1)))
      )     )              )
     (setq *second-value* rem)
     (return quot)))

)

(compiletime

(defcmacro *jumpunsignedlessp)

(de *jumpunsignedlessp (Lbl ArgOne ArgTwo)
  (prog (ResultingCode*)
    (return (CMacroPatternExpand
           (list (ResolveOperand '(Reg t1) ArgOne)
                 (ResolveOperand '(Reg t2) ArgTwo)
                 (ResolveOperand '(Reg Error) Lbl)
                 )

        '(( 
            (SUBcc ArgOne ArgTwo (reg 0))  
            (ADDX (reg 0)(reg 0)(reg t6))
            (*jumpeq  ArgThree (reg t6)(quote 1))
         ))))))
)

% (lap '((*entry hugo expr 2)
%        (*jumpunsignedlessp (label lessp)(reg 1)(reg 2))
%        (*move (quote nil) (reg 1)) (*exit 0)
%  lessp (*move (quote T) (reg 1)) (*exit 0) ))

 (lap '((*entry wquotientdouble expr 3)
%  (prog (quot rem locarry hicarry i)
%          t1   t2   t3      t4    5
         %    (setq quot 0)
     (*move (quote 0) (reg t1))
         %    (setq rem arg1)
     (*move (reg 1)(reg t2))
         %    (ifor (from i 0 31 1) (do
     (*move (quote -33) (reg 5))
loop
     (*wplus2 (reg 5)(quote 1))
     (*jumpeq (label finish) (reg 5)(quote 0))
         %  (progn
         %   (setq quot (wshift quot 1))
     (*wshift (reg t1) 1)
         %   (setq locarry (wshift arg2 -31))
     (*move (reg 2)(reg t3))
     (*wshift (reg t3) -31)
         %   (setq arg2 (wshift arg2 1))
     (*wshift (reg 2) 1)
         %   (setq hicarry (wshift rem -31))
     (*move (reg t2) (reg t4))
     (*wshift (reg t4) -31)
         %   (setq rem (wplus2 (wshift rem 1) locarry))
     (*wshift (reg t2) 1)
     (*wplus2 (reg t2) (reg t3))
         %   (when (or (not (ugreaterp arg3 rem))
         %             (eq hicarry 1))
     (*wdifference (reg t4)(quote 1))
     (*jumpeq (label yes)(reg t4)(quote 0))
     (*jumpunsignedlessp (label loop) (reg t2)(reg 3))
  yes
         %         (setq rem (wdifference rem arg3))
     (*wdifference (reg t2)(reg 3))
         %         (setq quot (wplus2 quot 1)))
     (*wplus2 (reg t1) (quote 1))
     (*jump (label loop))
  finish
        %  )     )              )
        %  (setq *second-value* rem)
     (*move (reg t2)($fluid *second-value*))
        % (return quot)))
     (*move (reg t1)(reg 1))
     (*exit 0)
))

