%						Michael Hucka 6/5/87
%
% Quick hack for creating interface functions between PSL and C programs.
% The basic idea is to use define-c-call to create a Lisp function for every
% C function to be used.  The C function must be defined such that its
% parameters are pointers, not actual values.  For example, the following
% is a simple C function accepting as argument a pointer to a double, 
% setting the value of that double to 99.34 and returning the value:
% 
% double *test_float(f)
% double *f;
% {
%   printf("C   : %f\n", *f); 
%   *f = 99.34;
%   return(f);
% }
%
% To make the above callable from PSL, it must first be placed in a C file
% (call it "test.c") then compiled with the following command
%	/com/cc test.c -nmgbl
% loaded into a running PSL with
% 	(aegisload "test")
% and finally, an equivalent Lisp function "test_float" must be created with
% 	(define-c-call test_float 1 f)
%
% The "1" informs define-c-call that test_float takes one argument, f.  If
% we had a function test_float2 which required two arguments "first" and
% "second" we could invoke define-c-call with
%  	(define-c-call test_float2 2 first second)
% 
% Once the function is defined in Lisp it can be called like any other Lisp
% function.  For example, the following shows a sample usage session:
%
% lisp> (load all_syscalls compiler)			% necessary
% lisp> (aegisload "test")
% lisp> (define-c-call test_float 1 f)
% lisp> (setq x 1.0)
% lisp> x
%  1.0
% lisp> (test_float x)
%  C   : 1.000000
%  NIL
% lisp> x
%  99.34			% see code for test_float
%
% One final note:  if a C file is reloaded with aegisload, the function which
% it contains must be redefined with "aegisredefine" as in
%	(aegisredefine 'test_float)
%


(load useful)

(defmacro define-c-call (name number-of-args . args)
  `(de ,name ,args
     ,@(for (from i 1 number-of-args 1)
         (collect (list 'putvalue i (list 'get-var-addr (nth args i)))))
     (syscall ',name ,number-of-args)
     nil))

(de get-var-addr (var)
  (plus2 4 (fltinf var)))
