%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:  PXU/DISASSEMBLE.SL -  disassembler for MC68000
%
% Author: H. Melenk , ZIB Berlin
%
% Date :  5-Apr-1989
%
% Copyright (c)  all rights reserved
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Revisions:
%

          (de getwrd(a)
            (if (eq (wshift (wshift a -2) 2)a)
               (wshift (getmem (wshift (wshift a -2) 2)) -16)
               (wand (getmem (wshift (wshift a -2) 2)) 65535) ))
          (de getfunctionaddress(fkt)
             (wshift(wshift (cdr (getd fkt)) 8) -8))

          (de idnumberp(x)
           (cond ((not (posintp x)) nil)
                 ((greaterp x (lshift maxsymbols 1)) nil)
                 ((stringp (symnam x)) x)
                 (t nil)))
          (de safe!-int2id(x) (if (idnumberp x) (mkid x) (mkid 32)))
          (copyd 'ttab 'tab)


(de word2addr (n) (times n 4))
(de addr2word (n) (quotient n 4))
(de jump2word (n) (quotient n 6))


(compiletime (setq !*syslisp t))

(fluid '(aregs dreg instrs !*comment fktend !*hardjump))

(setq aregs  '("a0" "a1" "a2" "a3" "a4" "a5" "a6" "a7"))
(setq dregs  '("d0" "d1" "d2" "d3" "d4" "d5" "d6" "d7"))


(de decode(p1 p2 p3 addr)
  (prog(i j k lth effa regs code info found)
      (setq i instrs)
      (while (and (not found) i)
           (setq j (car i))
           (when (equal (logand p1 (car j)) (cadr j))
                 (setq found j))
           (setq i (cdr i)))
      (when (not found)(return nil))
      (return (apply (cadr (cddr j)) (list j p1 p2 p3 addr)))))


(de disassemble (fkt)
   (prog(base instr jk jk77 p1 p2 p3 lth pat x
         comment mem jmem symvalhigh symfnchigh frame
         argumentblockhigh labels label bstart bend breg com4 memp1
         !*lower lc)
         (setq !*lower t)

         (cond ((numberp fkt) (setq base fkt))
               ((pairp fkt) (setq base (car fkt))
                            (setq bend (cadr fkt))
                            (plus2 base bend)) %do an arithmetic test
               ((idp fkt)
                      (when (not (getd fkt)) (error 99 "not compiled"))
                      (when (not (codep (cdr (getd fkt))))(return nil))
                      (setq base (getfunctionaddress fkt))
         )     )
         (when (greaterp base nextbps) (return (error 99 "out of range")))
         (setq argumentblockhigh (plus2 argumentblock (word2addr 15)))
         (setq symvalhigh (plus2 symval (word2addr maxsymbols)))
         (setq symfnchigh (plus2 symfnc (word2addr maxsymbols)))
         (terpri)
   %     (putmem nextbps 0)            % safe endcondition
         (setq bstart base)
         (setq fktend nil)
  % first pass: find label references
loop1
         (setq p1 (getwrd base))
         (setq !*hardjump nil)
         (when (eq p1 0)(go continue1))
         (setq p2 (getwrd (plus2 base 2)))
         (setq p3 (getwrd (plus2 base 4)))
         (setq instr (decode p1 p2 p3 base))      % instruction
         (setq lth (atsoc 'LTH instr))
         (setq lth (if lth (cdr lth) 2))
         (setq jmem (atsoc 'addr instr))
         (when jmem (setq jmem (cdr jmem)))
         (cond ((not (assoc jmem labels))
                     (setq labels (cons (list jmem) labels)) ))
 next    (setq base (plus2 base lth))
         (when (and !*hardjump fktend (greaterp base fktend))
               (go continue1))
         (cond ((not bend ) (go loop1))
               ((greaterp base bend) (go continue1))
               (t (go loop1)))
 continue1
  % second pass: assign symbolic labels to jump targets
         (when (not bend) (setq bend base))
         (setq base bstart)
         (setq labels (labelsort (delete '(nil) labels)))
         (mapcar labels
                (function
                  (lambda(x)
                    (cond
                      ((and       % test within-range
                        (geq (car x) bstart)
                        (leq (car x) bend)
                       )(rplacd x (gensym)) )
                      (t (rplaca x nil))
         )      ) ) )
  % third pass: print instructions
         (prinblx (list "function: " fkt " base: " base))
         (terpri)
         (setq lc 0)
loop
         (cond ((assoc base labels)
                (ttab 22) (prin2 (cdr (assoc base labels)))
                (setq lc (add1 lc))
                (prin2t ":")))
         (setq p1 (getwrd base))
         (cond((eq p1 0)(return nil)))
         (setq p2 (getwrd (plus2 base 2)))
         (setq p3 (getwrd (plus2 base 4)))
         (setq !*comment nil)
         (setq instr (decode p1 p2 p3 base))      % instruction
         (setq lth (atsoc 'lth instr)) (setq lth (if lth (cdr lth) 2))
         (cond ((atsoc 'op2 instr)
                (setq pat (list (cdr (atsoc 'op1 instr)) ","
                                (cdr (atsoc 'op2 instr)) )))
               ((atsoc 'op1 instr)
                (setq pat (list (cdr (atsoc 'op1 instr)))))
               (t (setq pat nil)))

         (setq mem (atsoc 'addr instr)) (when mem (setq mem (cdr mem)))
         (setq jmem (assoc mem labels)) (when jmem(setq jmem (cdr jmem)))
      %  (when jmem (setq pat (subst (cdr jmem) mem pat)))
         (when jmem (setq pat (subst jmem mem pat))
                    (setq instr (cons (cons '!<effa!> jmem) instr)))

         (ttab 1)
         (prinbnx base 8)
         (prin2 " ")
         (prinbnx p1 4)   % binary first parcel
         (when (greaterp lth 2) (prin2 " ") (prinbnx p2 4))
         (when (greaterp lth 4) (prin2 " ") (prinbnx p3 4))
         (ttab 28)
         (setq p1 (atsoc 'name instr))
         (when p1 (prin2 (cdr p1)))
         (ttab 38)
         (prinblx (subla instr pat))
         (prin2 "    ")

         (cond
             (!*comment (setq comment !*comment))
             ((null mem) nil)
             ((and (geq mem argumentblock)
                   (leq mem argumentblockhigh))
               (setq comment (list 'extrareg
                    (plus2 (addr2word (difference mem argumentblock)) 6))) )
             ((and (geq mem symval)
                   (leq mem symvalhigh))
                (setq comment
                   (safe!-int2id (addr2word (difference mem symval)))) )
             ((and (geq mem symfnc )
                   (leq mem symfnchigh))
	      (setq comment(safe!-int2id (jump2word (difference mem symfnc)))
	     ))
         )
      %  (when (and (null comment)
      %             (setq mem(idnumberp (wrshift mem 4)))
      %             (greaterp mem 128))
      %             (setq comment (list 'xidloc (int2id mem))))
         (when comment (ttab 55) (prin2 comment))
         (setq comment nil)
         (setq base (plus2 base lth))
         (setq lc (add1 lc))
         (when (leq base bend)(go loop))
)   )


(de prinbl (l)                  % binary (octal) printing of a list
    (if (atom l)(prinb l)
                (mapc l (function prinbl))))

(de prinblx (l)                  % binary (hexa) printing of a list
     (if (atom l)(prinbx l)
                 (mapc l (function prinblx))))



(de prinb (it)                  % binary (octal) printing of an item
     (cond ((numberp it)(prinbo it))
           ((eq it 't1) (ttab 42))
           ((eq it 't2) (ttab 60))
           (t       (prin2 it))))

(de prinbx (it)                  % binary (hexa) printing of an item
     (cond ((numberp it)(prinbox it))
           ((eq it 't1) (ttab 42))
           ((eq it 't2) (ttab 60))
           (t       (prin2 it))))


(de prinbo (it)
      (cond ((lessp it 0) (prin2 "-") (prinbo (minus it)))
            ((geq it 8)  (prin2 "O'") (prinbn it 1))
            (t            (prinbn it 1))))

(de prinbox (it)
      (cond ((lessp it 0) (prin2 "-") (prinbox (minus it)))
            ((geq it 9)  (prin2 "0x") (prinbnx it 1) )
            (t            (prinbnx it 1))))

(de prinbn (it n)                % print an octal number
     (cond ((and (eq it 0) (leq n 0)) nil)
           (t (progn
                (prinbn (lshift it -3) (plus2 n -1))
                (prindig (logand it 7))
)    )     )  )

(de prinbnx (it n)                % print a hexa number
     (cond ((and (eq it 0) (leq n 0)) nil)
           (t (progn
                (prinbnx (quotient it 16) (plus2 n -1))
                (prindigx (logand it 15))
)    )     )  )



(de prindig (dig)        % print a numeric digit
     (writeChar  (plus2 dig 48)))

(fluid '(hexadigits))
(setq hexadigits
     '("0" "1" "2" "3" "4" "5" "6" "7" "8" "9" "a" "b" "c" "d" "e" "f"))
(de prindigx (dig) (prin2 (nth  hexadigits (add1 dig))))

(de labelsort (l)        % sort labels to ascending sequence
    (labelsort1 l nil))

(de labelsort1 (rest sorted)
    (cond ((null rest) sorted)
          (t (labelsort1(cdr rest) (labelsortin (car rest) sorted))) ))

(de labelsortin (object l)
    (cond ((null l)(list object))
          ((greaterp (car object)(caar l))
           (cons (car l) (labelsortin object (cdr l))) )
          (t (cons object l)) ))


%%---------------------------------------------
(fluid '(instrs instrfields ilength jaddr fktend dregs aregs !*comment
         !*hardjump))

(dm instr(u) (list 'instr1 (mkquote (cdr u))))

(de instr1 (u)
     (prog(form mask1 mask2 n cf x l name fn)
            % mask1 : mask out relevant bits
            % mask2 : instruction pattern
       (setq name (car u))
       (setq form (cadr u))
       (setq fn (caddr u))
       (setq mask1 0)(setq mask2 0) (setq n 16)
       (while form
          (setq cf (car form))(setq form (cdr form))
          (cond
            ((equal cf 0)
             (setq n (sub1 n))
             (setq mask1 (add1 (lshift mask1 1)))
             (setq mask2  (lshift mask2 1)))
            ((equal cf 1)
             (setq n (sub1 n))
             (setq mask1 (add1 (lshift mask1 1)))
             (setq mask2 (add1 (lshift mask2 1))))
            ((setq x (assoc cf instrfields))
             (setq l (cdr x))
             (setq n (difference n l))
             (setq mask1 (lshift mask1 l))
             (setq mask2 (lshift mask2 l)))
            (t (print u)
               (print (list "unknown operand field type:" cf)))))
      (setq instrs
        (nconc instrs
         (list (list mask1 mask2 name fn))))
       (return nil)))

(setq instrfields
  '((reg . 3)(effa . 6)
    (bit . 1) (data3 . 3) (data4 . 4) (data8 . 8) (data12 . 12)))

(de crackeffa(p1 p2 p3 pc l)
     % l=nil means that the length attribute can be found in the
     % standard position (bits 6 and 7 of p1)
   (prog (mode reg text adr regname regnr wl scale size disp)
      (setq mode (logand (lshift p1 -3) 7))
      (setq reg (logand p1 7))
      (when (eq mode 7)
            (cond
              ((eq reg   000)(setq adr `((, p2) ":w"))
                             (setq jaddr p2)
                             (setq ilength (plus ilength 2)))
	      ((eq reg   001)(setq adr `((,(plus (times p2 65536) p3)) ":l"))
                             (setq jaddr (plus (times p2 65536) p3))
                             (setq ilength (plus ilength 4)))
              ((eq reg     4)(when (null l)(setq l (logand (lshift p1 -6)3)))
                      (cond ((eq l 0)
                             (setq adr `(# ,p2))
                             (setq ilength (plus ilength 2)))
                            ((eq l 1)
                             (when (greaterp p2 32767) % 16#7fff
                                   (setq p2 (difference p2 65536))) %16#10000
                             (setq adr `(# ,p2 ))
                             (setq ilength (plus ilength 2)))
                            ((eq l 2)
                             (setq adr `(# ,(plus (times p2 65536) p3)))
                             (setq ilength (plus ilength 4)))))
              ((eq reg     2) (setq adr `(# ,(plus pc p2)))
                             (setq ilength (plus ilength 2)))
          ))
      (setq text
        (cond
         ((eq mode 0)`( D , reg))
         ((eq mode 1)`( A , reg))
         ((eq mode 2) (when (eq reg 7) (setq !*comment "(frame 0)"))
                     `( "(" A ,reg ")"))
         ((eq mode 3)`( "(" A ,reg ")+" ))
         ((eq mode 4)`( "-(" A ,reg ")" ))
         ((eq mode 5) (setq ilength (plus ilength 2))
                      (when (greaterp p2 32000)
                            (setq p2 (difference 65536 p2)))
                      (when (eq reg 7)
                            (setq !*comment (list "frame" (quotient p2 4))))
                     `(  "(" ,p2 ",A" ,reg ")" ))
         ((eq mode 6)
                      (setq ilength (plus ilength 2))
                      (setq regname (if (zerop (lshift p2 -15)) 'd 'a))
                      (setq regnr (logand (lshift p2 -12) 7))
                      (setq wl (if (zerop(logand (lshift p2 -11)1))
                                         ":w" "l"))
                      (setq scale (logand (lshift p2 -9) 3))
                      (setq scale (expt 2 scale))
                      (setq disp  (logand p2 255))
                      (when (greaterp disp 128)
                            (setq disp (difference 256 disp)))
                    (if (not (eq scale 1))
                     `( "(" ,disp ",a" ,reg ","
                          ,regname ,regnr "." ,size "*" ,scale ")" )
                     `( "(" ,disp ",a" ,reg ","
                          ,regname ,regnr ")" )))
         (t          adr)))
      (return text)))

(de crackeffa2(p1 p2 p3 pc l)
   % linke effa: Parameter umgedreht
   (prog (mode reg text adr)
      (setq p1 (lshift p1 -6))
      (setq mode (logand (lshift p1 -3) 7))
      (setq reg (logand p1 7))
      (return (crackeffa(plus (lshift reg 3) mode) p2 p3 pc l))))

(de formatimmediate(pat p1 p2 p3 addr)
    (prog (u name size op1)
       (setq size (logand (lshift p1 -6) 3))
       (setq ilength(if (eq size 2) 6 4))
       (setq op1 (formatdata p2 p3 size))
       (if (eq ilength 4)
           (progn (setq p2 p3) (setq p3 (getwrd (plus addr 4))))
           (progn (setq p2 (getwrd (plus addr 4)))
                  (setq p3 (getwrd (plus addr 6))) ))
       (setq u (crackeffa p1 p2 p3 addr nil))
       (setq name (caddr pat))
       (return `((name . , name)(lth . ,ilength)
                 (op1 .(# ,op1))  (op2 ., u )))))

(de formatdata (p1 p2 size)
    (cond((eq size 0)(logand p1 255))
         ((eq size 1) p1)
         (t (plus (times p1 65536) p2)) ))

(de formateffaeffa (pat p1 p2 p3 addr)
    (prog (name op1 op2 size)
       (setq size (logand (lshift p1 -12) 3))
       (setq size  % special size coding for move-operation
          (cond ((eq size 1) 0)
                ((eq size 3) 1)
                (t         size)))
       (setq ilength 2)
       (setq op1 (crackeffa p1 p2 p3 addr size))
       (when (eq ilength 4)
             (setq p2 p3)
             (setq p3 (getwrd (plus2 addr 6))))
       (when (eq ilength 6)
             (setq p2 (getwrd (plus2 addr 6)))
             (setq p3 (getwrd (plus2 addr 8))))
       (setq op2 (crackeffa2 p1 p2 p3 addr size))
       (setq name (caddr pat))
       (return `((name . , name)(lth . ,ilength) (op1 ., op1 )
		 (op2 ., op2)))))

(de formateffa(pat p1 p2 p3 addr)
    (prog (u name)
       (setq ilength 2)
       (setq u (crackeffa p1 p2 p3 addr nil))
       (setq name (caddr pat))
       (return `((name . , name)(lth . ,ilength) (op1 ., u )))))


(de formatmovem(pat p1 p2 p3 addr)
    (prog (u name dir reglist i sz)
       (setq ilength 4)
       (setq sz (if (zerop (logand (lshift p1 -6))) 1 2))
       (setq dir(logand (lshift p1 -10)))
       (setq u (crackeffa p1 p3 (getwrd (plus addr 2)) addr sz))
       (setq i 0)
       (while (lessp i 7)
              (when (not (zerop (logand (lshift p2 (minus i))1)))
                    (when reglist (setq reglist (cons "," reglist)))
                    (setq reglist (cons (nth dregs (add1 i))reglist)))
              (setq i (add1 i)))
       (setq p2 (lshift p2 -8))
       (setq i 0)
       (while (lessp i 7)
              (when (not (zerop (logand (lshift p2 (minus i))1)))
                    (when reglist (setq reglist (cons "," reglist)))
                    (setq reglist (cons (nth aregs (add1 i))reglist)))
              (setq i (add1 i)))
       (setq reglist (cons ")" reglist))
       (setq reglist (cons "(" (reverse reglist)))
       (setq p2 (lshift p2 -8))
       (setq name (caddr pat))
       (return
         (if (eq dir 0)
         `((name . , name)(lth . ,ilength) (op2 .,reglist)(op1 ., u ))
         `((name . , name)(lth . ,ilength) (op1 .,reglist)(op2 ., u)) ))))

(de formatregeffa(pat p1 p2 p3 addr)
    (prog (u op1 name)
       (setq ilength 2)
       (setq op1 (logand (lshift p1 -9) 7))
       (setq op1 (nth  dregs (add1 op1)))
       (setq u (crackeffa p1 p2 p3 addr nil))
       (setq name (caddr pat))
       (return `((name . , name)(lth . ,ilength) (op2 ., op1) (op1 ., u )))))

(de formatregeffaword(pat p1 p2 p3 addr)
    (prog (u op1 name)
       (setq ilength 2)
       (setq op1 (logand (lshift p1 -9) 7))
       (setq op1 (nth  dregs (add1 op1)))
       (setq u (crackeffa p1 p2 p3 addr 2))
       (setq name (caddr pat))
       (return `((name . , name)(lth . ,ilength) (op2 ., op1) (op1 ., u )))))

(de formatregeffalong(pat p1 p2 p3 addr)
    (prog (u op1 name)
       (setq ilength 2)
       (setq op1 (logand (lshift p1 -9) 7))
       (setq op1 (nth  dregs (add1 op1)))
       (setq u (crackeffa p1 p2 p3 addr 2))
       (setq name (caddr pat))
       (return `((name . , name)(lth . ,ilength) (op2 ., op1) (op1 ., u )))))



(de formatmoveq(pat p1 p2 p3 addr)
    (prog (u op1 name q)
       (setq ilength 2)
       (setq q (logand p1 255)) % 2#11111111
       (setq u (logand (lshift p1 -9) 7))
       (setq u (nth  dregs (add1 u)))
       (setq name (caddr pat))
       (return `((name . , name)(lth . ,ilength)
             (op1 . (# ,q)) (op2 . ,u))) ))

(de formatregeffainv(pat p1 p2 p3 addr)
    (prog (u op1 name)
       (setq ilength 2)
       (setq op1 (logand (lshift p1 -9) 7))
       (setq op1 (nth  dregs (add1 op1)))
       (setq u (crackeffa p1 p2 p3 addr nil))
       (setq name (caddr pat))
       (return `((name . , name)(lth . ,ilength) (op1 ., op1) (op2 ., u )))))


(de formataregeffa(pat p1 p2 p3 addr)
    (prog (u op1 name)
       (setq ilength 2)
       (setq op1 (logand (lshift p1 -9) 7))
       (setq op1 (nth aregs (add1 op1)))
       (setq u (crackeffa p1 p2 p3 addr nil))
       (setq name (caddr pat))
       (return `((name . , name)(lth . ,ilength) (op2 ., op1) (op1 ., u )))))


(de formataregeffaword(pat p1 p2 p3 addr)
    (prog (u op1 name)
       (setq ilength 2)
       (setq op1 (logand (lshift p1 -9) 7))
       (setq op1 (nth  aregs (add1 op1)))
       (setq u (crackeffa p1 p2 p3 addr 2))
       (setq name (caddr pat))
       (return `((name . , name)(lth . ,ilength) (op2 ., op1) (op1 ., u )))))

(de formataregeffalong(pat p1 p2 p3 addr)
    (prog (u op1 name)
       (setq ilength 2)
       (setq op1 (logand (lshift p1 -9) 7))
       (setq op1 (nth  aregs (add1 op1)))
       (setq u (crackeffa p1 p2 p3 addr 2))
       (setq name (caddr pat))
       (return `((name . , name)(lth . ,ilength) (op2 ., op1) (op1 ., u )))))


(de formatareg(pat p1 p2 p3 addr)
    (prog (u op1 name)
       (setq ilength 2)
       (setq op1 (logand p1 7))
       (setq op1 (nth  aregs (add1 op1)))
       (setq name (caddr pat))
       (return `((name . , name)(lth . ,ilength) (op1 ., op1) ))))

(de formatreg(pat p1 p2 p3 addr)
    (prog (u op1 name)
       (setq ilength 2)
       (setq op1 (logand p1 7))
       (setq op1 (nth  dregs (add1 op1)))
       (setq name (caddr pat))
       (return `((name . , name)(lth . ,ilength) (op1 ., op1) ))))

(de formatnop (pat p1 p2 p3 addr)
    (prog (u op1 name)
       (setq !*hardjump t)
       (setq ilength 2)
       (setq name (caddr pat))
       (return `((name . , name)(lth . ,ilength)(op1 . " ") ))))

(de formatjmp(pat p1 p2 p3 addr)
    (prog (u name)
       (setq jaddr nil)
       (setq ilength 2)
       (setq u (crackeffa p1 p2 p3 addr nil))
       (when (and (numberp jaddr) (or (null fktend) (greaterp jaddr fktend)))
             (setq fktend jaddr))
       (setq !*hardjump t)
       (setq name (caddr pat))
       (when (eq name 'JSR) (setq !*hardjump nil))
       (return `((name . , name) (addr ., jaddr) (lth . ,ilength)
		 (op1 ., u )))))


(de formatbra(pat p1 p2 p3 addr)
    (prog (u name)
       (setq addr (plus addr 2))
       (setq !*hardjump t)
       (setq jaddr nil)
       (setq ilength 2)
       (setq u (logand p1 255))  % 2#11111111
       (cond ((eq u 0) (setq ilength 4)
                       (setq u p2)
                       (when (greaterp u 32767)   % 16#7fff
                             (setq u (difference 65535   u)))) % 16#ffff
             ((eq u 255  ) (setq ilength 6)
                           (setq u (plus p3 (lshift p2 16))))
             ((greaterp u 127)(setq u (difference u 256))))
       (setq name (caddr pat))
       (setq jaddr (plus addr u))
       (when (eq name 'BSR) (setq !*hardjump nil))
       (when (and (numberp jaddr) (or (null fktend) (greaterp jaddr fktend)))
             (setq fktend jaddr))
       (return `((name . , name)(lth . ,ilength)
                (addr .,jaddr)(op1 . ,jaddr))) ))

(de formatbcc (pat p1 p2 p3 addr)
    (prog (u name cc)
       (setq addr (plus addr 2))
       (setq jaddr nil)
       (setq ilength 2)
       (setq u (logand p1 255))
       (cond ((eq u 0) (setq ilength 4)
                       (setq u p2)
                       (when (greaterp u 32767)
                             (setq u (difference 65535 u))))
             ((eq u 255  ) (setq ilength 6)
                           (setq u (plus p3 (lshift p2 16))))
             ((greaterp u 127  )(setq u (difference u 256))))
       (setq cc (logand (lshift p1 -8) 15))  % 2#1111
       (setq jaddr (plus addr u))
       (when (and (numberp jaddr) (or (null fktend) (greaterp jaddr fktend)))
             (setq fktend jaddr))
       (setq name (cdr (assoc cc
          '((4 .BCC)(5 .BCS)(7 .BEQ)(12 .BEQ)(14 .BGT)(2 .BHI)(15 .BLE)
            (3 .BLS)(13 .BLT)(11 .BMI)(6 .BNE)(10 .BPL)(8 .BVC)(9 .BVS)))))
       (return `((name . , name)(lth . ,ilength)
                (addr .,jaddr)(op1 . ,jaddr ))) ))


(de formatqeffa(pat p1 p2 p3 addr)
    (prog (u name op1)
       (setq ilength 2)
       (setq op1 (logand (lshift p1 -8) 15))
       (setq u (crackeffa p1 p2 p3 addr nil))
       (setq name (caddr pat))
       (return `((name . , name)(lth . ,ilength)
               (op1 ., op1)(op2 ., u )))))

(de formatcondeffa (pat p1 p2 p3 addr)
        (formatqeffa pat p1 p2 p3 addr))

(de formatshiftrotate(pat p1 p2 p3 addr)
    (prog (u name op2 op1 count dr size ir type reg)
       (setq ilength 2)
       (setq reg   (logand p1 7))
       (setq type  (logand (lshift p1 -3) 3))
       (setq ir    (logand (lshift p1 -5) 1))
       (setq size  (logand (lshift p1 -6) 3))
       (setq dr    (logand (lshift p1 -8) 1))
       (setq count (logand (lshift p1 -9) 7))
       (setq name
           (cdr (assoc type
                  (if (eq dr 0)
                     '((0 . AST)(1 . LSR)(2 . ROXR)(3 . ROR))
                     '((0 . ASR)(1 . LSL)(2 . ROXL)(3 . ROL))))))
       (when (eq size 3)
             (return (formatshiftrotatemem (name p1 p2 p3 ))))
       (setq name (compress (append (explode name)
                                   (cond ((eq size 0) 'b)
                                         ((eq size 1) 'w)
                                         ((eq size 2) 'l)
                                         (t '! )))))
       (setq op1 (nth dregs (add1 reg)))
       (setq op2 (if (izerop ir)
                     (if (izerop count) 8 count)
                     (nth dregs (add1 count))))
       (return `((name . , name)(lth . ,ilength)
                 (op1 ., op1)(op2 ., op2))) ))


(de formateffamultlong(pat p1 p2 p3 addr)
    (prog (u name size op1 di dh typ)
       (setq ilength 4)
       (setq op1 ` ("d" ,(logand  p1 7)))
       (setq di ` ("d" ,(logand (lshift p2 -12) 7)))
       (setq dh ` ("d" ,(logand p2 7)))
       (setq size (logand (lshift p2 -10) 1))
       (setq typ  (logand (lshift p2 -11) 1))
       (setq name (caddr pat))
       (setq name (bldmsg "%w%w.l" name (if (eq typ 1) "s" "u")))
       (return
          (if (eq size 0)
             `((name . , name)(lth . ,ilength)
                 (op1 . ,op1) (op2 . ,di))
             `((name . , name)(lth . ,ilength)
                 (op1 . ,op1) (op2 . (,dh ":" ,di)))))))

(de formateffadivlong(pat p1 p2 p3 addr)
    (prog (u name size op1 dq dr typ)
       (setq ilength 4)
       (setq op1 ` ("d" ,(logand  p1 7)))
       (setq dq ` ("d" ,(logand (lshift p2 -12) 7)))
       (setq dr ` ("d" ,(logand p2 7)))
       (setq size (logand (lshift p2 -10) 1))
       (setq typ  (logand (lshift p2 -11) 1))
       (setq name (caddr pat))
       (setq name (bldmsg "%w%w%w.l" name (if (eq typ 1) "s" "u")
                                          (if (eq size 0) "" "l")))
       (return
             `((name . , name)(lth . ,ilength)
                 (op1 . ,op1) (op2 . (,dr ":" ,dq))))))


(setq instrs nil)

(instr ORb   (0 0 0 0 0 0 0 0 0 0 effa) formatimmediate )
(instr ORw   (0 0 0 0 0 0 0 0 0 1 effa) formatimmediate )
(instr ORl   (0 0 0 0 0 0 0 0 1 0 effa) formatimmediate )
(instr ANDb  (0 0 0 0 0 0 1 0 0 0 effa) formatimmediate )
(instr ANDw  (0 0 0 0 0 0 1 0 0 1 effa) formatimmediate )
(instr ANDl  (0 0 0 0 0 0 1 0 1 0 effa) formatimmediate )
(instr SUBb  (0 0 0 0 0 1 0 0 0 0 effa) formatimmediate )
(instr SUBw  (0 0 0 0 0 1 0 0 0 1 effa) formatimmediate )
(instr SUBl  (0 0 0 0 0 1 0 0 1 0 effa) formatimmediate )
(instr ADDb  (0 0 0 0 0 1 1 0 0 0 effa) formatimmediate )
(instr ADDw  (0 0 0 0 0 1 1 0 0 1 effa) formatimmediate )
(instr ADDl  (0 0 0 0 0 1 1 0 1 0 effa) formatimmediate )
(instr EORb  (0 0 0 0 1 0 1 0 0 0 effa) formatimmediate )
(instr EORw  (0 0 0 0 1 0 1 0 0 1 effa) formatimmediate )
(instr EORl  (0 0 0 0 1 0 1 0 1 0 effa) formatimmediate )
(instr CMPb  (0 0 0 0 1 1 0 0 0 0 effa) formatimmediate )
(instr CMPw  (0 0 0 0 1 1 0 0 0 1 effa) formatimmediate )
(instr CMPl  (0 0 0 0 1 1 0 0 1 0 effa) formatimmediate )

(instr MOVEb (0 0 0 1 effa effa)        formateffaeffa)
(instr MOVEl (0 0 1 0 effa effa)        formateffaeffa)
(instr MOVEw (0 0 1 1 effa effa)        formateffaeffa)

(instr NEGXb  (0 1 0 0 0 0 0 0 0 0 effa) formateffa)
(instr NEGXl  (0 1 0 0 0 0 0 0 0 1 effa) formateffa)
(instr NEGXw  (0 1 0 0 0 0 0 0 1 0 effa) formateffa)
(instr CHKl   (0 1 0 0 reg 1 0 0 effa)   formatregeffa)
(instr CHKw   (0 1 0 0 reg 1 1 0 effa)   formatregeffa)
(instr LEA    (0 1 0 0 reg 1 1 1 effa)   formataregeffa)

(instr CLRb   (0 1 0 0 0 0 1 0 0 0 effa) formateffa)
(instr CLRw   (0 1 0 0 0 0 1 0 0 1 effa) formateffa)
(instr CLRl   (0 1 0 0 0 0 1 0 1 0 effa) formateffa)
(instr NEGb   (0 1 0 0 0 1 0 0 0 0 effa) formateffa)
(instr NEGl   (0 1 0 0 0 1 0 0 0 1 effa) formateffa)
(instr NEGw   (0 1 0 0 0 1 0 0 1 0 effa) formateffa)
(instr NOTb   (0 1 0 0 0 1 1 0 0 0 effa) formateffa)
(instr NOTl   (0 1 0 0 0 1 1 0 0 1 effa) formateffa)
(instr NOTw   (0 1 0 0 0 1 1 0 1 0 effa) formateffa)

(instr MOVEtoSR
              (0 1 0 0 0 1 1 0 1 1 effa) formateffa)
(instr NBCD   (0 1 0 0 1 0 0 0 0 0 effa) formateffa)

(instr SWAP   (0 1 0 0 1 0 0 0 0 1 0 0 0 reg) formatareg)
(instr EXTw   (0 1 0 0 1 0 0 0 1 0 0 0 0 reg) formatreg)
(instr EXTl   (0 1 0 0 1 0 0 0 1 1 0 0 0 reg) formatreg)
(instr EXTBl  (0 1 0 0 1 0 0 1 1 1 0 0 0 reg) formatreg)
(instr MOVEMw (0 1 0 0 1 bit 0 0 1 0 effa)    formatmovem)
(instr MOVEMl (0 1 0 0 1 bit 0 0 1 1 effa)    formatmovem)
(instr TSTb   (0 1 0 0 1 0 1 0 0 0 effa)      formateffa)
(instr TSTw   (0 1 0 0 1 0 1 0 0 1 effa)      formateffa)
(instr TSTl   (0 1 0 0 1 0 1 0 1 0 effa)      formateffa)
(instr TAS    (0 1 0 0 1 0 1 0 1 1 effa)      formateffa)
(instr MUL    (0 1 0 0 1 1 0 0 0 0 effa)      formateffamultlong)
(instr DIV    (0 1 0 0 1 1 0 0 0 1 effa)      formateffadivlong)
(instr LINKw  (0 1 0 0 1 1 1 0 0 1 0 1 0 reg) formatareg)
(instr UNLK   (0 1 0 0 1 1 1 0 0 1 0 1 1 reg) formatareg)

(instr NOP    (0 1 0 0 1 1 1 0 0 1 1 1 0 0 0 1) formatnop)
(instr STOP   (0 1 0 0 1 1 1 0 0 1 1 1 0 0 1 0) formatnop)
(instr RTE    (0 1 0 0 1 1 1 0 0 1 1 1 0 0 1 1) formatnop)
(instr RTS    (0 1 0 0 1 1 1 0 0 1 1 1 0 1 0 1) formatnop)
(instr RTR    (0 1 0 0 1 1 1 0 0 1 1 1 0 1 0 1) formatnop)

(instr JSR   (0 1 0 0 1 1 1 0 1 0 effa)        formatjmp)
(instr JMP   (0 1 0 0 1 1 1 0 1 1 effa)        formatjmp)

(instr ADDQb  (0 1 0 1 data3 0 0 0 effa)       formatqeffa)
(instr ADDQw  (0 1 0 1 data3 0 0 1 effa)       formatqeffa)
(instr ADDQl  (0 1 0 1 data3 0 1 0 effa)       formatqeffa)

(instr SUBQb (0 1 0 1 data3 1 0 0 effa)        formatqeffa)
(instr SUBQw (0 1 0 1 data3 1 0 1 effa)        formatqeffa)
(instr SUBQl (0 1 0 1 data3 1 1 0 effa)        formatqeffa)

(instr S     (0 1 0 1 data4 1 1 effa)          formatcondeffa)
(instr DB    (0 1 0 1 data4 1 1 0 0 1 reg)     formatcondreg)

(instr BRA   (0 1 1 0 0 0 0 0 data8)           formatbra)
(instr BSR   (0 1 1 0 0 0 0 1 data8)           formatbra)
(instr B     (0 1 1 0 data4 data8)             formatbcc)

(instr MOVEQ (0 1 1 1 reg 0 data8)             formatmoveq)

(instr ORb    (1 0 0 0 reg 0 0 0 effa)         formatregeffa)
(instr ORw    (1 0 0 0 reg 0 0 1 effa)         formatregeffaword)
(instr ORl    (1 0 0 0 reg 0 1 0 effa)         formatregeffalong)
(instr ORb    (1 0 0 0 reg 1 0 0 effa)         formatregeffainv)
(instr ORw    (1 0 0 0 reg 1 0 1 effa)         formatregeffainv)
(instr ORl    (1 0 0 0 reg 1 1 0 effa)         formatregeffainv)

(instr DIVUw (1 0 0 0 reg 0 1 1 effa)          formatregeffaword)
(instr DIVSw (1 0 0 0 reg 1 1 1 effa)          formatregeffaword)

(instr SUBb   (1 0 0 1 reg 0 0 0 effa)         formatregeffa)
(instr SUBw   (1 0 0 1 reg 0 0 1 effa)         formatregeffaword)
(instr SUBl   (1 0 0 1 reg 0 1 0 effa)         formatregeffalong)
(instr SUBb   (1 0 0 1 reg 1 0 0 effa)         formatregeffainv)
(instr SUBw   (1 0 0 1 reg 1 0 1 effa)         formatregeffainv)
(instr SUBl   (1 0 0 1 reg 1 1 0 effa)         formatregeffainv)
(instr SUBAw  (1 0 0 1 reg 0 1 0 effa)         formataregeffaword)
(instr SUBAl  (1 0 0 1 reg 1 1 1 effa)         formataregeffalong)

(instr CMPMb  (1 0 1 1 reg 1 0 0 0 0 1 reg)    formatregreg)
(instr CMPMw  (1 0 1 1 reg 1 0 1 0 0 1 reg)    formatregreg)
(instr CMPMl  (1 0 1 1 reg 1 1 0 0 0 1 reg)    formatregreg)

(instr CMPb   (1 0 1 1 reg 0 0 0 effa)         formatregeffa)
(instr CMPw   (1 0 1 1 reg 0 0 1 effa)         formatregeffaword)
(instr CMPl   (1 0 1 1 reg 0 1 0 effa)         formatregeffalong)
(instr CMPAw  (1 0 1 1 reg 0 1 1 effa)         formataregeffaword)
(instr CMPAl  (1 0 1 1 reg 1 1 1 effa)         formataregeffalong)
(instr EORw   (1 0 1 1 reg 1 0 1 effa)         formatregeffaword)
(instr EORl   (1 0 1 1 reg 1 1 0 effa)         formatregeffalong)

(instr ANDb   (1 1 0 0 reg 0 0 0 effa)         formatregeffa)
(instr ANDw   (1 1 0 0 reg 0 0 1 effa)         formatregeffaword)
(instr ANDl   (1 1 0 0 reg 0 1 0 effa)         formatregeffalong)
(instr ANDb   (1 1 0 0 reg 1 0 0 effa)         formatregeffainv)
(instr ANDw   (1 1 0 0 reg 1 0 1 effa)         formatregeffainv)
(instr ANDl   (1 1 0 0 reg 1 1 0 effa)         formatregeffainv)

(instr MULUw  (1 1 0 0 reg 0 1 1 effa)         formatregeffaword)
(instr MULSw  (1 1 0 0 reg 1 1 1 effa)         formatregeffaword)


(instr ADDb   (1 1 0 1 reg 0 0 0 effa)         formatregeffa)
(instr ADDw   (1 1 0 1 reg 0 0 1 effa)         formatregeffaword)
(instr ADDl   (1 1 0 1 reg 0 1 0 effa)         formatregeffalong)
(instr ADDb   (1 1 0 1 reg 1 0 0 effa)         formatregeffainv)
(instr ADDw   (1 1 0 1 reg 1 0 1 effa)         formatregeffainv)
(instr ADDl   (1 1 0 1 reg 1 1 0 effa)         formatregeffainv)
(instr ADDAw  (1 1 0 1 reg 0 1 0 effa)         formataregeffa)
(instr ADDAl  (1 1 0 1 reg 1 1 1 effa)         formataregeffa)

(instr SHL    (1 1 1 0 data12)                 formatshiftrotate)


