%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           i860-disassemble.sl
% Title:          i860 PSL disassembler 
% Author:         Winfried Neun, ZIB Berlin
% Created:        27 May 1992
% Mode:           Lisp
% Package:        Compiler
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% main driver
%            either  (disassemble 'hugo)      or
%                    (disassemble '16#4711)

(de disassemble (fkt)
   (prog (name word op op2 i ll )  
      
     (setq name fkt  i 0  labellist* nil)
     
     (when (and (idp fkt) (wlessp (idinf fkt) maxsymbols))
		 (when (not(fcodep fkt)) (return nil))
		 (setq fkt (codeinf (getfcodepointer fkt))))

     (setq outputbase* 16)
     (prin2t (list " ** i860 function " name " at" fkt)) (terpri)
     (setq outputbase* 10)
   loop
     (setq word (wgetv fkt i))
     (unixputn (loc (wgetv fkt i)))
     (console-print-string "     ")
     (unixputn word)
     (console-print-string "     ")
     (console-print-string (id2string (getv instrvect (wshift word -26))))
     (console-newline)
     (setq i (wplus2 1 i))
     (when (eq 0 word) (return NIL))
     (go loop)
)   ) 

(de print-format-1 (fkt i dummy)
   (prog (word word1)
     (setq word (wgetv fkt i))
     (prinbn (wplus2 fkt (wtimes2 4 i)) 8)
     (tab 12)
     (prinbn word 8)
     (tab 24)
     (setq word (ashift (wshift word 2) -2))  % for gc
     (prin2 "call")
     (tab 35)
     (setq word1 (wshift word 2)) % byte address
     (setq word1 (wplus2 (wplus2 fkt (wtimes2 i 4)) word1)) % absolute
     (if (withinsymfnc word1)
	 (setq word (int2id (wquotient (wdifference word1 symfnc) 8)))
	 (progn
	   (setq word (wquotient word1 4))
	   (setq word (assoc word labellist*))
	   (when word (setq word (cdr word)))))

     (prinb word)
     (terpri)) )          % some comments later

(de print-format-2 (fkt i)
   (prog (opcode pattern op2 rd a cond imm22 comment)
     (setq comment "")
     (setq imm22 (wgetv fkt i))
     (setq op2   (wshift (wshift imm22 7) -29))
     (setq rd    (regname (wshift (wshift imm22 2) -27)))
     (setq cond  (wshift (wshift imm22 3) -28))
     (setq a     (wshift (wshift imm22 2) -31))
     (prinbn (wplus2 fkt (wtimes2 4 i)) 8) 
     (tab 12) 
     (prinbn imm22 8) 
     (tab 24)
     (setq imm22 (ashift (wshift imm22 10) -10))
     (setq opcode  op2)
     (setq pattern (getv sun4-instructions opcode))
     (cond ((null pattern) (setq pattern "?????????"))
	   ((atom pattern) NIL)
	   ((and (pairp pattern) (idp (car pattern)))
		(return (apply (car pattern)
		   (list (cdr pattern) rd a cond imm22 i fkt))))
	   ((pairp pattern) (if (and (pairp (cdr pattern)) (eq 0 i))
		               (setq pattern (cadr pattern))
		               (setq pattern (car pattern))))
     )
     (when (pairp pattern)     % some special cases for sethi 
       (when (and (equal (car pattern) "sethi")
		  (eq imm22 0)   % NOP case
		  (eq rd '!r0))
		(return (prin2t "nop")))
       (if (and (equal (car pattern) "sethi") (eq rd '!r16))
	      % Alloc 'marker'
	     (setq comment (bldmsg "Alloc %d" (wand imm22 127)))
        (when (equal (car pattern) "sethi")          % SYMVAL access ?
	     (setq comment (wgetv fkt (wplus2 i 1)))
	     (setq comment (wor (wshift imm22 10) (wand comment 16#3ff)))
		     % the address
	     (if (and (wgreaterp comment SYMVAL)
		      (wlessp comment (wplus2 SYMVAL (wtimes2 4 maxsymbols))))
		(setq comment (symnam (wquotient
		                             (wdifference comment symval) 4)))
		(setq comment ""))))
       (prinbl
	(subla
	 (pair '(opcode op2 rd a cond imm22 comment)
		(list opcode op2 rd a cond imm22 comment))
	 pattern)))

     (terpri)))  % bis auf weiteres
 
(de label-format-1 (fkt i dummy)
   (prog (word word1)
     (setq word (wgetv fkt i))
     (setq word (ashift (wshift word 2) -2))  % for gc
     (setq word1 (wshift word 2)) % byte address
     (setq word1 (wplus2 (wplus2 fkt (wtimes2 i 4)) word1))
     (return
     (if (withinsymfnc word1)     nil
       (when (and (posintp word1) (wleq word1 nextbps))
	(push (cons (wquotient word1 4) "nowhere") labellist*))))))
 
(de label-format-2 (fkt i)
   (prog (imm22 word word1)
     (setq imm22 (wgetv fkt i))
     (setq imm22 (ashift (wshift imm22 10) -10))
     (setq word1 (wshift imm22 2)) % byte address
     (setq word1 (wplus2 (wplus2 fkt (wtimes2 i 4))
		         word1))
(return
     (if (withinsymfnc word1)     nil
      (if (eq 0 imm22) nil  % a NOP instr 
	  (push (cons (wquotient word1 4) "anywhere") labellist*))))))

(de sun4-cond-instr-handler (vect rd a cond imm22 i fkt)
   (prog (comment pattern)
     (setq comment "")
     (setq cond (wor (wshift a 4) cond))
     (setq pattern (getv vect cond))
     (cond ((null pattern) (setq pattern "?????????"))
	   ((atom pattern) NIL)
     )
     (setq imm22 (wtimes2 imm22 4))
     (setq vect (wplus2 imm22 (wplus2 (wtimes2 i 4) fkt)))
     (if (withinsymfnc vect)
	 (setq imm22
	     (int2id (wquotient (wdifference vect symfnc) 8)))
	 (progn
	   (setq vect (wquotient vect 4))
	   (setq vect (assoc vect labellist*))
	   (when vect (setq imm22 (cdr vect)))))

     (when (pairp pattern)
       (prinbl
	(subla
	 (pair '(rd a cond disp22 comment)
		(list rd a cond imm22 comment))
	 pattern)))
 
     (terpri))) 

(de print-format-3 (op fkt ii)
   (prog (pattern opcode i frd crd op3 rd asi rs1 rs2 opf simm13 comment)
     (setq comment "")
     (setq simm13 (wgetv fkt ii))
     (setq rd    (regname (wshift (wshift simm13 2) -27)))
     (setq frd   (assoc (wshift (wshift simm13 2) -27) *nice-frame-NAMES*))
       (when frd   (setq comment (cdr frd)))
     (setq frd   (fregname rd))
     (setq crd   (cregname rd))
     (setq op3   (wshift (wshift simm13 7) -26))
     (setq rs1   (regname (wshift (wshift simm13 13) -27)))
     (setq i     (assoc (wshift (wshift simm13 13) -27) *nice-frame-NAMES*))
       (when i   (if (equal comment "") (setq comment (cdr i))
		     (setq comment (bldmsg "%w/%w" comment (cdr i)))))
     (setq i     (wshift (wshift simm13 18) -31))
     (setq asi   (wshift (wshift simm13 19) -24))
     (setq rs2   (regname (wand simm13 16#1f)))
     (setq opf   (wshift (wshift simm13 18) -23))
     (prinbn (wplus2 fkt (wtimes2 4 ii)) 8)
     (tab 12)
     (prinbn simm13 8) 
     (setq simm13 (ashift (wshift simm13 19) -19))
     (tab 24)
     (setq opcode (wor (wshift op 6) op3))
     (setq pattern (getv sun4-instructions opcode))
     (cond ((null pattern) NIL)
	   ((idp (car pattern))
		 (return (apply (car pattern)
		            (list rd rs1 rs2 opf ))))
	   ((atom pattern) NIL)
	   ((pairp pattern) (if (and (pairp (cdr pattern)) (eq 0 i))
		               (setq pattern (cadr pattern))
		               (setq pattern (car pattern))))
     )
	   % special case : the VIV variables access
     (when (and (eq i 1) (eq rs1 '!r7) (assoc simm13 *viv-names*))
	   (setq comment (cdr (assoc simm13 *viv-names*))))
    
     (when (pairp pattern)
       (prinbl
	(subla
	 (pair '(opcode op3 rd i asi rs1 rs2 opf simm13 f_rd c_rd comment)
		(list opcode op3 rd i asi rs1 rs2 opf simm13 frd crd comment))
	 pattern))
       (terpri))

     ))  % bis auf weiteres

(de sun4-float-instructionhandler (rd rs1 rs2 opf)
   (prog (f_rd f_rs1 f_rs2 pattern)
      (setq f_rd (fregname rd))
      (setq f_rs1 (fregname rs1))
      (setq f_rs2 (fregname rs2)) 
      (setq pattern (assoc opf sun4-float-instructions))
      (when (not pattern) (return nil))
      (setq pattern (cdr pattern))
      (when (pairp pattern)
       (prinbl
	(subla
	 (pair '(f_rd f_rs1 f_rs2 ) (list f_rd f_rs1 f_rs2 ))
	 pattern))
       (terpri))
))

(de prinbl (l)                  % binary (octal) printing of a list
     (mapc l (function prinb)))
 
(de prinb (it)                  % binary (octal) printing of an item
     (cond ((numberp it)(prinbo it))
	   ((eq it 't1) (tab 35))
	   ((eq it 't2) (tab 60))
	   (t       (prin2 it))))
 
(de prinbo (it)
      (cond ((wlessp it 0) (prin2 "-") (prinbo (wminus it)))
	    ((wgeq it 10)  (prin2 "x'") (prinbn it 1))
    (t            (prinbn it 1))))
 
(de prinbn (it n)                % print an octal number
     (cond ((and (weq it 0) (wleq n 0)) nil)
	    (t (progn
		  (prinbn (wshift it -4) (wplus2 n -1))
		  (prindig (wand it 15))
)    )     )  )
 
(de prindig (dig)        % print a numeric digit
     (if (wlessp dig 10) (writeChar  (wplus2 dig 48))
	   (writechar (wplus2 65 (wdifference dig 10)))))
 
(de labelsort (l)        % sort labels to ascending sequence
   (prog (so)
     (setq so (labelsort1 l nil))
     (return so)))
 
(de labelsort1 (rest sorted)
    (cond ((null rest) sorted)
	   (t (labelsort1(cdr rest) (labelsortin (car rest) sorted))) ))
 
(de labelsortin (object l)
    (cond ((null l) (list object))
	  ((wgreaterp (car object)(caar l))
		         (cons (car l) (labelsortin object (cdr l))) )
	  (t (cons object l)) ))
(fluid '(instrvect))

(setq instrvect (mkvect 64))
(compiletime (setq instrvect (mkvect 64)))

(dm define-instr (u)
   (prog (name opcode spec1 spec2)
	(setq name (cadr u))
	(setq opcode (caddr u))
        (when (cdddr u) (setq spec1 (cadddr u)))
	(when (cddr (cddr u)) (setq spec2 (cadr (cdddr u))))
	(if (eq spec1 'I)
		(return `(progn
			(putv instrvect ,opcode ',name)
			(putv instrvect ,(add1 opcode) ',name)))
	(return `(putv instrvect ,opcode ',name)))))

(define-instr	ld!.b	2#000000 I)
(define-instr	ld!.s	2#000100 I)
(define-instr	ld!.l	2#000100 I)

(define-instr	st!.b	2#000011 format3)
(define-instr	st!.s	2#000111 format3)
(define-instr	st!.l	2#000111 format3)

(define-instr	ixfr	2#000010)
(define-instr	reserved1	2#000110)

(define-instr	fld!.x	2#001000 I)
(define-instr	fst!.x	2#001010 I)
(define-instr	flush	2#001101)
(define-instr	pst!.d	2#001111)
(define-instr	ld!.c	2#001100)
(define-instr	st!.c	2#001110)

(define-instr	bri 	2#010000)
(define-instr	trap	2#010001)

(define-instr	FP_escape	2#010010)
(define-instr	core_escape	2#010011 core)

(define-instr	bte	2#010110 I format4)
(define-instr	btne	2#010100 I format4)
(define-instr	pfld!.y	2#011100 I)

(define-instr	ctrl1	2#011000)
(define-instr	ctrl2	2#011001)
(define-instr	ctrl3	2#011010)
(define-instr	ctrl4	2#011011)
(define-instr	ctrl5	2#011100)
(define-instr	ctrl6	2#011101)
(define-instr	ctrl7	2#011110)
(define-instr	ctrl8	2#011111)

(define-instr	addu	2#100000 I)
(define-instr	adds	2#100100 I)
(define-instr	subu	2#100010 I)
(define-instr	subs	2#110110 I)
(define-instr	shl 	2#101000 I)
(define-instr	shr 	2#101010 I)
(define-instr	shrd	2#101100)
(define-instr	bla	2#101101 format3)
(define-instr	shra	2#101110 I)
(define-instr	and 	2#110000 I)
(define-instr	andh	2#110010 I)
(define-instr	andnot	2#110100 I)
(define-instr	andnoth	2#110110 I)
(define-instr	or	2#111000 I)
(define-instr	orh	2#111010 I)
(define-instr	xor	2#111100 I)
(define-instr	xorh	2#111110 I)

(define-instr	reserved2	2#110010)
(define-instr	reserved3	2#110110)
(define-instr	reserved4	2#111010)
(define-instr	reserved5	2#111110)
